/* $Id: textures.c,v 1.6 2004/04/01 23:23:50 garbled Exp $ */

/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <sys/param.h>
#include <sys/time.h>
#include <inttypes.h>

#include <rm/rm.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "pngloader.h"
#include "textures.h"
#include "defines.h"
#include "utils.h"
#include "xmlparse.h"
#include "weatherdefs.h"

int numgtextures;

static void parsegtexture(xmlDocPtr doc, xmlNodePtr cur, int num, tmap_t *txt)
{
    char *c;

    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "path"))
	    parsestring(doc, cur, &txt[num].path);
	if (!strcmp(cur->name, "name"))
	    parsestring(doc, cur, &txt[num].name);
	if (!strcmp(cur->name, "type")) {
	    c = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	    if (!strncmp(c, "JPEG", 4))
		txt[num].tformat = TFORM_JPEG;
	    free(c);
	}
	cur = cur->next;
    }
}

static void parse_skycond(xmlDocPtr doc, xmlNodePtr cur, skycond_t *cond)
{
    int r, g, b, a, gotfog=0;

    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "mapfile")) {
	    parsestring(doc, cur, &cond->path);
	    if (strncmp(strrchr(cond->path, '.'), ".jpg", 4) == 0)
		cond->tformat = TFORM_JPEG;
	    if (strncmp(strrchr(cond->path, '.'), ".png", 4) == 0)
		cond->tformat = TFORM_PNG;
	}
	if (!strcmp(cur->name, "color")) {
	    parsequadint(doc, cur, &r, &g, &b, &a);
	    cond->sky.r = (float)r/255.0;
	    cond->sky.g = (float)g/255.0;
	    cond->sky.b = (float)b/255.0;
	    cond->sky.a = (float)a/255.0;
	}
	if (!strcmp(cur->name, "fog")) {
	    parsequadint(doc, cur, &r, &g, &b, &a);
	    cond->fog.r = (float)r/255.0;
	    cond->fog.g = (float)g/255.0;
	    cond->fog.b = (float)b/255.0;
	    cond->fog.a = (float)a/255.0;
	    gotfog++;
	}
	cur = cur->next;
    }
    if (!gotfog)
	cond->fog = cond->sky;
}

tmap_t *parsegtextures(char *filename, char *path)
{
    int textures;
    tmap_t *txts;
    xmlDocPtr doc;
    xmlNodePtr cur;
    char file[MAX_PATH];

    sprintf(file, "%s/%s", path, filename);
    doc = xmlParseFile(file);
    if (doc == NULL)
        bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        fprintf(stderr, "empty document\n");
        xmlFreeDoc(doc);
        bailout();
    }
    if (strcmp(cur->name, "GroundTexture")) {
        fprintf(stderr, 
		"document of the wrong type, root node != GroundTexture");
        xmlFreeDoc(doc);
        bailout();
    }

    cur = cur->xmlChildrenNode;
    txts = safer_malloc(sizeof(tmap_t));

    textures = 0;
    while (cur != NULL) {
        if (!strcmp(cur->name, "Terrain")) {
            txts = realloc(txts, sizeof(tmap_t)*(textures+2));
            parsegtexture(doc, cur, textures, txts);
	    textures++;
        }
        cur = cur->next;
    }
    numgtextures = textures-1;
    xmlFreeDoc(doc);
    return(txts);
}

/**
   \brief parse the skymaps file
   \param filename file name to parse
   \param path where to find it
   \return an array of skymaps
*/

skymap_t *parse_skymaps(char *filename, char *path)
{
    int curmap, fday, fnight, fdawn, fdusk;
    skymap_t *skys;
    xmlDocPtr doc;
    xmlNodePtr cur, ncur, xcur;
    char file[MAX_PATH], *c;

    sprintf(file, "%s/%s", path, filename);
    doc = xmlParseFile(file);
    if (doc == NULL)
        bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        fprintf(stderr, "empty document\n");
        xmlFreeDoc(doc);
        bailout();
    }
    if (strcmp(cur->name, "Skymaps")) {
        fprintf(stderr, "document of the wrong type, root node != Skymaps");
        xmlFreeDoc(doc);
        bailout();
    }

    cur = cur->xmlChildrenNode;
    skys = safer_malloc(sizeof(skymap_t) * SKY_NROF);

    curmap = 0;
    while (cur != NULL) {
	if (!strcmp(cur->name, "weather")) {
	    ncur = cur->xmlChildrenNode;
	    while (ncur != NULL) {
		if (!strcmp(ncur->name, "sky")) {
		    c = xmlGetProp(ncur, "clouds");
		    curmap = atoi(c);
		    free(c);
		    if (curmap > SKY_NROF) {
			fprintf(stderr, "Too many sky cloud conditions!");
			bailout();
		    }
		    xcur = ncur->xmlChildrenNode;
		    fday = fnight = fdawn = fdusk = 0;
		    while (xcur != NULL) {
			if (!strcmp(xcur->name, "day")) {
			    skys[curmap].day = smalloc(skycond_t);
			    parse_skycond(doc, xcur, skys[curmap].day);
			    fday++;
			}
			if (!strcmp(xcur->name, "night")) {
			    skys[curmap].night = smalloc(skycond_t);
			    parse_skycond(doc, xcur, skys[curmap].night);
			    fnight++;
			}
			if (!strcmp(xcur->name, "dusk")) {
			    skys[curmap].dusk = smalloc(skycond_t);
			    parse_skycond(doc, xcur, skys[curmap].dusk);
			    fdusk++;
			}
			if (!strcmp(xcur->name, "dawn")) {
			    skys[curmap].dawn = smalloc(skycond_t);
			    parse_skycond(doc, xcur, skys[curmap].dawn);
			    fdawn++;
			}
			xcur = xcur->next;
		    }
		    if (!fday) {
			fprintf(stderr, "Skymap missing day condition!\n");
			bailout();
		    }
		    /* if some are missing, fill them in with what we have */
		    if (!fnight)
			skys[curmap].night = skys[curmap].day;
		    if (!fdusk && fdawn)
			skys[curmap].dusk = skys[curmap].dawn;
		    if (!fdawn && fdusk)
			skys[curmap].dawn = skys[curmap].dusk;
		    if (!fdusk)
			skys[curmap].dusk = skys[curmap].day;
		    if (!fdawn)
			skys[curmap].dawn = skys[curmap].day;
		}
		ncur = ncur->next;
	    }
	}
        cur = cur->next;
    }
    xmlFreeDoc(doc);
    return(skys);
}
