/* $Id: language.c,v 1.2 2004/03/19 07:36:49 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#include <ctype.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "defines.h"
#include "xmlparse.h"
#include "utils.h"
#include "language.h"

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif

/**
   \file language.c
   \brief Game language translation functions
   \author Tim Rightnour
*/

extern TAILQ_HEAD(, _lang_t) langs;

/** 
    \brief parse a lang file
    \param path path to language files
    \param lang the language structure to populate
*/

void parse_language(char *path, lang_t *lang)
{
    char file[MAX_PATH];
    xmlDocPtr doc;
    xmlNodePtr cur;
    int lmall;

    sprintf(file, "%s/%s.lang", path, lang->name);
    doc = xmlParseFile(file);
    if (doc == NULL) {
	LOG(LOG_WARNING, "Warning, cannot find Language file %s", file);
	return;
    }
    /* if we got this file, we found one */

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        LOG(LOG_WARNING, "Language file %s is empty", file);
        xmlFreeDoc(doc);
	return;
    }
    if (strcmp(cur->name, "Language")) {
        LOG(LOG_WARNING,"%s: root node != Language", file);
        xmlFreeDoc(doc);
	return;
    }

    /* we malloc and realloc these in chunks of 100, to keep from having to
       realloc on every single line in the language file */

    lang->orig = safer_malloc(sizeof(char *) * SYL_MALLOC_CHUNK);
    lang->lang = safer_malloc(sizeof(char *) * SYL_MALLOC_CHUNK);
    lmall = SYL_MALLOC_CHUNK;

    lang->syls = 0;
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "syl")) {
	    if (lang->syls == lmall) {
		lang->orig = realloc(lang->orig, sizeof(char *) *
				     (lmall+SYL_MALLOC_CHUNK));
		lang->lang = realloc(lang->lang, sizeof(char *) *
				     (lmall+SYL_MALLOC_CHUNK));
		lmall += SYL_MALLOC_CHUNK;
	    }
	    lang->orig[lang->syls] = xmlGetProp(cur, "orig");
	    lang->lang[lang->syls] = 
		xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	    if (lang->orig[lang->syls] != NULL &&
		lang->lang[lang->syls] != NULL)
		lang->syls++;
	}
	cur = cur->next;
    }
    xmlFreeDoc(doc);
}

/**
   \brief Parse the main languages file and all the languages
   \param path path to the language files
   \param filename full path filename to languages file.
*/

void parse_languages(char *path, char *filename)
{
    xmlDocPtr doc;
    xmlNodePtr cur;
    int numlangs;
    lang_t *lang;

    doc = xmlParseFile(filename);
    if (doc == NULL) {
	LOG(LOG_WARNING, "Warning, cannot find Languages file %s", filename);
	return;
    }
    /* if we got this file, we found one */

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        LOG(LOG_WARNING, "Languages file %s is empty", filename);
        xmlFreeDoc(doc);
	return;
    }
    if (strcmp(cur->name, "Languages")) {
        LOG(LOG_WARNING,"%s: root node != Languages", filename);
        xmlFreeDoc(doc);
	return;
    }

    /* language 0 is common */

    numlangs = 1;
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "lang")) {
	    if (TAILQ_EMPTY(&langs))
		TAILQ_INIT(&langs);
	    lang = smalloc(lang_t);
	    lang->id = numlangs;
	    numlangs++;
	    parsestring(doc, cur, &lang->name);
	    parse_language(path, lang);
	    TAILQ_INSERT_HEAD(&langs, lang, next);
	}
	cur = cur->next;
    }
    xmlFreeDoc(doc);
}

/**
   \brief Translate a string into a language
   \param input original string
   \param langid language number to translate to
   \param knowledge how well do we know it?
   \return a malloc'd translated string, which must be freed
*/

char *translate_string(char *input, int langid, int knowledge)
{
    char *buf, *p;
    int maxstring, found=0, j;
    lang_t *lang;
    size_t offset, bufsize, bufpos, off, pos;

    if (knowledge >= 100)
	return(input);

    TAILQ_FOREACH(lang, &langs, next) {
	if (lang->id == langid) {
	    found++;
	    break;
	}
    }
    if (!found)
	return(input);

    maxstring = (lang->syls*(100-knowledge))/100;

    /* XXX
       we really cannot know how large the buffer will be, as the translated
       texts might extend the buffer.  We use a minimum of 256 however.
       This will work, as long as we do not have any translations that are
       primarily more than 3 times in length.
    */
    bufsize = MAX(256, strlen(input) * 3);
    buf = safer_malloc(sizeof(char) * bufsize);
    offset = 0;
    bufpos = 0;
    strcpy(buf, "");

    while (*(input+offset) && (bufsize-bufpos > 1)) {
	for (j=0, found=0; j < maxstring; j++) {
	    if (!strncasecmp(lang->orig[j], input+offset, strlen(lang->orig[j]))) {
		strncat(buf, lang->lang[j], bufsize-bufpos-1);
		off = strlen(lang->orig[j]);
		if (off) {
		    p = input+offset;
		    for (pos=0; pos <= off; pos++, p++)
			if (isupper(*p))
			    buf[pos+bufpos] = toupper(buf[pos+bufpos]);
		    offset += off;
		} else
		    offset++;
		if (strlen(lang->lang[j]))
		    bufpos += strlen(lang->lang[j]);
		found++;
		break;
	    }
	}
	if (!found && (bufsize-bufpos > 1)) {
	    strncat(buf, input+offset, 1);
	    offset++;
	    bufpos++;
	}
    }
    /* lets be sure, just in case we took it right to the edge */
    buf[bufsize-1] = '\0';
    return(buf);
}
