/* $Id: netparser.c,v 1.18 2004/04/01 23:23:50 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <stdarg.h>
#include <inttypes.h>
#include <ctype.h>
#if defined(__linux__)
#define __USE_GNU
#endif
#include <errno.h>
#include <math.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif
#include <sys/param.h>

#include "defines.h"
#include "shstr.h"
#include "utils.h"
#include "netparser.h"
#include "uid.h"

/**
   \file netparser.c
   \brief Network parsing engine
   \author Tim Rightnour
   \author Dave Watson
*/

extern argtable_t commtable[];
extern size_t comm_size;

/**
   \brief the argument to type mapping
*/

argtable_t argtable[] = {
    {"id", ID, PTUID},
    {"nrof", NROF, PTINT},
    {"draw", DRAW, PTINT},
    {"st", STATE, PTINT},
    {"nm", NAME, PTCHAR},
    {"nmp", NAMEPL, PTCHAR},
    {"mdl", MODEL, PTCHAR},
    {"fx", FACEX, PTFLOAT},
    {"fy", FACEY, PTFLOAT},
    {"fz", FACEZ, PTFLOAT},
    {"dx", DIRX, PTFLOAT},
    {"dy", DIRY, PTFLOAT},
    {"dz", DIRZ, PTFLOAT},
    {"cx", COORDX, PTFLOAT},
    {"cy", COORDY, PTFLOAT},
    {"cz", COORDZ, PTFLOAT},
    {"sx", SCALEX, PTFLOAT},
    {"sy", SCALEY, PTFLOAT},
    {"sz", SCALEZ, PTFLOAT},
    {"sex", SEX, PTINT},
    {"type", TYPE, PTINT},
    {"sub", SUBTYPE, PTINT},
    {"wght", WEIGHT, PTINT},
    {"lvl", LEVEL, PTINT},
    {"xp", EXP, PTINT},
    {"mat", MATERIAL, PTINT},
    {"val", VALUE, PTINT},
    {"str", STR, PTSTAT},
    {"dex", DEX, PTSTAT},
    {"con", CON, PTSTAT},
    {"wis", WIS, PTSTAT},
    {"int", INT, PTSTAT},
    {"per", PER, PTSTAT},
    {"fame", FAME, PTINT},
    {"appr", APPEAR, PTINT},
    {"hp", HP, PTINT},
    {"mhp", MAXHP, PTINT},
    {"sp", SP, PTINT},
    {"msp", MAXSP, PTINT},
    {"stm", STAMINA, PTINT},
    {"mstm", MAXSTAMINA, PTINT},
    {"food", FOOD, PTINT},
    {"wat", WATER, PTINT},
    {"ac", AC, PTINT},
    {"spd", SPEED, PTINT},
    {"act", ACT, PTINT},
    {"cont", CONTAINER, PTUID},
    {"buk", BUCKET, PTINT},
    {"eq", EQUIP, PTUID},
    {"sl", SLOT, PTINT},
    {"tar", TARGET, PTUID},
    {"v", VELOCITY, PTFLOAT},
    {"user", USERNAME, PTCHAR},
    {"pass", PASSWORD, PTCHAR},
    {"plexists", PLAYER_EXISTS, PTINT},
    {"lang", LANGUAGE, PTINT},
    {"data", DATA, PTCHAR},
    {"volume", VOLUME, PTINT},
    {"servid", SRVID, PTCHAR},
    {"ipaddr", IPADDR, PTCHAR},
    {"portnum", PORTNUM, PTINT},
    {"desc", DESC, PTCHAR},
    {"mapinfo", MAPINFO, PTCHAR},
    {"hostname", HOSTNAME, PTCHAR},
    {"srvversion", SRVVERSION, PTCHAR},
    {"webpage", WEBPAGE, PTCHAR},
    {"ownername", OWNERNAME, PTCHAR},
    {"owneremail", OWNEREMAIL, PTCHAR},
    {"lastup", LASTUPDATE, PTLONG},
    {"firstup", FIRSTUPDATE, PTLONG},
    {"numusers", NUMUSERS, PTINT},
    {"userhwm", USERHWM, PTINT},
    {"maxusers", MAXUSERS, PTINT},
    {"preferip", PREFERIP, PTINT},
    {"width", WIDTH, PTINT},
    {"height", HEIGHT, PTINT},
    {"bin", BINARY, PTINT},
    {"mapid", MAPID, PTINT},

};

/** \brief size of the args table */
const size_t args_size = sizeof(argtable) / sizeof(argtable_t);


/**
   \brief compare two argtable members
   \param a first member
   \param b second member
   \return see strcmp()
*/
static int compare_argtable(const void *a, const void *b)
{
    return strcmp(((argtable_t *)a)->name, ((argtable_t *)b)->name);
}

/**
   \brief initialize the argument and command tables
*/
void init_argcomm(void)
{
    qsort((char *)argtable, args_size, sizeof(argtable_t), compare_argtable);
    qsort((char *)commtable, comm_size, sizeof(argtable_t), compare_argtable);
}

/**
   \brief parse a command from the network
   \param words a list of words to parse
   \param command command name
   \param count how many words
   \return a pargs_t of arguments
   \note if this returns NULL, the command was invalid.
   \note Frees words upon exit
*/

pargs_t *parse_command(char **words, int *command, int count)
{
    argtable_t *asp, dummy;
    char *cp, *tmp, *p;
    size_t span;
    int i, numargs, cur;
    pargs_t *args;

    if (words == NULL || words[0] == NULL) {
	for (i=0; i < count && words[i] != NULL; i++)
	    free(words[i]);
	free(words);
	return(NULL);
    }

    /* the command is words[0] */
    for (p=words[0]; *p; p++) {
	*p = tolower(*p);
    }
    dummy.name = words[0];
    asp = (argtable_t *)bsearch((void *)&dummy, (void *)commtable,
				comm_size, sizeof(argtable_t),
				compare_argtable);
    if (asp)
	*command = asp->num;
    else
	*command = 0;

    numargs = count;

    args = safer_malloc(sizeof(pargs_t)*numargs);
    cur = 0;
    for (i=1; i < count && words[i] != NULL && *words[i]; i++) {
	span = strcspn(words[i], ":");
        tmp = strchr(words[i], ':');
        tmp++;
	cp = words[i];
	cp[span] = '\0';
	for (p=cp; *p; p++) {
	    *p = tolower(*p);
	}

	dummy.name = cp;
	asp = (argtable_t *)bsearch((void *)&dummy, (void *)argtable,
				    args_size, sizeof(argtable_t),
				    compare_argtable);
	if (asp) {
	    args[cur].cword = asp->num;
	    args[cur].type = asp->type;
	    switch (asp->type) {
	    case PTDOUBLE: args[cur].arg.d = atof(tmp); break;
	    case PTFLOAT: args[cur].arg.f = (float)atof(tmp); break;
	    case PTCHAR: args[cur].arg.c = add_string(tmp); break;
	    case PTINT: args[cur].arg.i = atoi(tmp); break;
	    case PTUID:
		if (!uid_sread(tmp, &args[cur].arg.uid))
		    uid_clear(&args[cur].arg.uid);
		break;
	    case PTSTAT: {
		p = strchr(tmp, '/');
		if (p == NULL) {
		    args[cur].arg.s.base = atoi(tmp);
		    args[cur].arg.s.pct = 0;
		} else {
		    p = '\0';
		    p++;
		    args[cur].arg.s.pct = atoi(p);
		    args[cur].arg.s.base = atoi(tmp);
		}
		break; }
	    }
	    cur++;
	} else {
	    LOG(LOG_ERROR, "Invalid command recieved: %s", words[i]);
	}
    }

    args[cur].cword = -1;
    for (i=0; i < count && words[i] != NULL; i++)
	free(words[i]);
    free(words);

    return(args);
}

/**
   \brief parse a string into a command
   \param buf string to parse
   \param arg_count how many words did we find?
   \return a list of words
*/

char **parse_netcommand(char *buf, int *arg_count)
{
    char **ret_str;
    char *buf2 = 0, *buf2_head = 0;
    int i = 0;

    if (!*buf)
        return(0);

    buf2 = (char *)safer_malloc(strlen(buf)+1);
    buf2_head = buf2;
    ret_str = (char **)safer_malloc(sizeof(char *) * 2);

    for (;;) {
        if (*buf == '\"') {
            buf++;

            while (*buf && *buf != '\"') {
                if (*buf == '\\') {
                    buf++;

                    if (*buf == '\"')
                        *buf2++ = *buf++;
                    else {
                        buf--;
                        *buf2++ = *buf++;
                    }
                } else
                    *buf2++ = *buf++;
            }

            if (*buf == '\"')
                buf++;
        }

        if (*buf == ' ' || !*buf) {
            *buf2 = '\0';
            buf2 = buf2_head;
            ret_str = (char **)realloc(ret_str, sizeof(char *) * (i + 1));
            ret_str[i] = strdup(buf2);
            buf2 = buf2_head;
            i++;

            if (!*buf) {
                /* Free temporary buffer */
                free(buf2);
                *arg_count = i;
                return(ret_str);
            } else
                buf++;
        } else {
            *buf2++ = *buf++;
        }
    }
}

/**
   \brief Find an argument in the args by it's ID
   \param args arguments to search
   \param id id to look for
   \return argument number if found, -1 if not
*/

int find_arg_by_id(pargs_t *args, int id)
{
    int i;

    for (i=0; args[i].cword != -1; i++)
	if (args[i].cword == id)
	    return(i);
    return(-1);
}
