/* $Id: pngloader.c,v 1.7 2004/03/17 03:43:53 garbled Exp $ */

/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <sys/param.h>

#include <png.h>

#include "defines.h"
#include "pngloader.h"
#include "utils.h"

/**
   \file pngloader.c
   \brief PNG loading routines
   \author Tim Rightnour
*/

/**
   \brief parse an 8bit PNG
   \param file_name full path filename to parse
   \param width width of image
   \param height height of image
   \return a 2d array of uint8_t's
   parse a png file containing map data, such as an elevation map, or
   a terrain map.
*/

uint8_t **parse_pngmap8bit(char* file_name, int width, int height)
{
    char header[8];
    FILE *fp;
    png_structp png_ptr;
    png_infop info_ptr;
    int number_of_passes, x, y, j;
    png_bytep *row_pointers;
    png_byte color_type;
    png_byte bit_depth;
    uint8_t **map;

    /* open file and test for it being a png */
    fp = fopen(file_name, "rb");
    if (!fp)
	LOG(LOG_FATAL, "file %s could not be opened for reading.", file_name);

    fread(header, 1, 8, fp);
    if (png_sig_cmp(header, 0, 8))
	LOG(LOG_FATAL, "file %s is not a PNG file.", file_name);

    /* initialize stuff */
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	
    if (!png_ptr)
	bailout();

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
	bailout();

    if (setjmp(png_jmpbuf(png_ptr)))
	bailout();

    png_init_io(png_ptr, fp);
    png_set_sig_bytes(png_ptr, 8);

    png_read_info(png_ptr, info_ptr);

    if (width != png_get_image_width(png_ptr, info_ptr) || height != png_get_image_height(png_ptr, info_ptr))
	LOG(LOG_FATAL, "width/height of map do not match that of image");
    color_type = png_get_color_type(png_ptr, info_ptr);
    bit_depth = png_get_bit_depth(png_ptr, info_ptr);

    number_of_passes = png_set_interlace_handling(png_ptr);
    png_read_update_info(png_ptr, info_ptr);

    /* read file */
    if (setjmp(png_jmpbuf(png_ptr)))
	bailout();

    row_pointers = (png_bytep*) malloc(sizeof(png_bytep) * height);
    for (y=0; y<height; y++)
	row_pointers[y] = (png_byte*) malloc(png_get_rowbytes(png_ptr, info_ptr));

    png_read_image(png_ptr, row_pointers);

    if (png_get_color_type(png_ptr, info_ptr) != PNG_COLOR_TYPE_GRAY)
	LOG(LOG_FATAL, "Image %s is not grayscale", file_name);

    map = safer_malloc(sizeof(uint8_t *) * height);
    for (y = 0; y < height; y++)
	map[y] = safer_malloc(sizeof(uint8_t) * width);
    for (y=0, j=height-1; y<height && j>=0; y++, j--) {
	png_byte* row = row_pointers[y];
	for (x=0; x<width; x++) {
	    png_byte* ptr = &(row[x*1]);
/*	    printf("Pixel at position [ %d - %d ] has the following RGB values: %d - %d - %d\n", x, y, ptr[0], ptr[1], ptr[2]); */
	    map[x][j] = ptr[0];
	}
    }
    return(map);
}

/**
   \brief Parse a 24bit png into a 16bit map
   \param width width of image
   \param height height of image
   \return a 2d array of int16_t's
   \note this parses a 24bit image into a 16bit value set. Blue is ignored
*/

int16_t **parse_pngmap16bit(char* file_name, int width, int height)
{
    char header[8];
    FILE *fp;
    png_structp png_ptr;
    png_infop info_ptr;
    int number_of_passes, x, y;
    png_bytep *row_pointers;
    png_byte color_type;
    png_byte bit_depth;
    int16_t **map;

    /* open file and test for it being a png */
    fp = fopen(file_name, "rb");
    if (!fp)
	LOG(LOG_FATAL, "file %s could not be opened for reading.", file_name);

    fread(header, 1, 8, fp);
    if (png_sig_cmp(header, 0, 8))
	LOG(LOG_FATAL, "file %s is not a PNG file.", file_name);

    /* initialize stuff */
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	
    if (!png_ptr)
	bailout();

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
	bailout();

    if (setjmp(png_jmpbuf(png_ptr)))
	bailout();

    png_init_io(png_ptr, fp);
    png_set_sig_bytes(png_ptr, 8);

    png_read_info(png_ptr, info_ptr);

    if (width != png_get_image_width(png_ptr, info_ptr) || height != png_get_image_height(png_ptr, info_ptr))
	LOG(LOG_FATAL, "width/height of map do not match that of image");
    color_type = png_get_color_type(png_ptr, info_ptr);
    bit_depth = png_get_bit_depth(png_ptr, info_ptr);

    number_of_passes = png_set_interlace_handling(png_ptr);
    png_read_update_info(png_ptr, info_ptr);

    /* read file */
    if (setjmp(png_jmpbuf(png_ptr)))
	bailout();

    row_pointers = (png_bytep*) malloc(sizeof(png_bytep) * height);
    for (y=0; y<height; y++)
	row_pointers[y] = (png_byte*) malloc(png_get_rowbytes(png_ptr, info_ptr));

    png_read_image(png_ptr, row_pointers);

    if (png_get_color_type(png_ptr, info_ptr) != PNG_COLOR_TYPE_RGB)
	LOG(LOG_FATAL, "Image %s is not RGB", file_name);

    map = safer_malloc(sizeof(int16_t *) * height);
    for (y = 0; y < height; y++)
	map[y] = safer_malloc(sizeof(int16_t) * width);
    for (y=0; y<height; y++) {
	png_byte* row = row_pointers[y];
	for (x=0; x<width; x++) {
	    png_byte* ptr = &(row[x*3]);
/*	    printf("Pixel at position [ %d - %d ] has the following RGB values: %d - %d - %d\n", x, y, ptr[0], ptr[1], ptr[2]); */
	    map[y][x] = (ptr[0] + 256*ptr[1]) - 32767;
	}
    }
    return(map);
}

/*
  \brief parse a 32bit RGBA PNG image
   \param width width of image
   \param height height of image
   \return an array of uint8_t's
   \note This is for texture use (ie: client)
*/

uint8_t *parse_pngRGBA(char* file_name, int *width, int *height)
{
    char header[8];
    FILE *fp;
    png_structp png_ptr;
    png_infop info_ptr;
    int number_of_passes, y, row_stride, indx;
    png_bytep *row_pointers;
    png_byte color_type;
    png_byte bit_depth;
    uint8_t *map;

    /* open file and test for it being a png */
    fp = fopen(file_name, "rb");
    if (!fp)
	LOG(LOG_FATAL, "file %s could not be opened for reading.", file_name);

    fread(header, 1, 8, fp);
    if (png_sig_cmp(header, 0, 8))
	LOG(LOG_FATAL, "file %s is not a PNG file.", file_name);

    /* initialize stuff */
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	
    if (!png_ptr)
	bailout();

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
	bailout();

    if (setjmp(png_jmpbuf(png_ptr)))
	bailout();

    png_init_io(png_ptr, fp);
    png_set_sig_bytes(png_ptr, 8);

    png_read_info(png_ptr, info_ptr);

    /*if (width != png_get_image_width(png_ptr, info_ptr) || height != png_get_image_height(png_ptr, info_ptr))
      LOG(LOG_FATAL, "width/height of map do not match that of image");*/
    color_type = png_get_color_type(png_ptr, info_ptr);
    bit_depth = png_get_bit_depth(png_ptr, info_ptr);

    number_of_passes = png_set_interlace_handling(png_ptr);
    png_read_update_info(png_ptr, info_ptr);

    /* read file */
    if (setjmp(png_jmpbuf(png_ptr)))
	bailout();

    row_pointers = (png_bytep*) malloc(sizeof(png_bytep) * png_get_image_height(png_ptr, info_ptr));
    for (y=0; y < png_get_image_height(png_ptr, info_ptr); y++)
	row_pointers[y] = (png_byte*) malloc(png_get_rowbytes(png_ptr, info_ptr));

    png_read_image(png_ptr, row_pointers);

    if (png_get_color_type(png_ptr, info_ptr) != PNG_COLOR_TYPE_RGBA)
	LOG(LOG_FATAL, "Image %s is not RGBA", file_name);

    row_stride = png_get_image_width(png_ptr, info_ptr) * 4;

    map = safer_malloc(sizeof(uint8_t) * png_get_image_height(png_ptr, info_ptr) *
		       png_get_image_width(png_ptr, info_ptr) * 4);

/*    indx = row_stride * (png_get_image_height(png_ptr, info_ptr)-1); */
    indx = 0;
    for (y=0; y < png_get_image_height(png_ptr, info_ptr); y++) {
	png_byte* row = row_pointers[y];
	memcpy(map+indx, row, row_stride);
	indx += row_stride;
    }

    *height = png_get_image_height(png_ptr, info_ptr);
    *width = png_get_image_width(png_ptr, info_ptr);

    return(map);
}
