/* $Id: quadric.c,v 1.4 2004/03/11 08:07:19 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

/**
   \file quadric.c
   \brief Quadric functions used by the lodder
   \author Bruno 'Beosil' Heidelberger
   \author Tim Rightnour

    This code is based directly on Quadric.cpp from Cal3D.
    Copyright (C) 2001, 2002 Bruno 'Beosil' Heidelberger
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <inttypes.h>
#include <math.h>
#include <time.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include "defines.h"
#include "utils.h"
#include "quadric.h"

/**
   \brief add quadric b to a
   \param a target quadric
   \param b quadric to add
*/

void add_quadric(quadric_t *a, quadric_t *b)
{
    a->m_aa += b->m_aa;
    a->m_ab += b->m_ab;
    a->m_ac += b->m_ac;
    a->m_ad += b->m_ad;

    a->m_bb += b->m_bb;
    a->m_bc += b->m_bc;
    a->m_bd += b->m_bd;

    a->m_cc += b->m_cc;
    a->m_cd += b->m_cd;

    a->m_dd += b->m_dd;

    a->m_area += b->m_area;
}

/**
   \brief set quadric a to the value of b
   \param a target quadric
   \param b source quadric
*/

void set_quadric(quadric_t *a, quadric_t *b)
{
    a->m_aa = b->m_aa;
    a->m_ab = b->m_ab;
    a->m_ac = b->m_ac;
    a->m_ad = b->m_ad;

    a->m_bb = b->m_bb;
    a->m_bc = b->m_bc;
    a->m_bd = b->m_bd;

    a->m_cc = b->m_cc;
    a->m_cd = b->m_cd;

    a->m_dd = b->m_dd;

    a->m_area = b->m_area;
}

/**
   \brief Evaluate a quadric
   \param x float
   \param y float
   \param z float
   \param a quadric to evaluate
   \return evaulation
*/

float eval_quadric(float x, float y, float z, quadric_t *a)
{
    return(x * x * a->m_aa + 2.0 * x * y * a->m_ab + 2.0 * x * z *
	a->m_ac + 2.0 * x * a->m_ad + y * y * a->m_bb + 2.0 * y * z *
	a->m_bc + 2.0 * y * a->m_bd + z * z * a->m_cc + 2.0 * z *
	a->m_cd + a->m_dd);
}

/**
   \brief Scale a quadric
   \param factor factor to scale by
   \param a quadric to scale
*/

void scale_quadric(float factor, quadric_t *a)
{
    a->m_aa *= factor;
    a->m_ab *= factor;
    a->m_ac *= factor;
    a->m_ad *= factor;

    a->m_bb *= factor;
    a->m_bc *= factor;
    a->m_bd *= factor;

    a->m_cc *= factor;
    a->m_cd *= factor;

    a->m_dd *= factor;
}

/**
   \brief Set the plane of a quadric
   \param nx x plane
   \param ny y plane
   \param nz z plane
   \param d float
   \param area area
   \param a quadric to set plane of
*/

void set_quadric_plane(float nx, float ny, float nz, float d, float area,
	quadric_t *a)
{
    a->m_aa = nx * nx;
    a->m_ab = nx * ny;
    a->m_ac = nx * nz;
    a->m_ad = nx * d;

    a->m_bb = ny * ny;
    a->m_bc = ny * nz;
    a->m_bd = ny * d;

    a->m_cc = nz * nz;
    a->m_cd = nz * d;

    a->m_dd =  d * d;

    a->m_area = area;
}

/**
   \brief set the edge of a quadric
   \param x1 float
   \param y1 float
   \param z1 float
   \param x2 float
   \param y2 float
   \param z2 float
   \param nx float
   \param ny float
   \param nz float
   \param a quadric to set edge of
*/

void set_quadric_edge(float x1, float y1, float z1, float x2, float y2,
	float z2, float nx, float ny, float nz, quadric_t *a)
{
    float dirx, diry, dirz;
    float vx, vy, vz;
    float length, d, weight;

    dirx = x2 - x1;
    diry = y2 - y1;
    dirz = z2 - z1;

    vx = ny * dirz - diry * nz;
    vy = nz * dirx - dirz * nx;
    vz = nx * diry - dirx * ny;

    length = sqrt(vx * vx + vy * vy + vz * vz);
    vx /= length;
    vy /= length;
    vz /= length;

    d = -(vx * x2 + vy * y2 + vz * z2);

    set_quadric_plane(vx, vy, vz, d, 1.0, a);

    weight = dirx * dirx + diry * diry + dirz * dirz;

    scale_quadric(weight, a);
}



