/* $Id: stattables.c,v 1.7 2004/03/16 07:42:42 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <time.h>
#include <stdarg.h>
#include <math.h>
#include <sys/param.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include "defines.h"
#include "shstr.h"
#include "utils.h"

/**
   \file stattables.c
   \author Tim Rightnour
   \brief Table defines for object stats
*/

float vision_range[19] = {
    1.0, 3.0, 6.0, 9.0, 12.0, 15.0, 18.0, 21.0, 24.0, 27.0, 30.0,
    33.0, 36.0, 39.0, 40.0, 43.0, 46.0, 49.0, 52.0};
float vision_padd = 10.0;
float maxarmspeed[19];
float maxliftweight[19];


/**
   \brief range of vision for perception
   \param per Perception stat
   \return range
*/
float vision_dist(stat_t per)
{
    if (per.pct)
	return((vision_range[per.base] + vision_padd/(float)per.pct)*SCALE_XZ);
    else
	return((vision_range[per.base])*SCALE_XZ);
}

/**
   \brief range of hearing for perception
   \param per Perception stat
   \return range
*/
float hearing_dist(stat_t per)
{
    if (per.pct)
	return((vision_range[per.base] + vision_padd/(float)per.pct)*SCALE_XZ*3.0);
    else
	return((vision_range[per.base])*SCALE_XZ*3.0);
}

/**
   \brief LOD for a given per at range x 0-1.0
   \param per Perception stat
   \param dist distance in points
   \return 0.0-1.0, level of detail
*/
float lod_dist(stat_t per, float dist)
{
    float lod;

    lod = dist/(vision_dist(per)*3.0/2.0);
    if (lod > 1.0)
	lod = 1.0;

    lod = 1.0-lod;

    return(lod);
}

/**
   \brief Returns the maximum weight a player can carry. (Kg)
   \note init_max_lift_weight() should be called before this.
   \param str Strength stat
   \return the maximum weight a player can carry. (Kg)
*/
float max_lift_weight(stat_t str)
{
    return maxliftweight[str.base] + 18*str.pct*(51.0/50.0);
}

/**
   \brief initialize the maximum lift weight table
*/
void init_max_lift_weight(void)
{
    int i=0;
    maxliftweight[0] = 0.5;
    maxliftweight[1] = 1.0;
    
    for (i=2; i <= 12; i++) 
	maxliftweight[i]=maxliftweight[i-1] + (float)i;
    for (i=13; i <= 15; i++) 
	maxliftweight[i]=maxliftweight[i-1] + (float)i*1.3;
    for (i=16; i <= 18; i++) 
	maxliftweight[i]=maxliftweight[i-1] + (float)i*1.5;
}

/**
   \brief Returns the maximum speed a player can move his arm.
   \note init_max_arm_speed() should be colled before this.
   \param dex Dexterity stat
   \return the maximum speed a player can move his arm.
*/
float max_arm_speed(stat_t dex)
{
    return maxarmspeed[dex.base] + 1.8*(float)dex.pct*(51.0/50.0);
}

/**
   \brief initialize the maximum armspeed table
*/
void init_max_arm_speed(void)
{
    int i=0;
    maxarmspeed[0] = 0.5;
    maxarmspeed[1] = 1.0;
    
    for (i=2; i <= 12; i++) 
	maxarmspeed[i]=maxarmspeed[i-1] + (float)i*0.23;
    for (i=13; i <= 15; i++) 
	maxarmspeed[i]=maxarmspeed[i-1] + (float)i*0.26;
    for (i=16; i <= 18; i++) 
	maxarmspeed[i]=maxarmspeed[i-1] + (float)i*0.32;
}

/**
   \brief Return the encumbrance penalty for a given weight
   \param str Strength stat
   \param carrying_weight Weight carried
   \return encumbrance
*/
float encumbrance_penalty(stat_t str, int carrying_weight)
{
    return 1.0/powf(powf((float)carrying_weight/max_lift_weight(str), 2.0)+1.0, 3.0);
}

/**
   \brief Return the throw penalty for a given item
   \param str Strength stat
   \param item_weight Weight of item to be thrown
   \return penalty
*/
float throw_item_w_penalty(stat_t str, int item_weight)
{
    return 1.0-powf(((float)item_weight/max_lift_weight(str)), 2.0);
}
