/* $Id: utils.c,v 1.18 2004/03/30 06:45:54 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <stdarg.h>
#include <inttypes.h>
#if defined(__linux__)
#define __USE_GNU
#endif
#include <errno.h>
#include <math.h>
#include <sys/param.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include "defines.h"
#include "shstr.h"
#include "utils.h"

/**
   \file utils.c
   \author Tim Rightnour
   \brief General utility functions
*/

FILE *logfile = NULL;

const char *loglevels[LOG_FATAL+1] = {
    "",
    "DEBUG",
    "WARNING",
    "ERROR",
    "FATAL"
};

/**
   \brief open the logfile for writing
*/
void openlog(void)
{
    /* XXX make this a setting */
    logfile = fopen("genecys.log", "a");
    if (logfile == NULL)
	bailout();
}

/**
   \brief close the logfile
*/
void closelog(void)
{
    fclose(logfile);
}

/**
   \brief Log something
   \param severity Severity of log, see LOG_WARNING etc
   \param s format string
   \param ... arguments
*/
void LOG(int severity, const char *s, ...)
{
    char *t, *timestamp;
    time_t now;
    va_list args;

    now = time(NULL);
    t = safer_malloc(sizeof(char)*26);
    timestamp = safer_malloc(sizeof(char)*26);
    t = ctime_r(&now, t); /* cut the day off */
    (void)strncpy(timestamp, t+4, 15); /* cut the year off */
    free(t);

    if (logfile != NULL) {
	va_start(args, s);
	fprintf(logfile, "%s [%s]:", timestamp, loglevels[severity]);
	free(timestamp);
	vfprintf(logfile, s, args);
	fprintf(logfile, "\n");
	fflush(logfile);
	va_end(args);
    } else {
	va_start(args, s);
	fprintf(stderr, "%s [%s]:", timestamp, loglevels[severity]);
	free(timestamp);
	vfprintf(stderr, s, args);
	fprintf(stderr, "\n");
	fflush(stderr);
	va_end(args);
    }
    if (severity == LOG_FATAL)
	exit(EXIT_FAILURE);
}


/**
   \brief exit on a failure
   \param line line number we bailed on
*/
void _bailout(char *file, int line)
{
    const char *pname = "";

#if defined(__NetBSD__) || defined(__DragonFly__) || defined(__FreeBSD__) || defined(__NetBSD__)
    pname = getprogname();
#elif defined(__linux__)
    pname = program_invocation_name;
#endif

    /* the fprintf catches cases where malloc failed, as it will likely fail
       again in LOG() */
    (void)fprintf(stderr, "%s: Failed in %s:%d: %s %d\n",
		  pname, file, line, strerror(errno), errno);
    (void)LOG(LOG_FATAL, "%s: Failed in %s:%d: %s %d\n",
	      pname, file, line, strerror(errno), errno);
}

/**
   \brief return a zero'd malloc with error checking
   \param size size to malloc
   \param line line number we were called from
   \return malloc'd data
*/
void *_safer_malloc(size_t size, char *file, int line)
{
    void *stuff;

    stuff = malloc(size);
    if (stuff == NULL)
	_bailout(file, line);
    memset(stuff, 0, size);
    return(stuff);
}

/*  X East-west ?? height
    Z North-south ?? width
    Y elevation
*/

/**
   \brief Return groundheight at x,z for a given elevation map
   \param map a map array
   \param x x coord
   \param z z coord
   \param h height of map
   \param w width of map
   \return groundheight at x,z
*/

float groundheight_at_xz(uint8_t **map, float x, float z, int h, int w)
{
    int fx, fz;
    float lx, lz, y0, y1;

    fx = (int)x/SCALE_XZ;
    fz = (int)z/SCALE_XZ;
    x /= SCALE_XZ;
    z /= SCALE_XZ;

    if (fx < h-1 && fz < w-1 && fx > -1 && fz > -1) {
	lx = x - fx;
	lz = z - fz;
	/* Bilinearly interpolate from surrounding points */
	y0 = map[fx][fz]*SCALE_Y + (map[fx][fz+1]*SCALE_Y -
				    map[fx][fz]*SCALE_Y) * lz;
    	y1 = map[fx+1][fz]*SCALE_Y + (map[fx+1][fz+1]*SCALE_Y -
				      map[fx+1][fz]*SCALE_Y)*lz;
	return (y0 + (y1 - y0) * lx);
    }
    return(0.0); /* XXX ??? */
}

/**
   \brief returns height of ground at coordinates. usable only on 8bit maps
   \param map a map array
   \param coord coordinates to check
   \param h height of map
   \param w width of map
   \return groundheight at coord
*/

float groundheight_at_coord(uint8_t **map, vertex_t coord, int h, int w)
{
    return(groundheight_at_xz(map, coord.x, coord.z, h, w));
}

/**
   \brief is point x,y within polygon xp,yp ?
   \param npol number of corners
   \param xp x coordinates
   \param yp y coordinates
   \param x x coord to check
   \param y y coord to check
   \return true/false
*/

int pnpoly(int npol, float *xp, float *yp, float x, float y)
{
    int i, j, c = 0;
    for (i = 0, j = npol-1; i < npol; j = i++) {
	if ((((yp[i]<=y) && (y<yp[j])) ||
	     ((yp[j]<=y) && (y<yp[i]))) &&
	    (x < (xp[j] - xp[i]) * (y - yp[i]) / (yp[j] - yp[i]) + xp[i]))

	    c = !c;
    }
    return c;
}

/**
   \brief square a value
   \param f number to square
   \return f^2
*/
static float square(float f) {
    return (f*f);
}

/**
   \brief is line xy1 xy2 intersected by circle xyr, and where?
   \param x1 P1 coordinates (point of line)
   \param y1 P1 coordinates (point of line)
   \param x2 P2 coordinates (point of line)
   \param y2 P2 coordinates (point of line)
   \param x3 P3 coordinates and radius (circle)
   \param y3 P3 coordinates and radius (circle)
   \param r P3 coordinates and radius (circle)
   \param x x intersection coordinate
   \param y y intersection coordinate
   \return This function returns a pointer array which first index indicates
   the number of intersection point, followed by coordinate pairs.

   atelier iebele abel - 2001
   atelier@iebele.nl
   http://www.iebele.nl
   sphere_line_intersection function adapted from:
   http://astronomy.swin.edu.au/pbourke/geometry/sphereline
   Paul Bourke pbourke@swin.edu.au
*/

float *circle_line_intersection(float x1, float y1, float x2, float y2,
				 float x3, float y3, float r)
{
    float a=0.0, b=0.0, c=0.0, mu, i, *p;

    p = (float *)safer_malloc(5*sizeof(float));

    if (x1 == x2 && y1 == y2) {
	i = -1.0;
    } else {
	a = square(x2 - x1) + square(y2 - y1);
	b = 2 * ((x2 - x1)*(x1 - x3) + (y2 - y1)*(y1 - y3)) ;
	c = square(x3) + square(y3) + square(x1) + square(y1) -
	    2 * (x3*x1 + y3*y1) - square(r);
	i =  b * b - 4 * a * c ;
    }
    if (i < 0.0) {
	/* no intersection */
	p[0] = 0.0;
    } else if (i == 0.0) {
	/* one intersection */
	p[0] = 1.0;

	mu = -b/(2*a) ;
	p[1] = x1 + mu*(x2-x1);
	p[2] = y1 + mu*(y2-y1);
	if (!(fabs(p[1]) > fabs(x1) && fabs(p[1]) < fabs(x2)))
	    p[0] = 1.0;
    } else if (i > 0.0) {
	/* two intersections */
	p[0] = 2.0;

	/* first intersection */
	mu = (-b + sqrt(square(b) - 4*a*c)) / (2*a);
	p[1] = x1 + mu*(x2-x1);
	p[2] = y1 + mu*(y2-y1);
	/* second intersection */
	mu = (-b - sqrt(square(b) - 4*a*c)) / (2*a);
	p[3] = x1 + mu*(x2-x1);
	p[4] = y1 + mu*(y2-y1);
	if (!(fabs(p[3]) > fabs(x1) && fabs(p[3]) < fabs(x2)))
	    p[0] = 1.0;
	if (!(fabs(p[1]) > fabs(x1) && fabs(p[1]) < fabs(x2))) {
	    p[0] = 1.0;
	    p[1] = p[3];
	    p[2] = p[4];
	}
    }
    return(p);
}

/**
   \brief find the farthest fall rotation possible
   \param nrof number of line segments
   \param x x coords of each line
   \param y y coords of each line
   \param r radius of circle
   \return angle of rotation
   \note Circle is centered at first point of the line segment.
*/

float farthest_fall_rotation(int nrof, float *x, float *y, float r)
{
    int i;
    float xx, yy, *p;

    xx = yy = -999999.0;

    for (i=0; i < nrof-1; i++) {
	p =circle_line_intersection(x[i], y[i], x[i+1], y[i+1], x[0], y[0], r);
	if (p[0] == 0.0) {
	    free(p);
	    continue;
	} else if (p[0] == 2.0) {
	    if (p[2] > yy) {
		xx = p[1];
		yy = p[2];
	    }
	    if (p[4] > yy) {
		xx = p[3];
		yy = p[4];
	    }
	} else if (p[0] == 1.0) {
	    if (p[2] > yy) {
		xx = p[1];
		yy = p[2];
	    }
	}
	free(p);
    }
    /* calculate the angle between xy[0] and xxyy */
    return(add_euler(90.0, -RADIANS_TO_DEGREES(atan((yy-y[0])/(xx-x[0])))));
    /* return(add_euler(0.0, RADIANS_TO_DEGREES(atan2(yy-y[0], xx-x[0])))); */
}


/**
   \brief find the triangle on the heightfield that the point coord lies within
   \param map 8bit heightfield array
   \param coord point to find triangle for
   \return a triangle of heightfield points
   \note caller should free the returned triagnle.
*/

triangle_t *find_closest_hf_triangle(uint8_t **map, vertex_t coord)
{
    float xf, xc, zf, zc, xp[3], yp[3];
    triangle_t *tri;

    xf = floor(coord.x);
    zf = floor(coord.z);
    if (xf == coord.x)
	coord.x += 0.6;
    if (zf == coord.z)
	coord.z += 0.6;
    xc = ceil(coord.x);
    zc = ceil(coord.z);
    tri = smalloc(triangle_t);

    xp[0] = xf;
    xp[1] = xf;
    xp[2] = xc;
    yp[0] = zf;
    yp[1] = zc;
    yp[2] = zf;
    if (pnpoly(3, xp, yp, coord.x, coord.z)) {
	tri->side[0].x = xf;
	tri->side[0].z = zf;
	tri->side[1].x = xf;
	tri->side[1].z = zc;
	tri->side[2].x = xc;
	tri->side[2].z = zf;
    } else {
	tri->side[0].x = xf;
	tri->side[0].z = zc;
	tri->side[1].x = xc;
	tri->side[1].z = zf;
	tri->side[2].x = xc;
	tri->side[2].z = zc;
    }
    tri->side[0].y = (float)map[(uint8_t)(rint(fmod(tri->side[0].x/SCALE_XZ,256.0)))]
	[(uint8_t)(rint(fmod(tri->side[0].z/SCALE_XZ,256.0)))]*SCALE_Y;
    tri->side[1].y = (float)map[(uint8_t)(rint(fmod(tri->side[1].x/SCALE_XZ,256.0)))]
	[(uint8_t)(rint(fmod(tri->side[1].z/SCALE_XZ,256.0)))]*SCALE_Y;
    tri->side[2].y = (float)map[(uint8_t)(rint(fmod(tri->side[2].x/SCALE_XZ,256.0)))]
	[(uint8_t)(rint(fmod(tri->side[2].z/SCALE_XZ,256.0)))]*SCALE_Y;
    return(tri);
}

/**
   \breif Given a src and dst coord, return the euler angle between them
   (from the origin)
   \param src source coordinates
   \param dst destination coordinates
   \param angle returned angle between the two
*/

void angle_between(vertex_t src, vertex_t dst, vertex_t *angle)
{
    float a, o;

    dst.x -= src.x;
    dst.y -= src.y;
    dst.z -= src.z;

    memset(angle, 0, sizeof(vertex_t));

    if (rint(dst.x * 100.0) == 0.0 && rint(dst.z * 10.0) == 0.0 &&
	rint(dst.y * 100.0) == 0.0) {
	angle->x = 0.0;
	angle->y = 0.0;
	angle->z = 0.0;
	return;
    }

    if (floor(fabs(dst.x) * 10.0) == 0.0 && floor(fabs(dst.y) * 10.0) == 0.0)
	angle->z = 0.0;
    else {
	a = fabs(dst.x);
	o = dst.y;
	angle->z = add_euler(0.0, rint(RADIANS_TO_DEGREES(atan(o/a))));
    }
    if (floor(fabs(dst.x) * 10.0) == 0.0 && floor(fabs(dst.z) * 10.0) == 0.0)
	angle->y = 0.0;
    else {
	a = dst.x;
	o = dst.z;
	angle->y = add_euler(0.0, RADIANS_TO_DEGREES(atan(o/a)));
    }
    if (floor(fabs(dst.z) * 10.0) == 0.0 && floor(fabs(dst.y) * 10.0) == 0.0)
	angle->x = 0.0;
    else {
	a = dst.z;
	o = dst.y;
	angle->x = add_euler(0.0, RADIANS_TO_DEGREES(atan(o/a)));
    }
}


/**
   \brief return the lowest point on the underlying HF triangle
   \param tri a heightfield triangle
   \return the lowest point on the underlying HF triangle
*/

int find_lowest_point_tri(triangle_t *tri)
{
    int lp=0;

    if (tri->side[1].y < tri->side[lp].y)
	lp = 1;
    if (tri->side[2].y < tri->side[lp].y)
	lp = 2;
    return(lp);
}

/**
   \brief return the highest point on the underlying HF triangle
   \param tri a heightfield triangle
   \return the highest point on the underlying HF triangle
*/

int find_highest_point_tri(triangle_t *tri)
{
    int lp=0;

    if (tri->side[1].y > tri->side[lp].y)
	lp = 1;
    if (tri->side[2].y > tri->side[lp].y)
	lp = 2;
    return(lp);
}

/**
   \brief Find the center of a heightfield triangle
   \param tri triangle to find center of
   \param center found center
*/

void find_center_tri(triangle_t *tri, vertex_t *center)
{
    center->x = (tri->side[0].x + tri->side[1].x + tri->side[2].x) / 3.0;
    center->y = (tri->side[0].y + tri->side[1].y + tri->side[2].y) / 3.0;
    center->z = (tri->side[0].z + tri->side[1].z + tri->side[2].z) / 3.0;
}

/**
   \brief create a bounding box for an object, used in gravity simulations
   \param bmin minimum corner
   \param bmax maximum corner
   \param cog center of gravity
   \param coord coordinates of box
   \return a bounding box
*/

bbox_t *create_bbox(vertex_t bmin, vertex_t bmax, vertex_t cog, vertex_t coord)
{
    bbox_t *box;
    vertex_t bx, bm;

    box = smalloc(bbox_t);

    bx = coord;
    bm = coord;

    bx.x += bmax.x;
    bx.y += bmax.y;
    bx.z += bmax.z;
    bm.x += bmin.x;
    bm.y += bmin.y;
    bm.z += bmin.z;

    box->side[0] = bx;
    box->side[7] = bm;

    box->side[1] = bx;
    box->side[1].z = bm.z;
    box->side[2] = bx;
    box->side[2].y = bm.y;
    box->side[3] = bx;
    box->side[3].z = bm.z;
    box->side[3].y = bm.y;

    box->side[4] = bm;
    box->side[4].y = bx.y;
    box->side[4].z = bx.z;
    box->side[5] = bm;
    box->side[5].y = bx.y;
    box->side[6] = bm;
    box->side[6].z = bx.z;

    box->cog = cog;
    box->cog.x += coord.x;
    box->cog.y += coord.y;
    box->cog.z += coord.z;
    box->center.x = (bmax.x+bmin.x)/2.0 + coord.x;
    box->center.y = (bmax.y+bmin.y)/2.0 + coord.y;
    box->center.z = (bmax.z+bmin.z)/2.0 + coord.z;

    return(box);
}

/**
   \brief Another convenience function.  Returns a number between min and max.
   \param min minimum number to return
   \param max maximum number to return
   \return a random number between min and max

  It is suggested one use these functions rather than RANDOM()%, as it
  would appear that a number of off-by-one-errors exist due to improper
  use of %.  This should also prevent SIGFPE.
*/

int rndm(int min, int max)
{
  int diff;

  diff = max - min + 1;
  if (max < 1 || diff < 1)
    return(min);

  return(RANDOM()%diff+min);
}

/**
   \brief print a stat out
   \param stat stat to print
   \return a string with the stat printed
   \warning caller must free returned string
*/

char *print_stat(stat_t stat)
{
    char buf[20];

    if (stat.pct > 0)
	sprintf(buf, "%d/%d", stat.base, stat.pct);
    else
	sprintf(buf, "%d", stat.base);
    return(strdup(buf));
}

/**
   \brief this adds two eulers, preventing 361 degree turns
   \param base base euler
   \param add euler to add to base
   \param new angle
*/

float add_euler(float base, float add)
{
    base += add;
    while (base > 360.0)
	base -= 360.0;
    if (base < 0.0)
	base += 360.0;
    if (base == 360.0)
	base = 0.0;
    return(base);
}

/**
   \brief average between 2 angles
   \param a first angle
   \param b second angle
   \return average of the two
*/

float avg_euler(float a, float b)
{
    float as, ac, bs, bc;

    as = sin(DEGREES_TO_RADIANS(a));
    ac = cos(DEGREES_TO_RADIANS(a));
    bs = sin(DEGREES_TO_RADIANS(b));
    bc = cos(DEGREES_TO_RADIANS(b));

    return(RADIANS_TO_DEGREES(atan2(as+bs, ac+bc)));
}

/**
   \brief smaller side difference between two angles
   \param a first angle
   \param b second angle
   \return difference
*/

float diff_euler(float a, float b)
{
    float diff;

    diff = fabs(a - b);
    if (diff > 180.0)
	diff = fabs(360.0 - diff);
	    /*fabs(fabs(a - 360.0) - b);*/
    return(diff);
}

/**
   \brief Find the coordinates where 2 lines intersect
   \param l1a coordinates of line 1 start
   \param l1b coordinates of line 1 end
   \param l2a coordinates of line 2 start
   \param l2b coordinates of line 2 end
   \param intersect coordinates of intersection (only use X and Y, Z is 0.0)
*/

void intersection_line(vertex_t l1a, vertex_t l1b, vertex_t l2a, vertex_t l2b,
		       vertex_t *intersect)
{
    float ua, ub;

    ua = ((l2b.x - l2a.x)*(l1a.y - l2a.y) - (l2b.y - l2a.y)*(l1a.x - l2a.x)) /
	((l2b.y - l2a.y)*(l1b.x - l1a.x) - (l2b.x - l2a.x)*(l1b.y - l1a.y));
    ub = ((l1b.x - l1a.x)*(l1a.y - l2a.y) - (l1b.y - l1a.y)*(l1a.x - l2a.x)) /
	((l2b.y - l2a.y)*(l1b.x - l1a.x) - (l2b.x - l2a.x)*(l1b.y - l1a.y));
    intersect->x = l1a.x + ua *(l1b.x - l1a.x);
    intersect->y = l1a.y + ub *(l1b.y - l1a.y);
    intersect->z = 0.0;
}

/**
   \brief distance between two points squared
   \param ca first point
   \param cb second point
   \return distance squared
*/

float dist_between_squared(vertex_t ca, vertex_t cb)
{
    float xd, yd, zd;

    xd = cb.x - ca.x;
    yd = cb.y - ca.y;
    zd = cb.z - ca.z;
    return(xd*xd + yd*yd + zd*zd);
}

/**
   \brief distance between two points
   \param ca first point
   \param cb second point
   \return distance
*/

float dist_between(vertex_t ca, vertex_t cb)
{
    float xd, yd, zd;

    xd = cb.x - ca.x;
    yd = cb.y - ca.y;
    zd = cb.z - ca.z;
    return(sqrt(xd*xd + yd*yd + zd*zd));
}

/**
   \brief distance between two points in a flat plane
   \param ca first point
   \param cb second point
   \return distance
*/

float dist_between_flat(vertex_t ca, vertex_t cb)
{
    float xd, zd;

    xd = cb.x - ca.x;
    zd = cb.z - ca.z;
    return(sqrt(xd*xd + zd*zd));
}

/**
   \brief get the 1d vector from two differing vectors
   \param aa angle A
   \param ma magnitude A
   \param ab angle B
   \param mb magnitude B
   \return the resultant 1d-vector
*/

vector1_t *vector_1d_cross(float aa, float ma, float ab, float mb)
{
    vector1_t *v;
    float as, ac, bs, bc, vma, vmb, lvm, ca, cb;

    v = smalloc(vector1_t);
    
    as = sin(DEGREES_TO_RADIANS(aa));
    ac = cos(DEGREES_TO_RADIANS(aa));
    bs = sin(DEGREES_TO_RADIANS(ab));
    bc = cos(DEGREES_TO_RADIANS(ab));

    v->a = RADIANS_TO_DEGREES(atan2((as*ma)+(bs*mb), (ac*ma)+(bc*mb)));

    if (ma == 0.0) {
	v->m = mb;
    } else {
	lvm = ma;
	if (fabs(mb) > fabs(lvm))
	    lvm = mb;

	vma = ma/lvm;
	vmb = mb/lvm;

	ca = as*vma+bs*vmb;
	cb = ac*vma+bc*vmb;

	if (ca < 0.0) {
	    v->m = (cb*cb) - (ca*ca);
	} else if (cb < 0.0) {
	    v->m = (ca*ca) - (cb*cb);
	} else {
	    v->m = (ca*ca) + (cb*cb);
	}
	v->m *= lvm;
#if 0


	as = sin(DEGREES_TO_RADIANS(aa));
	ac = cos(DEGREES_TO_RADIANS(aa));
	bs = sin(DEGREES_TO_RADIANS(ab));
	bc = cos(DEGREES_TO_RADIANS(ab));
	cs = sin(DEGREES_TO_RADIANS(v->a));
	cc = cos(DEGREES_TO_RADIANS(v->a));
    
	l1a.x = ac*ma;
	l1a.y = as*ma;
	l1a.z = 0.0;
	l1b.x = bc*mb;
	l1b.y = bs*mb;
	l1b.z = 0.0;

	l2a.x = 0.0;
	l2a.y = 0.0;
	l2a.z = 0.0;
	l2b.x = cc*999999.0;
	l2b.y = cs*999999.0;
	l2b.z = 0.0;

	intersection_line(l1a, l1b, l2a, l2b, &inter);
	v->m = dist_between(l2a, inter);
	if (diff_euler(aa, ab) < 45.1)
	    v->m += (ma+mb)/2.0;
	else if (diff_euler(aa, v->a) > 90.0)
	    v->m = -v->m;
#endif
    }
    /* this breaks at 45^ angles 
       v->m = (as+ac)*ma+(bs+bc)*mb;*/

    return(v);
}

/**
   \brief is a the same as b within a factor of factor?
   \param a first float
   \param b second float
   \param factor factor
   \return true/false
*/

int same_within(float a, float b, float factor)
{
    if (floor(a*factor) == floor(b*factor))
	return(1);
    return(0);
}
