/* $Id: xmlparse.c,v 1.11 2004/03/31 03:45:46 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/param.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include "defines.h"
#include "shstr.h"
#include "utils.h"
#include "xmlparse.h"

#include <assert.h>

/**
   \file xmlparse.c
   \brief XML parsing and writing routines
   \author Tim Rightnour
*/

/** \brief parse a string from the XML stuff */

void parsestring(xmlDocPtr doc, xmlNodePtr cur, char **string)
{
    char *tmp;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    *string = add_string(tmp);
    free(tmp);
}


/** \brief parse a float from the XML stuff */

void parsefloat(xmlDocPtr doc, xmlNodePtr cur, float *data)
{
    char *tmp;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    *data = atof(tmp);
    free(tmp);
}

/** \brief parse a double from the XML stuff */

void parsedouble(xmlDocPtr doc, xmlNodePtr cur, double *data)
{
    char *tmp;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    *data = atof(tmp);
    free(tmp);
}

/** \brief parse a long from the XML stuff */

void parselong(xmlDocPtr doc, xmlNodePtr cur, long *data)
{
    char *tmp;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    *data = atol(tmp);
    free(tmp);
}

/** \brief parse an int from the XML stuff */

void parseint(xmlDocPtr doc, xmlNodePtr cur, int *data)
{
    char *tmp;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    *data = atoi(tmp);
    free(tmp);
}

/** \brief parse three ints from the XML stuff */

void parsetriint(xmlDocPtr doc, xmlNodePtr cur, int *x, int *y, int *z)
{
    char *tmp;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    assert(sscanf(tmp, "%d %d %d", x, y, z)==3);
    free(tmp);
}

/** \brief parse four ints from the XML stuff */

void parsequadint(xmlDocPtr doc, xmlNodePtr cur,
		  int *w, int *x, int *y, int *z)
{
    char *tmp;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    assert(sscanf(tmp, "%d %d %d %d", w, x, y, z) == 4);
    free(tmp);
}

/** \brief parse three floats from the XML stuff */

void parsetrifloat(xmlDocPtr doc, xmlNodePtr cur, float *x, float *y, float *z)
{
    char *tmp;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    assert(sscanf(tmp, "%f %f %f", x, y, z) == 3);
    free(tmp);
}

/** \brief parse three doubles from the XML stuff */

void parsetridouble(xmlDocPtr doc, xmlNodePtr cur, double *x, double *y,
		    double *z)
{
    char *tmp;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    assert(sscanf(tmp, "%lf %lf %lf", x, y, z)==3);
    free(tmp);
}


/** \brief parse a vertex_t from the XML stuff */

void parsevector(xmlDocPtr doc, xmlNodePtr cur, vertex_t *vector)
{
    xmlNodePtr child;
    char *tmp;

    child = cur->xmlChildrenNode;
    while (child != NULL) {
	tmp = xmlNodeListGetString(doc, child->xmlChildrenNode, 1);
	if (!strcmp(child->name, "x"))
	    vector->x = atof(tmp);
	if (!strcmp(child->name, "y"))
	    vector->y = atof(tmp);
	if (!strcmp(child->name, "z"))
	    vector->z = atof(tmp);
	free(tmp);
	child = child->next;
    }
}

/** \brief parse a stat_t from the XML stuff */

void parsestat(xmlDocPtr doc, xmlNodePtr cur, stat_t *stat)
{
    char *tmp, *foo;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    foo = tmp;
    stat->base = atoi(strsep(&foo, "/"));
    if (foo != NULL)
	stat->pct = atoi(foo);
    free(tmp);
}

/** \brief parse an uid */
void parseuid(xmlDocPtr doc, xmlNodePtr cur, uniqid_t *data) {
    char *tmp;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    assert(uid_sread(tmp,data));
    free(tmp);
};


/* XML WRITERS */

/** \brief write a double */
void write_double(xmlNodePtr node, char *elm, double d)
{
    char buf[256];

    sprintf(buf, "%f", d);
    xmlNewChild(node, NULL, elm, buf);
}

/** \brief write a float */
void write_float(xmlNodePtr node, char *elm, float f)
{
    char buf[256];

    sprintf(buf, "%f", f);
    xmlNewChild(node, NULL, elm, buf);
}

/** \brief write a long */
void write_long(xmlNodePtr node, char *elm, long l)
{
    char buf[256];

    sprintf(buf, "%ld", l);
    xmlNewChild(node, NULL, elm, buf);
}

/** \brief write an int */
void write_int(xmlNodePtr node, char *elm, int i)
{
    char buf[256];

    sprintf(buf, "%d", i);
    xmlNewChild(node, NULL, elm, buf);
}

/** \brief write three ints */
void write_triint(xmlNodePtr node, char *elm, int i, int j, int k)
{
    char buf[256];

    sprintf(buf, "%d %d %d", i, j, k);
    xmlNewChild(node, NULL, elm, buf);
}

/** \brief write four ints */
void write_quadint(xmlNodePtr node, char *elm, int i, int j, int k, int l)
{
    char buf[256];

    sprintf(buf, "%d %d %d %d", i, j, k, l);
    xmlNewChild(node, NULL, elm, buf);
}

/** \brief write a string */
void write_string(xmlNodePtr node, char *elm, char *string)
{
    xmlNewChild(node, NULL, elm, string);
}

/** \brief write three floats */
void write_trifloat(xmlNodePtr node, char *elm, float x, float y, float z)
{
    char buf[256];

    sprintf(buf, "%f %f %f", x, y, z);
    xmlNewChild(node, NULL, elm, buf);
}

/** \brief write a uniq id */
void write_uid(xmlNodePtr node, char *elm, uniqid_t *data) {
    char buf[38];

    uid_sprint(buf,data);
    xmlNewChild(node, NULL, elm, buf);
}
