/* $Id: cmd_obj.c,v 1.27 2004/03/30 06:45:55 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#if defined(__linux__)
#include <values.h>
#include "queue.h"
#else
#include <sys/queue.h>
#endif
#include <unistd.h>
#include <math.h>
#include <sys/param.h>

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "object.h"
#include "map.h"
#include "shstr.h"
#include "utils.h"
#include "models.h"
#include "collision.h"
#include "tod.h"
#include "xmlparse.h"
#include "matrix.h"
#include "event.h"
#include "player.h"
#include "network.h"
#include "netparser.h"
#include "ninterface.h"
#include "text.h"
#include "commands.h"

/** \file cmd_obj.c
    \brief Object related commands
    \author Tim Rightnour

    Commands that have to do with objects.  Picking them up, wearing them,
    dropping them, etc.
*/

/* ----- Support Functions ----- */

/**
   \brief Find the closest object that can be picked up and return it.
   \param obj the object that wants to do the picking.
   \return pointer to the object to be picked.
*/

object_t *find_closest_pickable_obj(object_t *obj)
{
    maphead_t *maph;
    object_t *target, *tmp;
    float dist, d;

    dist = DBL_MAX;
    target = NULL;
    maph = obj->maphead;
    TAILQ_FOREACH(tmp, &maph->map->objs, next) {
	if (!QUERY_FLAG(tmp, FLAG_ALIVE) && !QUERY_FLAG(tmp, FLAG_NO_PICK)) {
	    d = dist_between_squared(obj->cli.coord, tmp->cli.coord);
	    if (d < dist) {
		dist = d;
		target = tmp;
	    }
	}
    }
    return(target);
}


/* ----- Do Commands ----- */


/**
   \brief pick up an object.
   \param obj object that is doing the picking
   \param target object to be picked up
   \return -2 too heavy -1 too far 0 unpickable 1 success
   \todo weight checking, encumbrance checking
   \todo make vertical reach depend on obj->stride
*/

int do_get(object_t *obj, object_t *target)
{
    vertex_t tloc;
    event_t *ev;

    if (target == obj || target == NULL)
	return(0);

    if (QUERY_FLAG(target, FLAG_NO_PICK))
	return(0);

    tloc = target->cli.coord;
    /* XXX we use 10.0 as a vertical reach factor */
    if (fabs(tloc.y - obj->cli.coord.y) < 10.0)
	tloc.y = obj->cli.coord.y;
    else
	return(-1);

    if (dist_between(obj->cli.coord, tloc) > 10.0)
	return(-1);

    remove_obj_from_map(target, target->maphead, POOM_SPACE, 0);
    put_obj_in_obj(target, obj, BUCKET_INV);
    if (obj->pl != NULL)
	tell_inv_obj(obj->pl, target, obj, BUCKET_INV);
    ev = new_obj_event(Q_PLAYER, 0, "obj_removed_from_map", target);
    ev->action = 1;
    queue_event(ev, target->mapgroup);

    return(1);
}

/**
   \brief drop an object.
   \param obj object that is doing the picking
   \param target object to be picked up
   \return 0 cursed, -1 you don't have one, 1 success
*/

int do_drop(object_t *obj, object_t *target)
{
    event_t *ev;
    
    if (target == obj || target == NULL)
	return(0);

    /* can't drop cursed objects */
    if (QUERY_FLAG(target, FLAG_CURSED))
	return(0);

    /* is the object in INVENTORY? */
    if (!verify_obj_in_bucket(target, obj, BUCKET_INV))
	return(-1);

    target->cli.coord = obj->cli.coord;
    put_obj_on_map_closeto(target, obj->maphead, POOM_SPACE);
    remove_obj_from_obj(target, obj, BUCKET_INV);
    if (obj->pl != NULL)
	tell_rminv_obj(obj->pl, target, obj, BUCKET_INV);
    ev = new_obj_event(Q_PLAYER, 0, "obj_teach_players_self", target);
    ev->action = 1;
    queue_event(ev, target->mapgroup);
    return(1);
}

/**
   \brief equip an object.
   \param obj the object doing the wearing
   \param target the object to be worn
   \param location the location to attempt to wear the object
   \return 0 you don't have one, -1 allready wearing one, -2 cannot wear,
   1 success
*/

int do_wear(object_t *obj, object_t *target, int location)
{
    event_t *ev;

    if (target == NULL || target == obj ||
	!verify_obj_in_bucket(target, obj, BUCKET_INV) ||
	!verify_obj_in_obj(obj, target))
	return(0);

    if (QUERY_EFLAG(obj->cli.equip, location))
	return(-1);

    if ((target->cli.stats.type == T_WEAPON && 
	!(location == EQ_WRHAND || location == EQ_WLHAND)) ||
	!QUERY_EFLAG(target->cli.equip, location))
	return(-2);

    target->cli.equip.loc = location;
    remove_obj_from_obj(target, obj, BUCKET_INV);
    put_obj_in_obj(target, obj, BUCKET_EQUIP);
    if (obj->pl != NULL) {
	tell_rminv_obj(obj->pl, target, obj, BUCKET_INV);
	tell_inv_obj(obj->pl, target, obj, BUCKET_EQUIP);
	tell_equip_obj(obj->pl, obj, target, location,
		       target->cli.equip.visible);
    }
    ev = new_obj_event(Q_PLAYER, 0, "obj_wore_obj", obj);
    ev->action = 1;
    evarg_setobj(ev,0,target);
    queue_event(ev, obj->mapgroup);
    
    return(1);
}

/**
   \brief unequip an object.  returns success/fail
   \param obj the object doing the unwearing
   \param target the object to be unworn
   \return -1 cursed, 0 object not being worn, 1 success
*/

int do_unwear(object_t *obj, object_t *target)
{
    event_t *ev;

    if (!verify_obj_in_obj(obj, target) || target == obj ||
	!verify_obj_in_bucket(target, obj, BUCKET_EQUIP))
	return(0);

    if (QUERY_FLAG(target, FLAG_CURSED))
	return(-1);

    target->cli.equip.loc = EQ_UNEQ;
    remove_obj_from_obj(target, obj, BUCKET_EQUIP);
    put_obj_in_obj(target, obj, BUCKET_INV);
    if (obj->pl != NULL) {
	tell_rminv_obj(obj->pl, target, obj, BUCKET_EQUIP);
	tell_inv_obj(obj->pl, target, obj, BUCKET_INV);
	tell_unequip_obj(obj->pl, obj, target,
			 target->cli.equip.visible);
    }
    ev = new_obj_event(Q_PLAYER, 0, "obj_unwore_obj", obj);
    ev->action = 1;
    evarg_setobj(ev,0,target);
    queue_event(ev, obj->mapgroup);
    
    return(1);
}

/* ----- Parsing Functions ----- */

/**
   \brief the command that attempts to drop an object
   \param obj object that is doing the dropping
   \param args arguments passed from the client to the drop command
   \return int success/fail
*/

int command_drop(object_t *obj, pargs_t *args)
{
    object_t *target;
    int i;
    event_t *event;

    i = find_arg_by_id(args, TARGET);
    if (i == -1)
        return(0);
    target = get_obj_from_uid(&args[i].arg.uid);

    if (target == NULL)
	return(0);

    event = new_obj_event(Q_PLAYER, ETYPE_NONE, "obj_drop_obj", obj);
    evarg_setobj(event,0,target);    
    queue_obj_event(event);

    printf("obj=%s is trying to DROP %s\n", obj->cli.name, target->cli.name);
    return(1);
}

/**
   \brief the command that attempts to pick up an object
   \param obj object that is doing the picking
   \param args arguments passed from the client to the get command
   \return int success/fail
*/

int command_get(object_t *obj, pargs_t *args)
{
    object_t *target;
    event_t *event;
    int i;

    i = find_arg_by_id(args, TARGET);
    if (i == -1)
        target = NULL;
    else
	target = get_obj_from_uid(&args[i].arg.uid);

    event = new_obj_event(Q_PLAYER, ETYPE_NONE, "obj_get_obj", obj);
    evarg_setobj(event,0,target);
    queue_obj_event(event);

    if (target)
	printf("obj=%s is trying to GET %s\n", obj->cli.name,
	       target->cli.name);
    return(1);
}

/**
   \brief parse the command to unwear an object
   \param obj the object to do the unwearing
   \param args the arguments from the client
   \return int success/fail
*/

int command_unwear(object_t *obj, pargs_t *args)
{
    object_t *target;
    int i;
    event_t *event;

    i = find_arg_by_id(args, TARGET);
    if (i == -1)
        return(0);
    target = get_obj_from_uid(&args[i].arg.uid);

    if (target == NULL)
	return(0);

    event = new_obj_event(Q_PLAYER, ETYPE_NONE, "obj_unwear_obj", obj);
    evarg_setobj(event,0,target);
    
    queue_obj_event(event);

    return(1);
}

/**
   \brief parse the command to wield a weapon
   \param obj the object to do the wielding
   \param args the arguments from the client
   \return int success/fail
*/

int command_wield(object_t *obj, pargs_t *args)
{
    object_t *target;
    int i;
    event_t *event;

    i = find_arg_by_id(args, TARGET);
    if (i == -1)
        return(0);
        target = get_obj_from_uid(&args[i].arg.uid);

    if (target == NULL)
	return(0);

    event = new_obj_event(Q_PLAYER, ETYPE_NONE, "obj_wear_obj", obj);
    evarg_setobj(event,0,target);
    /* for now, all players will just be right handed. */
    event->action = EQ_WRHAND;
    queue_obj_event(event);

    return(1);
}

/* ----- Events ----- */

/**
   \brief An object wants to pick up another object
   \param event event that called us
   \return event status, allways 1
*/

int obj_get_obj(event_t *event)
{
    object_t *obj, *target;
    char buf[256];

    obj = event->initiator;
    target = evarg_getobj(event,0);
    if (target == NULL)
	target = find_closest_pickable_obj(obj);

    if (target == NULL || target->mapgroup != event_mapg(event))
        return(1);

    if (obj->pl == NULL) {
	(void)do_get(obj, target);
	return(1);
    }

    switch (do_get(obj, target)) {
    case 1: 
	sprintf(buf, "You pick up the %s.", pluralize_name(target));
	break;
    case -1:
	sprintf(buf, "You cannot reach the %s.", pluralize_name(target));
	break;
    case -2:
	sprintf(buf, "You cannot lift the %s.", pluralize_name(target));
	break;
    case 0:
	sprintf(buf, "You cannot pick up the %s.", pluralize_name(target));
	break;
    }
    tell_text(obj->pl, buf, DATA_INFO, NULL);
    return(1);
}

/**
   \brief An object wants to drop another object
   \param event event that called us
   \return event status, allways 1
*/

int obj_drop_obj(event_t *event)
{
    object_t *obj, *target;
    char buf[256];

    obj = event->initiator;
    target = evarg_getobj(event,0);
    if (target->mapgroup != event_mapg(event))
        return(1);

    if (obj->pl == NULL) {
	(void)do_drop(obj, target);
	return(1);
    }
    switch (do_drop(obj, target)) {
    case 1:
	sprintf(buf, "You drop your %s.\n", pluralize_name(target));
	break;
    case -1:
	sprintf(buf, "You don't have any %s to drop.",
		pluralize_name(target));
	break;
    case 0:
	sprintf(buf, "For some reason, you cannot drop %s.",
		pluralize_name(target));
	break;
    }
    tell_text(obj->pl, buf, DATA_INFO, NULL);
    return(1);
}

/**
   \brief An object wants to wear another object
   \param event event that called us
   \return event status, allways 1
*/

int obj_wear_obj(event_t *event)
{
    object_t *obj, *target;
    char buf[256];

    obj = event->initiator;
    target = evarg_getobj(event,0);
    if (target->mapgroup != event_mapg(event))
        return(1);

    if (obj->pl == NULL) {
	(void)do_wear(obj, target, event->action);
	return(1);
    }
    switch (do_wear(obj, target, event->action)) {
    case 0:
	sprintf(buf, "You do not have any %s to wear.",
		pluralize_name(target));
	break;
    case -1:
	sprintf(buf, "You are allready wearing something there.");
	break;
    case -2:
	sprintf(buf, "You cannot wear the %s.", pluralize_name(target));
	break;
    case 1:
	sprintf(buf, "You wear the %s.", pluralize_name(target));
	break;
    }
    tell_text(obj->pl, buf, DATA_INFO, NULL);
    return(1);
}

/**
   \brief An object wants to unwear another object
   \param event event that called us
   \return event status, allways 1
*/

int obj_unwear_obj(event_t *event)
{
    object_t *obj, *target;
    char buf[256];

    obj = event->initiator;
    target = evarg_getobj(event,0);
    if (target->mapgroup != event_mapg(event))
        return(1);

    if (obj->pl == NULL) {
	(void)do_unwear(obj, target);
	return(1);
    }
    switch (do_unwear(obj, target)) {
    case -1:
	sprintf(buf, "The %s seems to be stuck to your body!",
		pluralize_name(target));
	break;
    case 0:
	sprintf(buf, "You are not wearing a %s.", pluralize_name(target));
	break;
    case 1:
	sprintf(buf, "You remove the %s.", pluralize_name(target));
	break;
    }
    tell_text(obj->pl, buf, DATA_INFO, NULL);
    return(1);
}
