/* $Id: collision.c,v 1.31 2004/04/02 02:39:00 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <math.h>
#include <time.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "collision.h"
#include "utils.h"
#include "netparser.h"
#include "ninterface.h"
#include "servdata.h"
#include "map.h"

extern npersist_data_t nonpersist;

/**
   \file collision.c
   \brief Collision callbacks
   \author Tim Rightnour

   Collision callbacks and things having directly to do with collision
*/

/**
   \brief Determine if two objects collide
   \param client_data the mapgroup to do a test upon
   \param obj1 The first object that has collided
   \param obj2 The second object that has collided
   \param coll_data Data detailing the nature of the collision
   \warning This is called by dtTest()

   If the two objects collide, we increment the collision count of both
   objects.  This is the standard behavior of the collision engine.
*/

/*  ARGSUSED */
void collide2(void *client_data, DtObjectRef obj1, DtObjectRef obj2,
	      const DtCollData *coll_data)
{
    mapgroup_t *mapg;

    mapg = (mapgroup_t *)client_data;

    if ((*(object_t *)obj1).mapgroup == mapg &&
	(*(object_t *)obj2).mapgroup == mapg) {

	(*(object_t *)obj1).collision++;
	(*(object_t *)obj2).collision++;
	LOG(LOG_DEBUG, "collide2:Object %s interferes with object %s",
	    (*(object_t *)obj1).cli.name, (*(object_t *)obj2).cli.name);
	printf("collide2:Object %s interferes with object %s\n",
	       (*(object_t *)obj1).cli.name, (*(object_t *)obj2).cli.name);
    }
}

/**
   \brief Determine how many collisions a specific object has currently
   \param client_data The object to check against
   \param obj1 The first object that has collided
   \param obj2 The second object that has collided
   \param coll_data Data detailing the nature of the collision
   \warning This is called by dtTest()

   If the client_data is set to an object, we only increment the collision
   count on that one object, ignoring all other collisions.  It also looks
   for an object named "collider" which is used by the create_collision_sphere
   call.  Because of this, only one collider object can be in use at any
   time.
*/

/*  ARGSUSED */
void collide_count(void *client_data, DtObjectRef obj1, DtObjectRef obj2,
	      const DtCollData *coll_data)
{
    /* only mark collision objects */
    if (strcmp((*(object_t *)obj1).cli.name, "collider") == 0 ||
	(object_t *)obj1 == (object_t *)client_data)
	(*(object_t *)obj1).collision++;
    if (strcmp((*(object_t *)obj2).cli.name, "collider") == 0 ||
	(object_t *)obj2 == (object_t *)client_data)
	(*(object_t *)obj2).collision++;

    LOG(LOG_DEBUG, "collide_count:Object %s interferes with object %s",
	    (*(object_t *)obj1).cli.name, (*(object_t *)obj2).cli.name);
    /*printf("collide_count:Object %s interferes with object %s\n",
      (*(object_t *)obj1).cli.name, (*(object_t *)obj2).cli.name);*/
}

/**
   \brief Determine what a specific object collides with
   \param client_data Client data passed to the collide call
   \param obj1 The first object that has collided
   \param obj2 The second object that has collided
   \param coll_data Data detailing the nature of the collision
   \warning This is called by dtTest()

   This function is desgined for use with functions like
   create_collision_sphere().  It finds all objects that collide with the
   collider object, and marks them with a collision count.  All other
   collisions are ignored.

   \todo This function should probably be extended to take a client_data
   argument like collide_count()
*/

/*  ARGSUSED */
void collide_marktarget(void *client_data, DtObjectRef obj1, DtObjectRef obj2,
	      const DtCollData *coll_data)
{
    /* only mark non-collision objects */
    if (strcmp((*(object_t *)obj1).cli.name, "collider") != 0)
	(*(object_t *)obj1).collision++;
    if (strcmp((*(object_t *)obj2).cli.name, "collider") != 0)
	(*(object_t *)obj2).collision++;
    LOG(LOG_DEBUG, "collide_marktarget:Object %s interferes with object %s",
	    (*(object_t *)obj1).cli.name, (*(object_t *)obj2).cli.name);
    printf("collide_mt:Object %s interferes with object %s\n",
	   (*(object_t *)obj1).cli.name, (*(object_t *)obj2).cli.name);
}

/**
   \brief Unmark all colliding objects
   \param client_data the mapgroup to act on
   \param obj1 The first object that has collided
   \param obj2 The second object that has collided
   \param coll_data Data detailing the nature of the collision
   \warning This is called by dtTest()

   This function finds all objects which are experiencing a collision, and
   unmarks the collision flag.

   \todo This function should probably be extended to take a client_data
   argument like collide_count()
*/

/*  ARGSUSED */
void collide_uncollide(void *client_data, DtObjectRef obj1, DtObjectRef obj2,
	      const DtCollData *coll_data)
{
    mapgroup_t *mapg;

    mapg = (mapgroup_t *)client_data;

    if ((*(object_t *)obj1).mapgroup == mapg &&
	(*(object_t *)obj2).mapgroup == mapg) {

	(*(object_t *)obj1).collision = 0;
	(*(object_t *)obj2).collision = 0;
	LOG(LOG_DEBUG, "collide_uncollide:Object %s interferes with object %s",
	    (*(object_t *)obj1).cli.name, (*(object_t *)obj2).cli.name);
	printf("collide_uncollide:Object %s interferes with object %s\n",
	       (*(object_t *)obj1).cli.name, (*(object_t *)obj2).cli.name);
    }
}

/**
   \brief Inform player of objects he collides with
   \param client_data A player_t representing the player to tell.
   \param obj1 The first object that has collided
   \param obj2 The second object that has collided
   \param coll_data Data detailing the nature of the collision
   \warning This is called by dtTest()

   This function locates all objects that a player collides with, and issues
   a tell_new_obj() for each one.  This is generally called by the
   player_find_surrounds() event.

*/

/*  ARGSUSED */
void collide_tell_player(void *client_data, DtObjectRef obj1, DtObjectRef obj2,
	      const DtCollData *coll_data)
{
    object_t *found;
    player_t *pl;

    pl = (player_t *)client_data;
    found = NULL;

    /* only tell about non collision objects (and not me)*/
    if (strcmp((*(object_t *)obj1).cli.name, "collider") != 0 &&
	(object_t *)obj1 != pl->obj)
	found = (object_t *)obj1;
    else if (strcmp((*(object_t *)obj2).cli.name, "collider") != 0 &&
	(object_t *)obj2 != pl->obj)
	found = (object_t *)obj2;

    if (found) {
	if (update_pl_knowledge(pl, found) == 0)
	    tell_new_obj(pl, found, NO_EXTENDED, DO_DRAW);
    }

    LOG(LOG_DEBUG, "collide_tell_player:Object %s interferes with object %s",
	    (*(object_t *)obj1).cli.name, (*(object_t *)obj2).cli.name);
    printf("collide_tp:Object %s interferes with object %s\n",
	   (*(object_t *)obj1).cli.name, (*(object_t *)obj2).cli.name);
}

/**
   \brief create a finder sphere, and tell the player about all objects we find
   \param center the player to center the sphere around
   \param radius the radius in points of the sphere
   \return the object we created (the sphere)
*/

object_t *create_finder_sphere(player_t *center, float radius)
{
    object_t *obj;
    obj = new_obj_from_arch("collider",DO_NOREC); /*XXX */

    if (obj->cli.nrof == 0)
        obj->cli.nrof = 1;
    obj->cli.coord = center->obj->cli.coord;
    put_obj_on_map(obj, center->obj->maphead, POOM_NOSPACE);

    obj->shape = dtSphere(radius); /* XXX perception? */
    dtCreateObject(obj, obj->shape);
    dtSelectObject(obj);
    dtLoadIdentity();
    dtTranslate(obj->cli.coord.x, obj->cli.coord.y, obj->cli.coord.z);
    dtSetObjectResponse(obj, collide_tell_player, DT_SIMPLE_RESPONSE,
			(void *)center);
    return(obj);
}

/**
   \brief Create a collision sphere object and place it on the map.
   \param coord the coordinates to place the sphere at.
   \param radius the radius of the sphere
   \param maph the maphead to place the sphere on
   \param colltype what kind of collision response does this sphere cause
   \return an object_t denoting the sphere we created
*/

object_t *create_collision_sphere(vertex_t coord, float radius,
				  maphead_t *maph, int colltype)
{
    object_t *obj;

    obj = new_obj_from_arch("collider", DO_NOREC);
    if (obj->cli.nrof == 0)
        obj->cli.nrof = 1;
    obj->cli.coord = coord;
    put_obj_on_map(obj, maph, POOM_NOSPACE);
    obj->shape = dtSphere(radius);
    dtCreateObject(obj, obj->shape);
    dtSelectObject(obj);
    dtLoadIdentity();
    dtTranslate(obj->cli.coord.x, obj->cli.coord.y, obj->cli.coord.z);
    switch (colltype) {
    case COLL_NORMAL:
	dtSetObjectResponse(obj, collide2, DT_SIMPLE_RESPONSE, NULL);
	break;
    case COLL_COUNT:
	dtSetObjectResponse(obj, collide_count, DT_SIMPLE_RESPONSE,
			    (void *)obj);
	break;
    case COLL_MARKTARGET:
	dtSetObjectResponse(obj, collide_marktarget, DT_SIMPLE_RESPONSE,
			    NULL);
	break;
    case COLL_UNCOLLIDE:
	dtSetObjectResponse(obj, collide_uncollide, DT_SIMPLE_RESPONSE, NULL);
	break;
    }
    return(obj);
}

/**
   \brief removes a collision sphere from the game
   \param obj the sphere to remove
*/

void remove_collision_obj(object_t *obj)
{
    remove_obj_from_map(obj, obj->maphead, POOM_SPACE, 1);
    remove_obj(obj);
}

/**
   \brief perform a collision test on an object
   \param obj object to test for collsions
   \param coord obj's coordinates
   \return 0 no collision, 1 collision
*/

int perform_collision_test(object_t *obj, vertex_t coord)
{
    float r1, r2, dist, dx, dz;
    object_t *tmp;
    int i;

    r1 = obj->radius;

    dx = (float)(obj->maphead->width) * SCALE_XZ * 0.66;
    dz = (float)(obj->maphead->height) * SCALE_XZ * 0.66;

    for (i=0; i<9; i++) {
	if (obj->maphead->tile[i] == NULL)
	    continue;
	if (!QUERY_FLAG(obj->maphead->tile[i], MAPFLAG_LOADED))
	    continue;
	if (i > 0 && (fabs(coord.x - obj->maphead->tile[i]->x) > dx ||
		      fabs(coord.z - obj->maphead->tile[i]->z) > dz))
	    continue;

	TAILQ_FOREACH(tmp, &obj->maphead->tile[i]->map->objs, next) {
	    if (tmp->cli.model == NULL || QUERY_FLAG(tmp, FLAG_NOCOLLIDE) ||
		tmp == obj)
		continue;
	    r2 = tmp->radius + r1;
	    if (fabs(coord.x - tmp->cli.coord.x) > r2)
		continue;
	    if (fabs(coord.y - tmp->cli.coord.y) > r2)
		continue;
	    if (fabs(coord.z - tmp->cli.coord.z) > r2)
		continue;
	    dist = dist_between_flat(coord, tmp->cli.coord);
	    if (r2 >= dist) {
		collision_map_obj(tmp, tmp->cli.coord);
		dtTestObjects(obj, tmp);
		dtDeleteObject(tmp);
		if (obj->collision)
		    return(1);
	    }
	}
    }
    return(0);
}
