/* $Id: event.c,v 1.49 2004/04/02 02:41:05 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/time.h>
#include <assert.h>

#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "shstr.h"
#include "utils.h"
#include "event.h"
#include "npc.h"
#include "timers.h"
#include "physics.h"
#include "text.h"
#include "settings.h"
#include "servdata.h"
#include "outnet.h"
#include "sutils.h"
#include "object.h"
#include "commands.h"

static pthread_mutex_t queue_mutex = PTHREAD_MUTEX_INITIALIZER;
/* server event queue */
event_queue_t server_queue[Q_MAX];

long max_time = USECS_PER_CLOCK;
struct timeval last_time;
int nrof_efunc_handlers;

extern persist_data_t persist;
extern npersist_data_t nonpersist;
extern TAILQ_HEAD(, _mapgroup_t) maps;
extern pthread_mutex_t maps_mutex;
extern settings_t settings;

void free_event(event_t*); 
/**
   \file event.c
   \brief The event engine
   \author Tim Rightnour

  about events:
  an event returns 0, if it does not consider itself complete.  On a return
  of 1, it will remove itself from the queue.
  an event returning 1 is expected to free necc. data. XXX sure of thats ? FreeEvent do it
*/

/* include the event table */
#include "eventtable.h"



/**
   \brief event flags
   \note these are sorted in init_flags() in object.c
*/
eflag_t evflags[] = {
    {"queued", EVFLAG_QUEUED},
    {"nosave", EVFLAG_NOSAVE},
    {"obj", EVFLAG_OBJ},
    {"follow_obj",EVFLAG_FOLLOW_OBJ},
    {"server_ignore",EVFLAG_SERVER_IGNORE},
    {"del_not_in_mapg",EVFLAG_DEL_NOT_IN_MAPG},
    {"initiator",EVFLAG_INITIATOR},

};

/** \brief size of evflags list */
const size_t evflag_size = sizeof(evflags) / sizeof(eflag_t);

static void evarg_clear(event_t * event,int num) ;

/* move an event to queue : lock must be set , queue can be null ->unqueue*/
static void event_queue_move_nl(event_t *event, event_queue_t *queue) {
    if (event->event_queue) { /*has an event queue */
	if (event == event->event_queue->_cur) { /* is curannt_item */
	    event->event_queue->_cur_action |=2;
	    event->event_queue->_cur_move = queue;
	    return;
	}
	if (event->event_queue == queue)
	    return; /* move to same queue */
	/* remove it from is queue */
	TAILQ_REMOVE(&event->event_queue->queue, event, next);
    }

    if (!queue) {
	CLEAR_FLAG(event, EVFLAG_QUEUED);
	/*event->mapg = NULL;*/
    } else {
	SET_FLAG(event, EVFLAG_QUEUED);
	TAILQ_INSERT_TAIL(&queue->queue, event, next);
	/*event->mapg = queue->mapg;*/
    }
    event->event_queue = queue;
}


static void delete_event_nl(event_t *event) {
    if (event->event_queue) { /*has an event queue */
	if (event == event->event_queue->_cur) { /* is curannt_item */
	    event->event_queue->_cur_action |=1;
	    return;
	}
	/* remove it from is queue */
	TAILQ_REMOVE(&event->event_queue->queue, event, next);
    }
	
    free_event(event);
}

/** \brief delete the event
*/
void delete_event(event_t *event) {
    pthread_mutex_lock(&queue_mutex);
    delete_event_nl(event);
    pthread_mutex_unlock(&queue_mutex);
}






/**
   \brief the empty function handler
   \return int 1
   \param event the event that called us
*/

int empty_func_handler(event_t *event)
{
    LOG(LOG_DEBUG, "Empty func handler called.");
    return(1);
}

/**
   \brief finds the function to call for a given event
   \param event the event to find a function for
   \return the event function record number for a given event. 0 if not found.
   \todo Eventually to be replaced by binary search.
*/

int find_efunc_by_event(event_t *event)
{
    int i;

    for (i=0; event_func_map[i].name != NULL; i++)
	if (event_func_map[i].name == event->func)
	    return(i);
    return(0);
}

/**
   \brief initialize the function handlers

   This changes all the strings in the funcrec to a shared string, and
   counts the number of funcrecs while its there.  This wastes ram but
   makes a few things easier.
*/

void init_efuncrec(void)
{
    int i;
    char *tmp;

    for (i=0; event_func_map[i].name != NULL; i++) {
	tmp = event_func_map[i].name;
	event_func_map[i].name = add_string(tmp);
    }
    nrof_efunc_handlers = i;
}

/**
  \brief initialize the timers for the server
*/

void init_time(void)
{
    nonpersist.pticks = 0;
    (void)gettimeofday(&last_time, NULL);
}

/**
   \brief initialize all the server event queues
*/

void init_server_queues(void)
{
    int i;

    for (i=0; i < Q_MAX; i++) {
	event_queue_init(&server_queue[i], NULL);
    }
}

/**
   \brief initialize the event queues for a mapgroup
   \param mapg the mapgroup to initialize queues for
*/

void init_queues(mapgroup_t *mapg)
{
    int i;

    for (i=0; i < Q_MAX; i++) {
	event_queue_init(&mapg->queue[i], mapg);
    }
}

/**
   \brief get the time of day
   \param tod a timeofday_t to be filled out by this call
   \todo should probably be in sutils.c, not event.c
*/

void get_tod(timeofday_t *tod)
{
    tod->year = persist.todtick/HOURS_PER_YEAR;
    tod->month = (persist.todtick/HOURS_PER_MONTH)%MONTHS_PER_YEAR;
    tod->day = (persist.todtick%HOURS_PER_MONTH)/DAYS_PER_MONTH;
    tod->dayofweek = tod->day%DAYS_PER_WEEK;
    tod->hour = persist.todtick%HOURS_PER_DAY;
    tod->minute = (nonpersist.pticks%PTICKS_PER_CLOCK)/(PTICKS_PER_CLOCK/58);
    if (tod->minute > 58)
	tod->minute = 58; /* it's imprecise at best anyhow */
    tod->weekofmonth = tod->day/WEEKS_PER_MONTH;
    if (tod->month < 3)
	tod->season = 0;
    else if (tod->month < 6)
	tod->season = 1;
    else if (tod->month < 9)
	tod->season = 2;
    else if (tod->month < 12)
	tod->season = 3;
    else
	tod->season = 4;
}

/**
   \brief initialize a new basic event
   \param queue what queue do we place this in (priority)
   \param type what type of event is this?
   \param func a string telling us the function name to call to process
   \return a new unqueued event
*/

event_t *new_event(int queue, int type, char *func)
{
    event_t *event;
    int n;

    event = smalloc(event_t);
    event->queue = queue;
    event->type = type;
    event->func = add_string(func);
    n = find_efunc_by_event(event);
    if (n<= 0) {
	LOG(LOG_ERROR, "Unknown event name '%s'",event->func);
    };
    memcpy(&event->flags,&event_func_map[n].flags,sizeof(event->flags));
    return(event);
}

/**
   \brief release memory used py event and decrease ref
   \param event event to free
   \warning don't remove it from any queue
*/
void free_event(event_t *event){
    int k;
    free_string(event->func);
    if (QUERY_FLAG(event, EVFLAG_OBJ) ) {
	TAILQ_REMOVE(&event->initiator->events, event, stack);
	obj_decref(event->initiator);
    }
    if (event->initiator) obj_decref(event->initiator);
    for ( k=0 ; k< EVARG_NBR ; k++ ) evarg_clear(event,k);
    free(event);
}




/**
   \brief initialize a new object event
   \param queue what queue do we place this in (priority)
   \param type what type of event is this?
   \param func a string telling us the function name to call to process
   \param obj initiator of this event
   \return a new unqueued event
*/

event_t *new_obj_event(int queue, int type, char *func, object_t *obj)
{
    event_t *event;

    event = new_event(queue, type, func);
    event->initiator = obj;
    obj_incref(obj);
    return(event);
}

/**
   \brief place event in the queue
   \param event event to place in the queue
   \param mapg mapgroup to place this event in, NULL for system queue
*/

void queue_event(event_t *event, mapgroup_t *mapg){
    pthread_mutex_lock(&queue_mutex);
    event_queue_move_nl(event, get_event_queue(mapg, event->queue));
    pthread_mutex_unlock(&queue_mutex);
}


static void add_event_obj_nl(event_t*event) {
    if(QUERY_FLAG(event,EVFLAG_OBJ)) {
	LOG(LOG_ERROR, "Duplicate calling add_event_obj");
	return;
    }
    if(!event->initiator) {
	LOG(LOG_ERROR, "Calling add_event_obj with initiator == NULL");
	return;
    }
    obj_incref(event->initiator);
    if (TAILQ_EMPTY(&event->initiator->events))
        TAILQ_INIT(&event->initiator->events);
    TAILQ_INSERT_TAIL(&event->initiator->events, event, stack);
    SET_FLAG(event, EVFLAG_OBJ);
}
/** \brief add event to obj->events
*/
void add_event_obj(event_t*event){
    pthread_mutex_lock(&queue_mutex);
    add_event_obj_nl(event);
    pthread_mutex_unlock(&queue_mutex);
}

/**
   \brief place an event in the queue, owned by obj.
   \note Only used by recurring events.
   \param event event to place
   \param obj object to place event into and base queue on
*/
void queue_obj_event(event_t *event)
{
    pthread_mutex_lock(&queue_mutex);
    add_event_obj_nl(event);
    event_queue_move_nl(event, get_event_queue(event->initiator->mapgroup, event->queue));
    pthread_mutex_unlock(&queue_mutex);
}

/**
  \brief remove an event from it's queue
  \param event event to remove
*/

void unqueue_event(event_t *event) 
{
    pthread_mutex_lock(&queue_mutex);
    event_queue_move_nl(event, NULL);
    pthread_mutex_unlock(&queue_mutex);
}

/** initialize events that must be reocurring, and seeded at startup */

void init_server_events(void)
{
    event_t *event;

    if (settings.mservfreq) {
	event = new_event(Q_MAINT, ETYPE_DNOB, "metaserver_update");
	event->dnob = persist.todtick + 6;
	queue_event(event, NULL);
    }
    event = new_event(Q_MAINT, ETYPE_DNOB, "save_persistent_data");
    event->dnob = persist.todtick + 5;
    queue_event(event, NULL);
}

/**
   \brief initialize events that are recurring and seeded at startup
   \param mapg mapgroup to seed events for
*/

void init_events(mapgroup_t *mapg)
{
    event_t *event;

    /* queue update_all_maps_object_motion */
    event = new_event(Q_TIMER, ETYPE_TIMER, "update_all_maps_object_motion");
    event->action = EACT_MOVEOBJS;
    event->dnob = PTICKS_PER_SPEED;
    queue_event(event, mapg);

    event = new_event(Q_TIMER, ETYPE_TIMER, "npc_life_handler");
    event->action = EACT_NPCLIFE;
    event->dnob = 500;
    queue_event(event, mapg);

    event = new_event(Q_PLAYER, ETYPE_TIMER, "player_update_all_motion");
    event->dnob = PTICKS_PER_SPEED;
    queue_event(event, mapg);
}

/**
   \brief write out persistent data to disk occasionally
   \param event event that called us
   \return event status
*/

int save_persistent_data(event_t *event)
{
    write_persist();
    event->dnob = persist.todtick + 50;
    return(0);
}

/**
   \brief update the metaserver with our server data
   \param event event that called us
   \return event status
*/

int metaserver_update(event_t *event)
{
    char *buf, *buf2;
    servout_t *out;
    size_t len=0;
    event_t *ogn;

    out = new_outgoing_connection(settings.mservhost, settings.mservport);
    if (out == NULL) {
	LOG(LOG_WARNING, "Could not connect to metaserver %s:%d",
	    settings.mservhost, settings.mservport);
	return(0);
    }
    len += ((settings.name) ? strlen(settings.name) : 0);
    len += ((settings.desc) ? strlen(settings.desc) : 0);
    len += ((settings.mapinfo) ? strlen(settings.mapinfo) : 0);
    len += ((settings.hostname) ? strlen(settings.hostname) : 0);
    len += ((settings.version) ? strlen(settings.version) : 0);
    len += ((settings.webpage) ? strlen(settings.webpage) : 0);
    len += ((settings.ownername) ? strlen(settings.ownername) : 0);
    len += ((settings.owneremail) ? strlen(settings.owneremail) : 0);
    len += 256;
    buf = safer_malloc(len * sizeof(char));
    buf2 = safer_malloc(len * sizeof(char));
    sprintf(buf, "up portnum:%d maxusers:%d", settings.bindport,
	    settings.maxusers);
    if (settings.name) {
	sprintf(buf2, " nm:\"%s\"", settings.name);
	strcat(buf, buf2);
    }
    if (settings.desc) {
	sprintf(buf2, " desc:\"%s\"", settings.desc);
	    strcat(buf, buf2);
    }
    if (settings.mapinfo) {
	sprintf(buf2, " mapinfo:\"%s\"", settings.mapinfo);
	strcat(buf, buf2);
    }
    if (settings.hostname) {
	sprintf(buf2, " hostname:\"%s\"", settings.hostname);
	    strcat(buf, buf2);
    }
    if (settings.version) {
	sprintf(buf2, " srvversion:\"%s\"", settings.version);
	strcat(buf, buf2);
    }
    if (settings.webpage) {
	sprintf(buf2, " webpage:\"%s\"", settings.webpage);
	strcat(buf, buf2);
    }
    if (settings.ownername) {
	sprintf(buf2, " ownername:\"%s\"", settings.ownername);
	strcat(buf, buf2);
    }
    if (settings.owneremail) {
	sprintf(buf2, " owneremail:\"%s\"", settings.owneremail);
	strcat(buf, buf2);
    }
    if (persist.userhwm) {
	sprintf(buf2, " userhwm:%d", persist.userhwm);
	strcat(buf, buf2);
    }
    if (settings.preferip)
	strcat(buf, " preferip:1\r");
    else
	strcat(buf, "\r");
    do_write_out(out, buf);
    free(buf);
    free(buf2);
    event->dnob = persist.todtick + (600 * settings.mservfreq);

    ogn = new_event(Q_MAINT, ETYPE_NONE, "process_outgoing_queue");
    ogn->out = (struct _servout_t *)out;
    queue_event(ogn, NULL);

    return(0);
}

/**
   \brief move objects around thier mapgroup
   \param event event that called us
   \return event status

   look at all loaded maps, and determine if any require objects in motion
   to be moved
*/

int update_all_maps_object_motion(event_t *event)
{
    maphead_t *maph;
    object_t *obj;
    mapgroup_t * mapg = event_mapg(event);
    assert(mapg);

    TAILQ_FOREACH(maph, &mapg->maps, next) {
	if (QUERY_FLAG(maph, MAPFLAG_LOADED)) {
	    TAILQ_FOREACH(obj, &maph->map->objs, next)
		if (obj->pl == NULL)
		    do_automove_obj(obj);
	}
    }
    return(0);
}

/**
   \brief process this queue
   \param q an event queue to process
   \param q_mutex the mutex for this queue
   \todo need to check if obj are in the current mapg (not sure)
*/

static void process_queue(event_queue_t *queue)
{
    int j;
    event_t * event;
    for (event = event_queue_first(queue); event != NULL; event = event_queue_next(queue)) {
	/* skip over timer events that aren't ripe */
	if ((event->type == ETYPE_TIMER && nonpersist.pticks%event->dnob != 0)
	    || (event->type == ETYPE_DNOB && persist.todtick < event->dnob)
	    || !QUERY_FLAG(event, EVFLAG_QUEUED))
	    continue;

	if (event->initiator && !obj_exists(event->initiator)) {
		delete_event(event);
		continue;
	}

	/* event with EVFLAG_INITIATOR need an initiator */
	if(QUERY_FLAG(event,EVFLAG_INITIATOR)  && !event->initiator) {
		delete_event(event);
		continue;
	};
	/* delete the event with flag DEL_NOT_IN_MAPG if the initiator isn't in the mapg */
	if( QUERY_FLAG(event,EVFLAG_DEL_NOT_IN_MAPG)
		&& (!event->initiator || event->initiator->mapgroup != queue->mapg)) {
		delete_event(event);
		continue;
	};
	/* event with EVFLAG_FOLLOW_OBJ follow the obj */
	if(QUERY_FLAG(event,EVFLAG_FOLLOW_OBJ)
		&& event->initiator && event->initiator->mapgroup != event_mapg(event)) {
		queue_event(event,event->initiator->mapgroup);
		continue;
	};
	/* event with EVFLAG_SERVER_IGNORE are ignored in server queue */
	if(QUERY_FLAG(event,EVFLAG_SERVER_IGNORE) && !queue->mapg)
		continue;
	j = find_efunc_by_event(event);

	if ( event_func_map[j].function(event) == 1) {
		delete_event(event);
	}
	sched_yield();
    }
}

/**
   \brief the mapgroup event loop
   \param mapgroup a void pointer to a mapgroup that this queue runs in
*/

void *event_loop(void *mapgroup)
{
    int i;
    mapgroup_t *mapg;

    mapg = (mapgroup_t *)mapgroup;

    while(1) {
	sleep_delta();
	for (i=0; i < Q_MAX; i++) {
/*	    if (TAILQ_EMPTY(&mapg->queue[i]))
		continue;*/
	    process_queue(&mapg->queue[i]);
	}
    }
}

/** The main server event loop */

void *server_event_loop(void *nada)
{
    int i;

    while(1) {
	sleep_delta();
	nonpersist.pticks++;
	if (nonpersist.pticks%PTICKS_PER_CLOCK == 0)
	    persist.todtick++;
	for (i=0;  i < Q_MAX; i++) {
	    process_queue(&server_queue[i]);
	}
    }
}


/*==========  function on event_queue ==============*/


static void _event_queue_clear_cur(event_queue_t *queue)
{
    event_t *evt;
	
    if (queue->_cur ) {
	evt = queue->_cur;
	queue->_cur = NULL;
	if (queue->_cur_action&1) {
	    delete_event_nl(evt);
	} else if (queue->_cur_action&2) {
	    event_queue_move_nl(evt, queue->_cur_move);
	}
	queue->_cur_action = 0;
    }
}


/**
   \brief innitialize data on event_queue 
   \param queue queue to initialize
*/

void event_queue_init(event_queue_t *queue, mapgroup_t *mapg)
{
    memset(queue, 0, sizeof(event_queue_t));
    TAILQ_INIT(&queue->queue);
    queue->mapg = mapg;
}

/**
   \brief get the first event in the queue
   \see event_queue_next
*/

event_t *event_queue_first(event_queue_t *queue)
{
    event_t *res;

    pthread_mutex_lock(&queue_mutex);
    _event_queue_clear_cur(queue);
    queue->_cur= res = TAILQ_FIRST(&queue->queue);
    pthread_mutex_unlock(&queue_mutex);
    return res;
}

/**
   \brief get the next event in the queue
   \see event_queue_first
*/

event_t *event_queue_next(event_queue_t *queue)
{
    event_t *res;

    pthread_mutex_lock(&queue_mutex);
    if (!queue->_cur) {
	pthread_mutex_unlock(&queue_mutex);
	return NULL;
    }
    res = TAILQ_NEXT(queue->_cur, next);
    _event_queue_clear_cur(queue);
    queue->_cur= res;
    pthread_mutex_unlock(&queue_mutex);
    return res;
}

/** get an event_queue from a mapgroup and an int */
event_queue_t * get_event_queue(mapgroup_t *mapg, int n)
{
    if (mapg)
	return &mapg->queue[n];
    else
	return &server_queue[n];
}

/** return the map_group of un event (or null) */
struct _mapgroup_t * event_mapg(event_t *e) {
	struct _mapgroup_t *res;
	pthread_mutex_lock(&queue_mutex);
	res = e->event_queue?e->event_queue->mapg:NULL;
	pthread_mutex_unlock(&queue_mutex);
	return res;
};

/** \brief delete all event in obj->events */
void delete_all_obj_event(struct _object_t *obj) {
    event_t *event,*nevent;
    pthread_mutex_lock(&queue_mutex);
    for(event = TAILQ_FIRST(&obj->events);event !=NULL;event=nevent) {
	nevent = TAILQ_NEXT(event,stack);
	delete_event_nl(event);
    };
    pthread_mutex_unlock(&queue_mutex);
};
/** \brief unqueue all event in obj->events */
void unqueue_all_obj_event(struct _object_t *obj) {
    event_t *event,*nevent;
    pthread_mutex_lock(&queue_mutex);
    for(event = TAILQ_FIRST(&obj->events);event !=NULL;event=nevent) {
	nevent = TAILQ_NEXT(event,stack);
	event_queue_move_nl(event,NULL);
    };
    pthread_mutex_unlock(&queue_mutex);
};
/** \brief queue all event in obj->events */
void queue_all_obj_event(struct _object_t *obj,struct _mapgroup_t *mapg) {
    event_t *event,*nevent;
    pthread_mutex_lock(&queue_mutex);
    for(event = TAILQ_FIRST(&obj->events);event !=NULL;event=nevent) {
	nevent = TAILQ_NEXT(event,stack);
	event_queue_move_nl(event,get_event_queue(mapg, event->queue));
    };
    pthread_mutex_unlock(&queue_mutex);
};


/*================ evarg_* func ========================= */
void evarg_clear(event_t * event,int num) {
    assert ( (num>=0) && ( num <= EVARG_NBR) );
    switch(event->arg_type[num]) {
	case EVARG_DSTRING : 
	    if ( event->arg_val[num].c )
		free (event->arg_val[num].c);
	    break;
	case EVARG_ASTRING : 
	    if ( event->arg_val[num].c )
		free_string(event->arg_val[num].c);
	    break;
	case EVARG_OBJ : 
	    if (event->arg_val[num].obj ) obj_decref(event->arg_val[num].obj);
	    break;
    };
    
    event->arg_type[num] = EVARG_NONE;
};

/** set event arg as int */
void evarg_setint (event_t * event,int num,int v) {
    assert ( (num>=0) && ( num <= EVARG_NBR) );
    evarg_clear(event,num);
    event->arg_val[num].i = v;
    event->arg_type[num] = EVARG_INT;
};
    
    
    
/** set event arg as float */
void evarg_setfloat(event_t * event,int num,float v) {
    assert ( (num>=0) && ( num <= EVARG_NBR) );
    evarg_clear(event,num);
    event->arg_val[num].f = v;
    event->arg_type[num] = EVARG_FLOAT;
};
/** set event arg as dynamically allocated string */
void evarg_setdstr(event_t * event,int num,char *s) {
    assert ( (num>=0) && ( num <= EVARG_NBR) );
    evarg_clear(event,num);
    event->arg_val[num].c = s;
    event->arg_type[num] = EVARG_DSTRING;
};
/** set event arg as static string */
void evarg_setsstr(event_t * event,int num,char *s) {
    assert ( (num>=0) && ( num <= EVARG_NBR) );
    evarg_clear(event,num);
    event->arg_val[num].c = s;
    event->arg_type[num] = EVARG_SSTRING;
};
/** set event arg as string allocated with add_string */
void evarg_setastr(event_t * event,int num,char *s) {
    assert ( (num>=0) && ( num <= EVARG_NBR) );
    evarg_clear(event,num);
    event->arg_val[num].c = s;
    event->arg_type[num] = EVARG_ASTRING;
};
/** \brief set event arg as object 
\note  call obj_incref
*/
void evarg_setobj(event_t * event,int num,struct _object_t *obj){
    assert ( (num>=0) && ( num <= EVARG_NBR) );
    if(obj) obj_incref(obj);
    evarg_clear(event,num);
    event->arg_val[num].obj = obj;
    event->arg_type[num] = EVARG_OBJ;
};

/** \brief get an event arg as int */
int evarg_getint(event_t * event,int num) {
    assert ( (num>=0) && ( num <= EVARG_NBR)  && (event->arg_type[num] ==EVARG_INT));
    return event->arg_val[num].i;
};

/** \brief get an event arg as float */
float evarg_getfloat(event_t * event,int num) {
    assert ( (num>=0) && ( num <= EVARG_NBR)  && (event->arg_type[num] ==EVARG_FLOAT));
    return event->arg_val[num].f;
};
/** \brief get an event arg as string */
char * evarg_getstr(event_t * event,int num){
    assert ( (num>=0) && ( num <= EVARG_NBR)  &&( 
	(event->arg_type[num] ==EVARG_DSTRING)||(event->arg_type[num] ==EVARG_ASTRING)
	    ||(event->arg_type[num] ==EVARG_SSTRING)));
    return event->arg_val[num].c;
};
/** \brief get an event arg as int */
struct _object_t * evarg_getobj(event_t * event,int num){
    assert ( (num>=0) && ( num <= EVARG_NBR)  && (event->arg_type[num] ==EVARG_OBJ));
    return event->arg_val[num].obj;
};


