/* $Id: object.h,v 1.30 2004/03/30 06:45:55 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/
#ifndef _OBJECT_H_
#define _OBJECT_H_

/**
   \file object.h
   \brief Object definitions
   \author Tim Rightnour
*/

#include "defines.h"
#include "uid.h"

/** 
    \brief Things that are transferred to the client.
    \note we need to make compatibility data to version the clients
*/

typedef struct _client_t {
    int nrof;		/**< \brief number of obj in pile */
    int state;		/**< \brief state of obj */
    char *name;		/**< \brief name */
    char *name_pl;	/**< \brief plural name */
    char *model;	/**< \brief filename of 3D model */
    vertex_t facing;	/**< \brief direction facing */
    vertex_t direction; /**< \brief direction moving */
    vertex_t coord;	/**< \brief coordinates */
    vertex_t prevc;	/**< \brief previous coordinates */
    vertex_t scale;	/**< \brief scale of object */
    stats_t stats;	/**< \brief stats of object */
    equip_t equip;	/**< \brief equipment data */
    float velocity;	/**< \brief how fast are we moving in direction? */
    uniqid_t uid;	/**< \brief uniq id*/
} client_t;

/** \brief object structure */
typedef struct _object_t object_t;

/**
   \brief Object buckets
We place these into a bucket to keep them all in one handy place.  This
avoids problems like in crossfire where you want to find a job, and have
to sort through the entire inventory of spells and items to find it.
*/

typedef struct _bucket_t {
    TAILQ_HEAD(, _object_t) spells;	/**< \brief ptr to first spell */
    TAILQ_HEAD(, _object_t) jobs;  	/**< \brief ptr to first job */
    TAILQ_HEAD(, _object_t) inv;   /**< \brief ptr to first obj in inventory */
    TAILQ_HEAD(, _object_t) skills;	/**< \brief ptr to first skill */
    TAILQ_HEAD(, _object_t) keys;	/**< \brief key items, forces */
    TAILQ_HEAD(, _object_t) equip;	/**< \brief equipment */
} bucket_t;

/**
   \brief Bits of useful information about this object, or that this object
   knows
*/

typedef struct _lore_t {
    char *text;	/**< \brief the actual lore */
    int chance;	/**< \brief chance out of 100 */
    TAILQ_ENTRY(_lore_t) next; /**< linked list */
} lore_t;

/**
   \brief keywords that trigger a dialog string
*/

typedef struct _dkeyword_t {
    char *key;				/**< \brief keyword */
    TAILQ_ENTRY(_dkeyword_t) next;	/**< \brief Next keyword */
} dkeyword_t;

/**
   \brief things this NPC has to say, when triggered by keywords
*/

typedef struct _dialog_t {
    char *text;	/**< \brief what the NPC has to say */
    TAILQ_HEAD(, _dkeyword_t) keywords;	/**< \brief what triggers him to say it */
    TAILQ_ENTRY(_dialog_t) next; /**< \brief next dialog entry */
} dialog_t;

/**
   \brief flag structure
*/

typedef struct _eflag_t {
    char *flag;	/**< \brief Flag name */
    int num;	/**< \brief flag number */
} eflag_t;

#define FLAGARRAYSZ 3	/**< \brief size of flags in object_t */
/**
   \brief basic obj structure, needs much much more
*/

struct _object_t {
    char *arch;			  /**< \brief archetype name */
    TAILQ_ENTRY(_object_t) next;  /**< \brief doubly linked sorted list */
    TAILQ_ENTRY(_object_t) stack; /**< \brief tail queue, stacking */
    struct _mapgroup_t *mapgroup; /**< \brief pointer to mapgroup obj is on */
    struct _maphead_t *maphead;	/**< \brief pointer to map obj is on */
    object_t *connected;       /**< \brief object this item is connected to. */
    object_t *archobj;		/**< \brief pointer to arch obj */
    object_t *owner;		/**< \brief object that owns this object */
    object_t *container;	/**< \brief container this object is in */
    object_t *master;		/**< \brief this object's master */
    int flags[FLAGARRAYSZ];	/**< \brief flags.. ala crossfire */
    char *msg;			/**< \brief description of object */
    DtShapeRef shape;		/**< \brief SOLID model */
    int shaperef;		/**< \brief reference count for shape */
    bucket_t bucket;	     /**< \brief the bucket that holds other objects */
    client_t cli;		/**< \brief client shared data */
    struct _player_t *pl;      	/**< \brief pointer to player record */
    int collision;		/**< \brief >0 if collision occurred */
    vertex_t bmin, bmax, coff;	/**< \brief bounding box points */
    vertex_t center;		/**< \brief rotational center */
    float radius;		/**< \brief radius of sphere enclosing obj */
    float gravity;		/**< \brief gravitational velocity */
    int refcount; /**< @brief number of reference on this object @see obj_incref,obj_decref */
    TAILQ_HEAD(, _lore_t) lore;	/**< \brief lore about the item */
    TAILQ_HEAD(, _dialog_t) dialog;	/**< \brief chatty NPC's */
    TAILQ_HEAD(, _event_t) events;	/**< \brief events owned by this obj */
    TAILQ_ENTRY(_object_t) hash_next;  /**< \brief in obj_table */

};

#include "player.h"
#include "map.h"


struct _object_t * new_obj_from_arch(char * arch,  int recursive);
 
object_t *find_arch_byname(char *name);
object_t *new_obj(void);
void remove_obj(object_t *obj);
void do_automove_obj(object_t *obj);
void put_obj_in_space(object_t *obj);
void collision_map_obj(object_t *obj, vertex_t coord);
int place_obj_and_test(object_t *obj, vertex_t coord);
void remove_obj_from_space(object_t *obj);
void put_obj_in_obj(object_t *obj, object_t *cont, int slot);
void remove_obj_from_obj(object_t *obj, object_t *cont, int slot);
int verify_obj_in_bucket(object_t *obj, object_t *cont, int slot);
int verify_obj_in_obj(object_t *obj, object_t *target);
void unload_obj_space(object_t *obj);
void duplicate_object(object_t *from, object_t *to, int recursive);
void parse_dialog(xmlDocPtr doc, xmlNodePtr cur, object_t *obj);
void init_flags(void);
void parse_equipflags(xmlDocPtr doc, xmlNodePtr cur, object_t *obj);
void parse_flags(xmlDocPtr doc, xmlNodePtr cur, object_t *obj);
void parse_lore(xmlDocPtr doc, xmlNodePtr cur, object_t *obj);
void parse_object(xmlDocPtr doc, xmlNodePtr cur, object_t *obj, int mode);
void parse_archfile(char *filename, char *path);

/** @brief increment the object refcount
    @note it use the objtable_mutex
	 @see obj_decref
**/
void obj_incref(object_t *obj);
/** @brief decrement the object refcount
* The memory will be freed if th recount goes to 0
    @note it use the objtable_mutex
	 @see obj_incref
**/
void obj_decref(object_t *obj);


/** @brief object exists (not destructed )
    @return 0 if object is destructed 
    @todo need to implement it corectly
**/
int obj_exists(object_t *obj);

/** @brief is object an arch
    @return 0 if object isn't an arch
**/
int obj_isarch(object_t *obj);


void init_uid_obj_table();
void add_uid_obj_table(object_t *obj);
void rem_uid_obj_table(object_t *obj);
struct _object_t *get_obj_from_uid(uniqid_t *id);
 
#endif /* _OBJECT_H_ */
