/* $Id: loader.c,v 1.30 2004/03/30 02:54:55 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "map.h"
#include "shstr.h"
#include "utils.h"
#include "event.h"
#include "player.h"
#include "network.h"
#include "netparser.h"
#include "collision.h"
#include "commands.h"
#include "settings.h"
#include "servdata.h"
#include "xmlparse.h"
#include "sutils.h"

/**
   \file loader.c
   \brief The main initialization of the server
   \author Tim Rightnour
*/

pthread_mutex_t maps_mutex;	/**< The mutex protecting all mapgroups */
pthread_mutex_t arch_mutex;	/**< The mutex protecting the archtable */
pthread_mutex_t players_mutex;	/**< The mutex protecting the playertable */
pthread_mutex_t dlist_mutex;	/**< The descriptor list mutex */
pthread_mutex_t objtable_mutex;	/**< The object table mutex */
pthread_mutex_t collision_mutex;/**< The collsion mutex */
const char *progrname;		/**< The name of this binary */
settings_t settings;		/**< Server settings */
persist_data_t persist;		/**< Persistent data */
npersist_data_t nonpersist;	/**< Non-Persistent data */

extern TAILQ_HEAD(, _mapgroup_t) maps;

/**
   \brief Initialize all non-persistent data
*/

void init_nonpersist(void)
{
    memset(&nonpersist, 0, sizeof(npersist_data_t));
    /* override any that should not be initted to 0 below */
    nonpersist.objectcount = 1;
}

/**
   \brief Write out all persistent data
*/

void write_persist(void)
{
    char file[MAX_PATH], *sf;
    xmlDocPtr doc;
    xmlNodePtr tree;

    sf = "serverdata.xml";

    sprintf(file, "%s/%s", settings.vardir, sf);
    doc = xmlNewDoc("1.0");
    doc->children = xmlNewDocNode(doc, NULL, "ServerData", NULL);
    tree = doc->children;
    write_long(tree, "todtick", persist.todtick);
    write_int(tree, "userhwm", persist.userhwm);
    write_uid(tree, "uid_seed", &(persist.uid_seed));
    xmlSaveFormatFile(file, doc, 1);
    xmlFreeDoc(doc);
}

/**
   \brief Initialize all persistent data
*/

void init_persist(void)
{
    char file[MAX_PATH], *sf;
    xmlDocPtr doc;
    xmlNodePtr cur;

    memset(&persist, 0, sizeof(persist_data_t));

    sf = "serverdata.xml";

    sprintf(file, "%s/%s", settings.vardir, sf);
    doc = xmlParseFile(file);
    if (doc == NULL) {
        /* we don't know any data yet.  now what? */
        LOG(LOG_ERROR, "No ServerData file at %s.", file);
        return;
    }
    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        LOG(LOG_ERROR, "ServerData file %s is empty.", file);
        xmlFreeDoc(doc);
        return;
    }
    if (strcmp(cur->name, "ServerData")) {
        LOG(LOG_ERROR, "%s: root node != ServerData", file);
        xmlFreeDoc(doc);
        return;
    }
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "todtick"))
	    parselong(doc, cur, &persist.todtick);
	if (!strcmp(cur->name, "userhwm"))
	    parseint(doc, cur, &persist.userhwm);
	if (!strcmp(cur->name, "uid_seed"))
	    parseuid(doc, cur, &persist.uid_seed);
	cur = cur->next;
    }
    xmlFreeDoc(doc);
}

/**
   \brief initialze server settings
*/

void init_settings(void)
{
    settings.archdir = PACKAGE_DATA_DIR ARCH_SUFX;
    settings.etcdir = PACKAGE_DATA_DIR ETC_SUFX;
    settings.mapdir = PACKAGE_DATA_DIR MAP_SUFX;
    settings.texdir = PACKAGE_DATA_DIR TEX_SUFX;
    settings.vardir = PACKAGE_DATA_DIR VAR_SUFX;
    settings.bindport = BIND_PORT;
    settings.mservport = 5502;
    settings.mservhost = "metaserver.genecys.org";
    /* by default, do not connect to metaserver */
    settings.mservfreq = 0;
    settings.name = NULL;
    settings.hostname = NULL;
    settings.mapinfo = NULL;
    settings.desc = NULL;
    /* eventually, this will be automated */
    settings.version = NULL;
    settings.webpage = NULL;
    settings.ownername = NULL;
    settings.owneremail = NULL;
    /* default to 5 max users.  for now */
    settings.maxusers = 5;
    settings.preferip = 0;
}

/**
   \brief parse the settings file
*/

void parse_settings(void)
{
    char file[MAX_PATH], *sf;
    xmlDocPtr doc;
    xmlNodePtr cur;

    sf = "genserverconfig.xml";

    sprintf(file, "./%s", sf);
    doc = xmlParseFile(file);
    if (doc == NULL) {
	sprintf(file, "%s/%s", settings.etcdir, sf);
	doc = xmlParseFile(file);
    }
    if (doc == NULL) {
	sprintf(file, "/etc/%s", sf);
	doc = xmlParseFile(file);
    }
    if (doc == NULL) {
	LOG(LOG_WARNING, "Warning, cannot find settings file %s", sf);
	return;
    }
    /* if we got this file, we found one */

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        LOG(LOG_WARNING, "Settings file %s is empty", file);
        xmlFreeDoc(doc);
	return;
    }
    if (strcmp(cur->name, "ServerSettings")) {
        LOG(LOG_WARNING,"%s: root node != ServerSettings", file);
        xmlFreeDoc(doc);
	return;
    }
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "archdir"))
	    parsestring(doc, cur, &settings.archdir);
	if (!strcmp(cur->name, "etcdir"))
	    parsestring(doc, cur, &settings.etcdir);
	if (!strcmp(cur->name, "vardir"))
	    parsestring(doc, cur, &settings.vardir);
	if (!strcmp(cur->name, "mapdir"))
	    parsestring(doc, cur, &settings.mapdir);
	if (!strcmp(cur->name, "texdir"))
	    parsestring(doc, cur, &settings.texdir);
	if (!strcmp(cur->name, "bindport"))
	    parseint(doc, cur, &settings.bindport);
	if (!strcmp(cur->name, "mservport"))
	    parseint(doc, cur, &settings.mservport);
	if (!strcmp(cur->name, "mservhost"))
	    parsestring(doc, cur, &settings.mservhost);
	if (!strcmp(cur->name, "mservfreq"))
	    parseint(doc, cur, &settings.mservfreq);
	if (!strcmp(cur->name, "name"))
	    parsestring(doc, cur, &settings.name);
	if (!strcmp(cur->name, "desc"))
	    parsestring(doc, cur, &settings.desc);
	if (!strcmp(cur->name, "mapinfo"))
	    parsestring(doc, cur, &settings.mapinfo);
	if (!strcmp(cur->name, "hostname"))
	    parsestring(doc, cur, &settings.hostname);
	if (!strcmp(cur->name, "version"))
	    parsestring(doc, cur, &settings.version);
	if (!strcmp(cur->name, "webpage"))
	    parsestring(doc, cur, &settings.webpage);
	if (!strcmp(cur->name, "ownername"))
	    parsestring(doc, cur, &settings.ownername);
	if (!strcmp(cur->name, "owneremail"))
	    parsestring(doc, cur, &settings.owneremail);
	if (!strcmp(cur->name, "maxusers"))
	    parseint(doc, cur, &settings.maxusers);
	if (!strcmp(cur->name, "preferip"))
	    parseint(doc, cur, &settings.preferip);
	cur = cur->next;
    }
    xmlFreeDoc(doc);
}

/**
   \brief Main
   \param argc argument count
   \param argv argument vector
   \return success/fail
*/

int main(int argc, char **argv)
{
    pthread_t network_thread;
    pthread_t server_event_thread;
    maphead_t *maph;
    mapgroup_t *mapg;

    xmlKeepBlanksDefault(0);
    init_uid_obj_table();
    init_hash_table();
#ifdef __NetBSD__
    progrname = getprogname();
    setproctitle(NULL);
#endif
    openlog();
    dtDisableCaching();
    dtSetDefaultResponse(collide2, DT_SIMPLE_RESPONSE, NULL);
    init_time();
    init_argcomm();
    init_settings();
    parse_settings();
    init_nonpersist();
    init_persist();
    init_efuncrec();
    init_server_queues();
    init_server_events();
    init_flags();
    init_commands();

    init_mutex(&maps_mutex);
    init_mutex(&arch_mutex);
    init_mutex(&players_mutex);
    init_mutex(&dlist_mutex);
    init_mutex(&objtable_mutex);
    init_mutex(&collision_mutex);

    parse_archfile("archetypes.xml", settings.etcdir);
    parse_mapgroups("mapgroups.xml", settings.mapdir);

    /* we don't have to lock here, as the threads aren't running yet */
    TAILQ_FOREACH(mapg, &maps, next) {
        TAILQ_FOREACH(maph, &mapg->maps, next) {
            if (QUERY_FLAG(maph, MAPFLAG_STARTING)) {
		set_map_loaded(maph);
            }
        }
    }

    init_network();

    /* fire up the net */
    pthread_create(&network_thread, NULL, network_loop, NULL);
    /* fire up the server event queue */
    pthread_create(&server_event_thread, NULL, server_event_loop, NULL);

    /* fire up all the mapgroup event queues */
    pthread_mutex_lock(&maps_mutex);
    TAILQ_FOREACH(mapg, &maps, next) {
	init_queues(mapg);
	init_events(mapg);
	pthread_create(&mapg->thread, NULL, event_loop, (void *)mapg);
    }
    pthread_mutex_unlock(&maps_mutex);

    /* we only join the server event thread, as that is the only thread that
       can issue a server shutdown. */

    pthread_join(server_event_thread, NULL);

    closelog();

    return(0);
}
