/* $Id: map.c,v 1.31 2004/03/30 06:45:55 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <math.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <png.h>
#include <pthread.h>

#include "defines.h"
#include "object.h"
#include "map.h"
#include "shstr.h"
#include "utils.h"
#include "models.h"
#include "pngloader.h"
#include "xmlparse.h"
#include "collision.h"
#include "matrix.h"
#include "servdata.h"
#include "sutils.h"
#include "stattables.h"

extern persist_data_t persist;
extern npersist_data_t nonpersist;
extern pthread_mutex_t maps_mutex;

TAILQ_HEAD(, _mapgroup_t) maps = TAILQ_HEAD_INITIALIZER(maps);
maphead_t *activemap;

/**
   \file map.c
   \brief Map manipulation routines
   \author Tim Rightnour
*/

/**
   \brief find a map by it's name
   \param name name to match
   \param mapg mapgroup to look in
   \return a pointer to a maphead_t of the located map, NULL if not found
*/

maphead_t *find_map_byname(char *name, mapgroup_t *mapg)
{
    maphead_t *maph;
    size_t len;

    len = strlen(name);
    TAILQ_FOREACH(maph, &mapg->maps, next)
	if (!strncmp(maph->name, name, len))
	    return(maph);

    return(NULL);
}

/**
   \brief find a map by it's mapnumber within the group
   \param num mapnumber to find
   \param mapg mapgroup to search
   \return a pointer to a maphead_t of the located map, NULL if not found
*/

maphead_t *find_map_bynum(int num, mapgroup_t *mapg)
{
    maphead_t *maph;

    TAILQ_FOREACH(maph, &mapg->maps, next) {
	if (maph->mapnumber == num) {
	    return(maph);
	}
    }
    return(NULL);
}

/**
   \brief Load a map
   \param maph maphead to load
   \note This function is nowhere near complete.  It still needs to load in
   the actual map elevation and terrain, rather than have that permanently
   loaded in the game.  In addition, objects should probably not be loaded
   on the map at all until this is called.
*/

void set_map_loaded(maphead_t *maph)
{
    object_t *obj;

    if (QUERY_FLAG(maph, MAPFLAG_LOADED))
	return;

    TAILQ_FOREACH(obj, &maph->map->objs, next) {
	put_obj_in_space(obj);
    }
    /* don't forget to collisionify the ground itself */
    SET_FLAG(maph, MAPFLAG_LOADED);
}

/**
   \brief Unload a map
   \param maph maphead to unload
   \note This function is nowhere near complete.  It still needs to unload in
   the actual map elevation and terrain, rather than have that permanently
   loaded in the game.  In addition, objects should probably be unloaded
   from the map when this is called.
*/

void set_map_unloaded(maphead_t *maph)
{
    object_t *obj;

    TAILQ_FOREACH(obj, &maph->map->objs, next) {
	remove_obj_from_space(obj);
    }
    /* remove the map's collision */
    CLEAR_FLAG(maph, MAPFLAG_LOADED);
}

/**
   \brief removes an object from a map
   \param obj object to remove from map
   \param maph maphead to remove from
   \param space should we remove the object from collision as well?
   \param group should we remove the object from the whole mapgroup?
*/

void remove_obj_from_map(object_t *obj, maphead_t *maph, int space, int group)
{
    event_t *event, *enext;

    TAILQ_REMOVE(&maph->map->objs, obj, next);
    obj->maphead = NULL;
    if (group) {
	/* if we are removing the object from the mapgroup, we need to
	   pack up our events, and pull them off the queue */
	for (event = TAILQ_FIRST(&obj->events); event != NULL; event = enext) {
	    enext = TAILQ_NEXT(event, stack);
	    unqueue_event(event);
	}
	obj->mapgroup = NULL;
    }

    if (space)
	remove_obj_from_space(obj);
}

/**
   \brief place an object on the map.
   \param obj object to place on map
   \param maph maphead to place object on
   \param space should we place the object into collision as well?

   This places it in the physical plane of the map, and instansiates the
   SOLID modeling for that object. It is expected that all vertex_t's for
   this object have been filled out prior to this call.  See POOM_*
*/

void put_obj_on_map(object_t *obj, maphead_t *maph, int space)
{
    if (TAILQ_EMPTY(&maph->map->objs))
	TAILQ_INIT(&maph->map->objs);
    /* put it on the linked list of objs on the map */

    obj->maphead = maph;
    obj->mapgroup = maph->mapg;
    TAILQ_INSERT_TAIL(&maph->map->objs, obj, next);

    /* a y value of -99999.0 means put this at groundheight */
    if (obj->cli.coord.y == -99999.0)
	obj->cli.coord.y = groundheight_at_mapcoord(maph, obj->cli.coord);

    if (space)
	put_obj_in_space(obj);
}

/**
   \brief place an object on the map as close as we can to a given location
   \param obj object to place on map
   \param maph maphead to place object on
   \param space should we place the object into collision as well?
   \note if space is set to FALSE, then we just leave it where we put it
   \warning Calls dtTest()
*/

void put_obj_on_map_closeto(object_t *obj, maphead_t *maph, int space)
{
    int i;
    vertex_t newcoord, moved, dir;
    float ghy, dis, zwidth;

    put_obj_on_map(obj, maph, space);

    if (space == POOM_NOSPACE)
	return;
    if (obj->cli.model == NULL || QUERY_FLAG(obj, FLAG_NOCOLLIDE))
	return;

    if (place_obj_and_test(obj, obj->cli.coord) == 0)
	return;

    obj->collision = 0;
    zwidth = fabs(obj->bmin.z - obj->bmax.z)+fabs(obj->bmin.x - obj->bmax.x);
    zwidth *= 3.0;
    printf("HELLO ZWIDTH %f\n", zwidth);
    for (i=0; i < 1000; i++) {
	newcoord = obj->cli.coord;
	dir.y = (float)rndm(0, 359);
	dir.x = dir.z = 0.0;
	dis = zwidth * ((float)rndm(-i-1, i+1) / 5.0);
	printf("DISTANCE %f\n", dis);
	new_coord(dir, dis, newcoord, &moved);
	newcoord.x += moved.x;
	newcoord.y += moved.y;
	newcoord.z += moved.z;
	ghy = groundheight_at_mapxz(obj->maphead, newcoord.x, newcoord.z);
	if (ghy > newcoord.y)
	    moved.y += ghy - newcoord.y;
	if (place_obj_and_test(obj, newcoord) > 0) {
	    obj->collision = 0;
	    continue;
	} else {
	    obj->cli.coord = newcoord;
	    remove_obj_from_map(obj, maph, space, 0);
	    put_obj_on_map(obj, maph, space);
	    return;
	}
    }
    LOG(LOG_WARNING, "Could not find room for obj %s on map! Removing.",
	obj->cli.name);
    remove_obj_from_map(obj, maph, space, 1);
    return;
}

/**
   \brief is the object in range from the given border?
   \param obj object to test
   \param border direction (0-8) to test
   \param range how far is "in range"?
   \return true/false
*/

int in_range_from_border(object_t *obj, int border, float range)
{
    float minx, minz, maxx, maxz;

    minx = obj->maphead->x;
    minz = obj->maphead->z;
    maxx = minx+(obj->maphead->width*SCALE_XZ);
    maxz = minz+(obj->maphead->height*SCALE_XZ);

    switch(border) {
    case 0: 
	return(0);
    case 1: 
	if (xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x+range,
				obj->cli.coord.z-range) ||
	    (obj->cli.coord.x+range > maxx && obj->cli.coord.z-range < minz))
	return(1);
    case 2:
	if (xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x+range,
			       obj->cli.coord.z) ||
	    obj->cli.coord.x+range > maxx)
	return(1);
    case 3:
	if (xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x+range,
			    obj->cli.coord.z+range) ||
	    (obj->cli.coord.x+range > maxx && obj->cli.coord.z+range < maxz))
	return(1);
    case 4:
	if (xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x,
			       obj->cli.coord.z+range) ||
	    obj->cli.coord.z+range > maxz)
	return(1);
    case 5:
	if (xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x-range,
			    obj->cli.coord.z+range) ||
	    (obj->cli.coord.x-range < minz && obj->cli.coord.z+range > maxz))
	return(1);
    case 6:
	if (xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x-range,
			       obj->cli.coord.z) ||
	    obj->cli.coord.x-range < minx)
	return(1);
    case 7:
	if (xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x-range,
			    obj->cli.coord.z-range) ||
	    (obj->cli.coord.x-range < minx && obj->cli.coord.z-range < minz))
	return(1);
    case 8:
	if (xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x,
			       obj->cli.coord.z-range) ||
	    obj->cli.coord.z-range < minz)
	return(1);
    }
    return(0);
}

/**
   \brief is an object close to a border?
   \param obj object to test
   \return true/false

   discover if an object is within the "red zone" of a map.  That being the
   edge of them, close to an adjoining map.  Returns an int, telling us
   which of the 8 directions we have wandered close to.  0 if we are
   "centered".  Uses the perception of the char to determine if he can see
   past an edge.
*/

int obj_close_to_border(object_t *obj)
{
    float minx, minz, maxx, maxz, d;

    minx = obj->maphead->x;
    minz = obj->maphead->z;
    maxx = minx+(obj->maphead->width*SCALE_XZ);
    maxz = minz+(obj->maphead->height*SCALE_XZ);
    d = vision_dist(obj->cli.stats.Per);

    if (!xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x+d,
			    obj->cli.coord.z-d) &&
	obj->cli.coord.x+d > maxx && obj->cli.coord.z-d < minz)
	return(1);
    if (!xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x+d,
			    obj->cli.coord.z+d) &&
	obj->cli.coord.x+d > maxx && obj->cli.coord.z+d > maxz)
	return(3);
    if (!xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x-d,
			    obj->cli.coord.z+d) &&
	obj->cli.coord.x-d < minz && obj->cli.coord.z+d > maxz)
	return(5);
    if (!xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x-d,
			    obj->cli.coord.z-d) &&
	obj->cli.coord.x-d < minx && obj->cli.coord.z-d < minz)
	return(7);
    if (!xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x+d, obj->cli.coord.z)
	&& obj->cli.coord.x+d > maxx)
	return(2);
    if (!xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x, obj->cli.coord.z+d)
	&& obj->cli.coord.z+d > maxz)
	return(4);
    if (!xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x-d, obj->cli.coord.z)
	&& obj->cli.coord.x-d < minx)
	return(6);
    if (!xz_out_of_mapgroup(obj->maphead, obj->cli.coord.x, obj->cli.coord.z-d)
	&& obj->cli.coord.z-d < minz)
	return(8);
    return(0);
}

/**
   \brief where is this map in relation to that map
   \param old map to look from
   \param new map to look to
   \return 0-8 direction, -1 for impossible

   compute the direction shift (1-8) between map old and map new.  return 0
   if they are the same.  returns -1 if you can't get there from here.
*/

int map_shift_direction(maphead_t *old, maphead_t *new)
{
    int i;

    for (i=0; i < 9; i++)
	if (old->tile[i] == new)
	    return(i);

    return(-1);
}

/**
   \brief find which maphead an object is on
   \param obj the object to locate
   \return the maphead that corresponds to the map that the object is truly
   on, given by his coordinates.  This can be used to update obj->maphead
   when an object transitions between two maps.  returns NULL if something
   is horribly wrong.
*/

maphead_t *obj_on_map(object_t *obj)
{
    maphead_t *maph;
    int mapnum;

    mapnum = mapnum_at_xz(obj->maphead, obj->cli.coord.x, obj->cli.coord.z);
    maph = find_map_bynum(mapnum, obj->maphead->mapg);
    return(maph);
}

/* MAP loading routines */

/**
   \brief allocate a new maphead
   \return a new maphead
*/

maphead_t *new_maphead(void)
{
    maphead_t *maph;

    maph = smalloc(maphead_t);
    return(maph);
}

/**
   \brief allocate a new map
   \return a new map
*/

map_t *new_map(void)
{
    map_t *map;

    map = smalloc(map_t);
    return(map);
}

/**
   \brief parse a mapheader into *maph
   \param doc an xmlDocPtr
   \param cur an xmlNodePtr
   \param maph maphead to fill in
*/

void parsemaphead(xmlDocPtr doc, xmlNodePtr cur, maphead_t *maph)
{
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "name"))
	    parsestring(doc, cur, &maph->name);
	if (!strcmp(cur->name, "reset"))
	    parseint(doc, cur, &maph->reset_timeout);
	if (!strcmp(cur->name, "width"))
	    parseint(doc, cur, &maph->width);
	if (!strcmp(cur->name, "height"))
	    parseint(doc, cur, &maph->height);
	if (!strcmp(cur->name, "difficulty"))
	    parseint(doc, cur, &maph->difficulty);
	cur = cur->next;
    }
}

/**
   \brief parse a map into *map
   \param doc an xmlDocPtr
   \param cur an xmlNodePtr
   \param map map to fill in
*/

void parsemap(xmlDocPtr doc, xmlNodePtr cur, map_t *map)
{
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "elevfile"))
	    parsestring(doc, cur, &map->elevfile);
	if (!strcmp(cur->name, "terrfile"))
	    parsestring(doc, cur, &map->terrfile);
	cur = cur->next;
    }
}


#if 0
void write_png_file(char* file_name)
{
    /* create file */
    FILE *fp = fopen(file_name, "wb");
    if (!fp)
	bailout();

    /* initialize stuff */
    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr)
	bailout();

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
	bailout();

    if (setjmp(png_jmpbuf(png_ptr)))
	bailout();

    png_init_io(png_ptr, fp);

    /* write header */
    if (setjmp(png_jmpbuf(png_ptr)))
	bailout();

    png_set_IHDR(png_ptr, info_ptr, width, height,
		 bit_depth, color_type, PNG_INTERLACE_NONE,
		 PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

    png_write_info(png_ptr, info_ptr);

    /* write bytes */
    if (setjmp(png_jmpbuf(png_ptr)))
	bailout();

    png_write_image(png_ptr, row_pointers);

    /* end write */
    if (setjmp(png_jmpbuf(png_ptr)))
	bailout();

    png_write_end(png_ptr, NULL);
}
#endif

/**
   \brief parse a mapfile
   \param filename file to parse
   \param path path to directory where map resides
   \param mapnum the mapnumber of this map
   \param mapg the mapgroup containing this map
   \param basex the base x coordinate of the 0,0,0 point of this map
   \param basey the base y coordinate of the 0,0,0 point of this map
   \param basez the base z coordinate of the 0,0,0 point of this map
   \param mapsize the size of the PNG that contains the actual map
*/

void parse_mapfile(char *filename, char *path, int mapnum, mapgroup_t *mapg,
		   float basex, float basey, float basez, int mapsize)
{
    xmlDocPtr doc;
    xmlNodePtr cur;
    maphead_t *maph;
    map_t *map;
    char *c, file[MAX_PATH];
    object_t *obj;

    doc = xmlParseFile(filename);
    if (doc == NULL)
	bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
	LOG(LOG_FATAL, "%s: empty document", filename);
	xmlFreeDoc(doc);
	bailout();
    }
    if (strcmp(cur->name, "Map")) {
	LOG(LOG_FATAL, "%s: root node != Map", filename);
	xmlFreeDoc(doc);
	bailout();
    }

    if (TAILQ_EMPTY(&mapg->maps))
	TAILQ_INIT(&mapg->maps);

    maph = NULL;
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "Header")) {
	    maph = new_maphead();
	    maph->mapg = mapg;
	    maph->width = mapsize;
	    maph->height = mapsize;
	    maph->mapnumber = mapnum;
	    maph->num = nonpersist.mapcount++;
	    maph->x = basex;
	    maph->y = basey;
	    maph->z = basez;
	    maph->path = add_string(path);
	    parsemaphead(doc, cur, maph);
	    TAILQ_INSERT_TAIL(&mapg->maps, maph, next);
	}
	if (!strcmp(cur->name, "Mapdata")) {
	    map = new_map();
	    parsemap(doc, cur, map);
	    /* we should have elevdata and terrdata, when do we parse that? */
	    if (map->elevfile != NULL) {
		sprintf(file, "%s/%s", path, map->elevfile);
		map->elevation = parse_pngmap8bit(file, maph->width,
						  maph->height);
	    }
	    if (map->terrfile != NULL) {
		sprintf(file, "%s/%s", path, map->terrfile);
		map->terrain = parse_pngmap8bit(file, maph->width,
						maph->height);
	    }
	    /* XXX not in proper order.. handle this better */
	    if (maph == NULL)
		bailout();
	    maph->map = map;
	}
	if (!strcmp(cur->name, "Objects")) {
	    cur = cur->xmlChildrenNode;
	    continue; /* skip the cur->next */
	    /* WARNING: obviously maps are order dependant */
	}
	if (!strcmp(cur->name, "Object")) {
	    c = xmlGetProp(cur, "arch");
	    if (c == NULL) {
		printf("FATAL: Obejct in map without arch set\n");
		bailout();
	    }
	    if (!strcmp(c, "starting_location"))
		SET_FLAG(maph, MAPFLAG_STARTING);
	    obj = new_obj_from_arch(c,DO_NOREC); /* XXX not sure with that  can we add in obj_table before parse_object ?*/
	    free(c);
	    parse_object(doc, cur, obj, PO_MAP);
	    if (obj->cli.nrof == 0)
		obj->cli.nrof = 1;
	    put_obj_on_map(obj, maph, POOM_NOSPACE);
	}
	cur = cur->next;
    }
    xmlFreeDoc(doc);
}

/**
   \brief fill in the tiling information for each map in the mapgroup
   \param mapg the mapgroup to tile
*/

void tile_maps(mapgroup_t *mapg)
{
    int i, j, mapnum, maxh, ii, mw;
    maphead_t *maph;

    if (mapg->nrofmaps == 1)
	return;

    maxh = mapg->nrofmaps/mapg->width;
    mw = mapg->width;

    for (i=0; i < maxh; i++) {
	for (j=0; j < mapg->width; j++) {
	    mapnum = i*mapg->width+j;
	    printf("tiling map %d\n", mapnum);
	    maph = find_map_bynum(mapnum, mapg);
	    if (maph == NULL)
		bailout();
	    maph->tile[0] = maph;
	    if (i > 0) {
		ii = i - 1;
		maph->tile[8] = find_map_bynum(ii*mw+j, mapg);
		if (j < mapg->width-1)
		    maph->tile[1] = find_map_bynum(ii*mw+j+1, mapg);
		if (j > 0)
		    maph->tile[7] = find_map_bynum(ii*mw+j-1, mapg);
	    }
	    if (j < mapg->width-1)
		maph->tile[2] = find_map_bynum(i*mw+j+1, mapg);
	    if (j > 0)
		maph->tile[6] = find_map_bynum(i*mw+j-1, mapg);
	    if (i < maxh) {
		ii = i + 1;
		maph->tile[4] = find_map_bynum(ii*mw+j, mapg);
		if (j < mapg->width-1)
		    maph->tile[3] = find_map_bynum(ii*mw+j+1, mapg);
		if (j > 0)
		    maph->tile[5] = find_map_bynum(ii*mw+j-1, mapg);
	    }
	}
    }
}

/**
   \brief parse a mapgroup file
   \param filename file to parse
*/

void parse_mapgroup(char *filename)
{
    xmlDocPtr doc;
    xmlNodePtr cur;
    char *path, *p;
    mapgroup_t *mapg;
    int mapsize, i, j, xa, xb;
    float baseelev, basex, basez;
    char *mapspec, mapfile[MAX_PATH], pre[256], mid[256], post[256];

    doc = xmlParseFile(filename);
    if (doc == NULL)
	bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
	LOG(LOG_FATAL, "%s: empty document", filename);
	xmlFreeDoc(doc);
	bailout();
    }
    if (strcmp(cur->name, "Mapgroup")) {
	LOG(LOG_FATAL, "%s: root node != Mapgroup", filename);
	xmlFreeDoc(doc);
	bailout();
    }

    path = strdup(filename);
    p = strrchr(path, '/');
    if (p == NULL)
	path = ".";
    else
	*p = '\0';

    mapg = smalloc(mapgroup_t);

    basex = 0.0;
    basez = 0.0;
    baseelev = 0.0;
    mapsize = 256;
    mapg->width = 1;
    mapg->nrofmaps = 1;
    mapspec = NULL;
    init_mutex(&(mapg->mutex));
    pthread_mutex_lock(&mapg->mutex);
   
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "name"))
	    parsestring(doc, cur, &mapg->name);
	if (!strcmp(cur->name, "nrofmaps"))
	    parseint(doc, cur, &mapg->nrofmaps);
	if (!strcmp(cur->name, "width"))
	    parseint(doc, cur, &mapg->width);
	if (!strcmp(cur->name, "mapsize"))
	    parseint(doc, cur, &mapsize);
	if (!strcmp(cur->name, "baseelev"))
	    parsefloat(doc, cur, &baseelev);
	if (!strcmp(cur->name, "mapspec"))
	    parsestring(doc, cur, &mapspec);
	cur = cur->next;
    }
    pthread_mutex_lock(&maps_mutex);
    TAILQ_INSERT_TAIL(&maps, mapg, next);
    pthread_mutex_unlock(&maps_mutex);
    xmlFreeDoc(doc);

    if (mapg->name == NULL || mapspec == NULL) {
	LOG(LOG_FATAL, "Need mapspec and mapname in mapgroup file %s.",
	    filename);
	bailout();
    }

    xa = 0;
    xb = 0;
    p = mapspec;
    i = 0;
    while (*p != 'X') {
	pre[i] = *p;
	p++;
	i++;
    }
    pre[i] = '\0';
    while (*p == 'X') {
	p++;
	xa++;
    }
    i = 0;
    while (*p != 'X') {
	mid[i] = *p;
	p++;
	i++;
    }
    mid[i] = '\0';
    while (*p == 'X') {
	p++;
	xb++;
    }
    i = 0;
    while (*p != '\0') {
	post[i] = *p;
	p++;
	i++;
    }
    post[i] = '\0';
	
    for (i=0; i < mapg->nrofmaps/mapg->width; i++) {
	basez = i * mapsize * SCALE_XZ;
	if (!same_within(basez, 0.0, 10.0))
	    basez -= SCALE_XZ;
	for (j=0; j < mapg->width; j++) {
	    basex = j * mapsize * SCALE_XZ;
	    if (!same_within(basex, 0.0, 10.0))
		basex -= SCALE_XZ;
	    sprintf(mapfile, "%s/map%s%.*d%s%.*d%s.xml", path, pre, xa, i,
		    mid, xb, j, post);
	    parse_mapfile(mapfile, path, i*mapg->width+j, mapg, basex,
			  baseelev, basez, mapsize);
	}
    }
    tile_maps(mapg);
    pthread_mutex_unlock(&mapg->mutex);
    free_string(mapspec);
    free(path);
}

/**
   \brief parse the mapgroups file
   \param filename name of file to parse
   \param path path to mapgroups file
*/

void parse_mapgroups(char *filename, char *path)
{
    xmlDocPtr doc;
    xmlNodePtr cur;
    char file[MAX_PATH], *c;

    sprintf(file, "%s/%s", path, filename);

    doc = xmlParseFile(file);
    if (doc == NULL)
	bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
	LOG(LOG_FATAL, "%s: empty document", file);
	xmlFreeDoc(doc);
	bailout();
    }
    if (strcmp(cur->name, "Mapgroups")) {
	LOG(LOG_FATAL, "%s: root node != Mapgroups", file);
	xmlFreeDoc(doc);
	bailout();
    }

    if (TAILQ_EMPTY(&maps)) {
	pthread_mutex_lock(&maps_mutex);
	TAILQ_INIT(&maps);
	pthread_mutex_unlock(&maps_mutex);
    }

    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "file")) {
	    parsestring(doc, cur, &c);
	    sprintf(file, "%s/%s", path, c);
	    parse_mapgroup(file);
	    free_string(c);
	}
	cur = cur->next;
    }
    xmlFreeDoc(doc);
}
