/* $Id: matrix.c,v 1.18 2004/04/02 02:39:00 garbled Exp $ */
/*
 * Copyright (C) 1997-2003, R3vis Corporation.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA, or visit http://www.gnu.org/copyleft/lgpl.html.
 *
 * Original Contributor:
 *   Wes Bethel, R3vis Corporation, Marin County, California
 * Additional Contributor(s):
 *
 * The OpenRM project is located at http://openrm.sourceforge.net/.
 */
#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <inttypes.h>
#include <math.h>
#include <string.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "object.h"
#include "models.h"
#include "utils.h"
#include "map.h"
#include "matrix.h"
#include "sutils.h"

/**
   \file matrix.c
   \brief Matrix functions, and math functions
   \author Wes Bethel, from rmmatrix.c of OpenRM
*/

/**
   \brief Assert that pointer is not NULL
   \param a pointer to check
   \param s string to write as an error if it is
   \return success/fail
*/

int ASSERT(const void *a, const char *s)
{
    if (a == NULL)
    {
        LOG(LOG_WARNING, s);
        return(0);
    }
    return(1);
}

/**
   \brief Allocate a new matrix
   \return pointer to a new matrix
*/

matrix_t *MatrixNew(void)
{
    matrix_t *m = (matrix_t *)malloc(sizeof(matrix_t));
    MatrixIdentity(m);
    return(m);
}

/**
   \brief copy contents of one matrix to another
   \param d desitnation matrix
   \param s source matrix
   \return success/fail
*/

int MatrixCopy(matrix_t *d, const matrix_t *s)
{
    if ((ASSERT(d, "MatrixCopy() error: the dest matrix_t pointer is NULL.") == 0) ||
	(ASSERT(s, "MatrixCopy() error: the src matrix_t pointer is NULL.") == 0))
	return (0);
    
    memcpy((char *)d, (char *)s, sizeof(matrix_t));
    return(1);
}

/**
   \brief Free a matrix
   \param m matrix to free
   \return success/fail
*/

int MatrixDelete(matrix_t *m)
{
    if (ASSERT(m, "MatrixDelete() error: the input matrix_t pointer is NULL.") == 0)
	return(0);
    
    free(m);
    return(1);
}

/**
   \brief Set this matrix to the identity matrix
   \param m matrix to set
   \return success/fail
*/

int MatrixIdentity(matrix_t *m)
{
    int i, j;

    if (ASSERT(m, "MatrixIdentity() error: the input matrix_t is NULL.") == 0)
	return(0);
    
    for (j = 0; j < 4; j++)
	for (i = 0; i < 4; i++)
	    m->m[i][j] = (i == j) ? 1.0 : 0.0;

    return(1);
}

/**
   \brief transpose a matrix
   \param s input matrix
   \param d destination matrix
   \note safe to call on the same matrix
   \return success/fail
*/

int MatrixTranspose(const matrix_t *s, matrix_t *d)
{
    int      i, j;
    matrix_t t;

    if ((ASSERT(s, "MatrixTranspose() error: the input matrix_t object is NULL") == 0) ||
	(ASSERT(d, "MatrixTranspose() error: the dest matrix_t object is NULL")) == 0)
	return(0);
    
    for (j = 0; j < 4; j++)
      for (i = 0; i < 4; i++)
        t.m[j][i] = s->m[i][j];
    memcpy(d, &t, sizeof(matrix_t));

    return(1);
}

/**
   \brief multiply two matrixes
   \param a input matrix 1
   \param b input matrix 2
   \param c output matrix
   \note Safe to call with aoutput and input being the same matrix
   \return success/fail
   \note take product of a and b, put result in c
*/

int MatrixMultiply(const matrix_t *a, const matrix_t *b, matrix_t *c)
{
    int      row, col;
    matrix_t tmp;

    if ((ASSERT(a, "MatrixMultiply() error: the input A matrix_t is NULL") == 0) ||
	(ASSERT(b, "MatrixMultiply() error: the input B matrix_t is NULL") == 0) ||
	(ASSERT(c, "MatrixMultiply() error: the destination matrix_t is NULL") == 0))
	return(0);

    for (row = 0; row < 4; row++)
    {
	for (col = 0; col < 4; col++)
	    tmp.m[row][col] = (a->m[row][0] * b->m[0][col]) + (a->m[row][1] * b->m[1][col]) + (a->m[row][2] * b->m[2][col]) + (a->m[row][3] * b->m[3][col]);
    }
    for (row = 0; row < 4; row++)
	for (col = 0; col < 4; col++)
	    c->m[row][col] = tmp.m[row][col];

    return(1);
}

#define MATSWAP(a,b) {temp=(a);(a)=(b);(b)=temp;}

/**
   \brief invert a matrix
   \param in input matrix
   \param out output matrix
   \return int succes/fail
*/

int MatrixInverse(const matrix_t *in, matrix_t *out)
{
    matrix_t ident;
    int i, j, k, l, ll;
    int icol=0, irow=0;
    int indxc[4], indxr[4], ipiv[4];
    float big, dum, pivinv, temp;
  
    for (i=0; i<4; i++) {
	for (j=0; j<4; j++) {
	    out->m[i][j] = in->m[i][j];
	    ident.m[i][j] = 0.0f;
	}
	ident.m[i][i]=1.0f;
    } 
    /* Gauss-jordan elimination with full pivoting.  Yes, folks, a 
       GL Matrix4 is inverted like any other, since the identity is 
       still the identity.
   
       from numerical recipies in C second edition, pg 39
    */
 
    for(j=0;j<=3;j++) ipiv[j] = 0;
    for(i=0;i<=3;i++) {
	big=0.0;
	for (j=0;j<=3;j++) {
	    if(ipiv[j] != 1) {
		for (k=0;k<=3;k++) {
		    if(ipiv[k] == 0) {
			if(fabs(out->m[j][k]) >= big) {
			    big = fabsf(out->m[j][k]);
			    irow=j;
			    icol=k;
			}
		    } else if (ipiv[k] > 1) return 1;
		} 
	    }
	}
	++(ipiv[icol]);
	if (irow != icol) {
	    for (l=0;l<=3;l++) MATSWAP(out->m[irow][l],out->m[icol][l]);
	    for (l=0;l<=3;l++) MATSWAP(ident.m[irow][l],ident.m[icol][l]);
	}
	indxr[i]=irow;
	indxc[i]=icol;
	if(out->m[icol][icol] == 0.0f) return 1; 
	pivinv = 1.0f / out->m[icol][icol];
	out->m[icol][icol]=1.0f;
	for (l=0;l<=3;l++) out->m[icol][l] *= pivinv;
	for (l=0;l<=3;l++) ident.m[icol][l] *= pivinv;
	for (ll=0;ll<=3;ll++) {
	    if (ll != icol) {
		dum=out->m[ll][icol];
		out->m[ll][icol]=0.0f;
		for (l=0;l<=3;l++) out->m[ll][l] -= out->m[icol][l]*dum;
		for (l=0;l<=3;l++) ident.m[ll][l] -= ident.m[icol][l]*dum;
	    }
	}
    }
    for (l=3;l>=0;l--) {
	if (indxr[l] != indxc[l]) {
	    for (k=0;k<=3;k++) {
		MATSWAP(out->m[k][indxr[l]],out->m[k][indxc[l]]);
	    }
	}
    }
    return 0;
}

/**
   \brief Print the contents of a matrix to stdout
   \param m matrix to print
   \return success/fail
*/

int PrintMatrix(const matrix_t *m)
{
    int i, j;

    if (ASSERT(m, "PrintMatrix() error: the input matrix_t object is NULL") == 0)
	return(0);
    
    for (j = 0; j < 4; j++)
    {
	for (i = 0; i < 4; i++)
	    printf("\t%g", m->m[j][i]);
	printf("\n");
    }
    return(1);
}

/**
   \brief convert a euler angle to a matrix
   \param euler the euler angle to convert
   \param mat target matrix
*/

void euler_to_matrix(vertex_t euler, matrix_t *mat)
{
    float a,b,c,d,e,f,ad,bd;

    MatrixIdentity(mat);

    a = cos(DEGREES_TO_RADIANS(euler.x));
    b = sin(DEGREES_TO_RADIANS(euler.x));
    c = cos(DEGREES_TO_RADIANS(euler.y));
    d = sin(DEGREES_TO_RADIANS(euler.y));
    e = cos(DEGREES_TO_RADIANS(euler.z));
    f = sin(DEGREES_TO_RADIANS(euler.z));
    ad = a*d;
    bd = b*d;

    mat->m[0][0] = c*e;
    mat->m[0][1] = -c*f;
    mat->m[0][2] = d;
    mat->m[1][0] = bd * e + a * f;
    mat->m[1][1] = -bd * f + a * e;
    mat->m[1][2] = -b * c;
    mat->m[2][0] = -ad * e + b * f;
    mat->m[2][1] = ad * f + b * e;
    mat->m[2][2] = a*c;
}

/**
   \brief convert a euler angle to a quaternion
   \param ax x angle
   \param ay y angle
   \param az z angle
   \param quat4 destination quaternion
   \note angles are in degrees
*/

void euler_to_quat4(float ax, float ay, float az, float quat4[4])
{
    float ex, ey, ez, x, y, z, w;	/* temp half euler angles */
    float cr, cp, cy, sr, sp, sy, cpcy, spsy; /* temp vars in roll,pitch yaw*/
    float norm;
    enum { X = 0, Y = 1, Z = 2, W = 3 };

    /*  convert to rads and half them ey = DEGTORAD(y) / 2.0; */
    ex = DEGREES_TO_RADIANS(ax)/2.0;
    ey = DEGREES_TO_RADIANS(ay)/2.0;
    ez = DEGREES_TO_RADIANS(az)/2.0;

    cr = cos(ex);
    cp = cos(ey);
    cy = cos(ez);

    sr = sin(ex);
    sp = sin(ey);
    sy = sin(ez);

    cpcy = cp * cy;
    spsy = sp * sy;

    x = sr * cpcy - cr * spsy;
    y = cr * sp * cy + sr * cp * sy;
    z = cr * cp * sy - sr * sp * cy;
    w = cr * cpcy + sr * spsy;

    /* Normalize(); Should always == 1 */
    norm = sqrt(x * x + y * y + z * z + w * w);

    quat4[X] = x / norm;
    quat4[Y] = y / norm;
    quat4[Z] = z / norm;
    quat4[W] = w / norm;

    return;
}

/**
   \brief given direction and distance, where do we end up?
   \param direction direction to travel in
   \param motion how far to go
   \param coord where are we now?
   \param newcoord where do we end up?
   \note after calling this routine, you should check the groundheight.

   take a direction, and distance to travel, and return the new coordinates
   in *newcoord. coord is expected to contain the current coordinates

     N=0
 W=270  E=90
    S=180
     
     unfortunately, the shape library gives us no easy way to draw out the
     shapes of each object.  making our lives somewhat annoying.
*/

void new_coord(vertex_t direction, float motion, vertex_t coord,
	       vertex_t *newcoord)
{
    matrix_t new, old;
    vertex_t hold;

    if (motion == 0.0) {
	newcoord->x = 0.0;
	newcoord->y = 0.0;
	newcoord->z = 0.0;
	return;
    }
    
    if (floor(direction.x*10.0) == 2700.0) {
	newcoord->x = 0.0;
	newcoord->y = -motion;
	newcoord->z = 0.0;
	return;
    } else if (floor(direction.x*10.0) == 900.0) {
	newcoord->x = 0.0;
	newcoord->y = motion;
	newcoord->z = 0.0;
	return;
    } else {
	MatrixIdentity(&new);
	prep_ymat(&old, add_euler(direction.y, -45.0));

	new.m[0][3] = -motion/1.414214;
	new.m[1][3] = -motion/1.414214;
	new.m[2][3] = motion/1.414214;
	MatrixMultiply(&old, &new, &old);
	/* PrintMatrix(&old); */

	hold.y = old.m[1][3];
    }

    if (direction.x != 0.0) {
	hold.x = old.m[0][3];
	hold.z = old.m[2][3];
	MatrixIdentity(&new);
	prep_xmat(&old, add_euler(direction.x, 45.0));

	new.m[0][3] = hold.x;
	new.m[1][3] = hold.y;
	new.m[2][3] = hold.z;

	MatrixMultiply(&old, &new, &old);
	/*PrintMatrix(&old); */
    }
    /*printf("direction.y = %f dir.x = %f\n", add_euler(direction.y, -45.0), direction.x);*/
    newcoord->x = -old.m[0][3];
    newcoord->y = hold.y;
    newcoord->z = old.m[2][3];
    /*printf("NEW: %f %f %f\n", newcoord->x, newcoord->y, newcoord->z);*/
}

/**
   \brief given an angle, prepare an x rotation matrix
   \param m matrix to prepare
   \param dir angle
*/

void prep_xmat(matrix_t *m, float dir)
{
    float d,c,s;

    MatrixIdentity(m);
    d = DEGREES_TO_RADIANS(dir);
    c = cos(d);
    s = sin(d);
    m->m[1][1] = c;
    m->m[1][2] = s;
    m->m[2][1] = -s;
    m->m[2][2] = c;
}

/**
   \brief given an angle, prepare a y rotation matrix
   \param m matrix to prepare
   \param dir angle
*/

void prep_ymat(matrix_t *m, float dir)
{
    float d,c,s;

    MatrixIdentity(m);
    d = DEGREES_TO_RADIANS(dir);
    c = cos(d);
    s = sin(d);
    m->m[0][0] = c;
    m->m[2][2] = c;
    m->m[0][2] = -s;
    m->m[2][0] = s;
}

/**
   \brief given an angle, prepare a z rotation matrix
   \param m matrix to prepare
   \param dir angle
*/

void prep_zmat(matrix_t *m, float dir)
{
    float d,c,s;

    MatrixIdentity(m);
    d = DEGREES_TO_RADIANS(dir);
    c = cos(d);
    s = sin(d);
    m->m[0][0] = c;
    m->m[0][1] = s;
    m->m[1][0] = -s;
    m->m[1][1] = c;
}

/* Bounding BOX:
      f      r
     1 0    5 4
     3 2    7 6
*/

/**
   \brief Rotate a point coord about the true center
   \param coord coordinate of box
   \param center center to rotate about
   \param dir euler to rotate in
   \param new new coordinates of point
*/

void rotate_point_on_center(vertex_t coord, vertex_t center, vertex_t dir,
			    vertex_t *new)
{
    matrix_t m1, m2;

    new->x = coord.x;
    new->y = coord.y;
    new->z = coord.z;

    if (dir.x != 0.0) {
	prep_xmat(&m2, dir.x);
	MatrixIdentity(&m1);
	m1.m[3][0] = coord.x-center.x;
	m1.m[3][1] = coord.y-center.y;
	m1.m[3][2] = coord.z-center.z;
	MatrixMultiply(&m1, &m2, &m1);
	new->x = m1.m[3][0]+center.x;
	new->y = m1.m[3][1]+center.y;
	new->z = m1.m[3][2]+center.z;
    }
    if (dir.y != 0.0) {
	prep_ymat(&m2, dir.y);
	MatrixIdentity(&m1);
	m1.m[3][0] = coord.x-center.x;
	m1.m[3][1] = coord.y-center.y;
	m1.m[3][2] = coord.z-center.z;
	MatrixMultiply(&m1, &m2, &m1);
	new->x = m1.m[3][0]+center.x;
	new->y = m1.m[3][1]+center.y;
	new->z = m1.m[3][2]+center.z;
    }
    if (dir.z != 0.0) {
	prep_zmat(&m2, dir.z);
	MatrixIdentity(&m1);
	m1.m[3][0] = coord.x-center.x;
	m1.m[3][1] = coord.y-center.y;
	m1.m[3][2] = coord.z-center.z;
	MatrixMultiply(&m1, &m2, &m1);
	new->x = m1.m[3][0]+center.x;
	new->y = m1.m[3][1]+center.y;
	new->z = m1.m[3][2]+center.z;
    }
    
}

/**
   \brief rotate a bounding box by the euler dir, about it's true center
   \param box bounding box to rotate
   \param dir euler to rotate in
   \param origin center to rotate about
*/

void rotate_bbox(bbox_t *box, vertex_t dir, vertex_t origin)
{
    int i;
    vertex_t newo;

    rotate_point_on_center(box->center, origin, dir, &newo);
    box->center = newo;
    rotate_point_on_center(box->cog, origin, dir, &newo);
    box->cog = newo;

    for (i=0; i<8; i++) {
	rotate_point_on_center(box->side[i], origin, dir, &newo);
	box->side[i] = newo;
    }
}

/**
   \brief find the lowest point on the bbox
   \return an int (the corner number)
   \param box box to examine
*/

int find_lowest_bbox_point(bbox_t *box)
{
    int i, j;

    for (i=0,j=0; i<8; i++) {
	if (box->side[i].y < box->side[j].y)
	    j = i;
    }
    return(j);
}

/**
   \brief find the bottom of a bounding box
   \param box box to find bottom of
   \param maph maphead where box is located
   \param side four points that make up the bottom
   \return the number of sides that are grounded
   find the "bottom" of the bounding box.  this is the side that is in contact
   with the ground, or close to it
*/

int find_lowest_bbox_side(bbox_t *box, maphead_t *maph, int side[4])
{
    int i, n, x;
    float lp, gdiff[8];

    for (i=0; i<8; i++) {
	gdiff[i] = box->side[i].y;
	gdiff[i] -=groundheight_at_mapxz(maph, box->side[i].x, box->side[i].z);
	gdiff[i] = fabs(gdiff[i]);
    }

    for (n=0, x=0; n<4; n++) {
	lp = 999999.0;
	for (i=0; i<8; i++) {
	    if (n > 0 && i == side[0])
		continue;
	    if (n > 1 && i == side[1])
		continue;
	    if (n > 2 && i == side[2])
		continue;
	    if (lp > gdiff[i]) {
		lp = gdiff[i];
		side[n] = i;
	    }
	}
	if (floor(gdiff[i] * 11.0) == 0.0)
	    x++;
    }
    /* we return the number of sides that are grounded.  giving us a
       stability factor */
    return(x);
}

/**
   \brief find the lowest point in a bbox and return it's groundheight diff
   \param box box to examine
   \param maph maphead to examine box on
   \param coord coordinates of object box represents
   \return groundheight difference of lowest box point, or coord.y whichever
   is lowest
*/

float bbox_groundheight_diff(bbox_t *box, maphead_t *maph, vertex_t coord)
{
    int i;
    float ldiff, lp;

    ldiff = coord.y - groundheight_at_mapxz(maph, coord.x, coord.z);

    for (i=0; i < 8; i++) {
	lp = box->side[i].y -
	    groundheight_at_mapxz(maph, box->side[i].x, box->side[i].z);
	if (lp < ldiff)
	    ldiff = lp;
    }
    return(ldiff);
}
