/* $Id: models.c,v 1.22 2004/04/01 07:29:54 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <math.h>
#include <time.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "object.h"
#include "map.h"
#include "shstr.h"
#include "utils.h"
#include "models.h"
#include "matrix.h"
#include "xmlparse.h"

/**
   \file models.c
   \brief Routines to parse model files for the collision engine
   \author Tim Rightnour

   This file only parses as much as we need.  The point is to get enough to
   collide off of realistically, not build a perfect picture.
*/

#if 0
/* parse a cal mesh file */
/* this is unbelievably slow for large meshes.  specifically, the paladin
   body is 800k, it just drags the game to a crawl.  In addition, the memory
   usage is through the roof.  COmmenting out for perhaps later use
*/

void parse_calmesh(char *filename, object_t *obj)
{
    xmlDocPtr doc;
    xmlNodePtr cur, ncur, xcur;
    float *v;
    int *coll;
    int x, y, z;
    char *tmp, *filedata;
    FILE *file;
    size_t fsize;

    /* this abhorrent little dance below, is because the CAL mesh models are
       not properly formed XML.  The HEADER bit at the top wrecks them. This
       silliness is to skip over the first line.  grrrrr */
    /* XXXXX THIS MEANS CALXML MODELS CANNOT BE STORED COMPRESSED ON SERVER */

    file = fopen(filename, "r");
    if (file == NULL)
	bailout();
    if (fseek(file, 0, SEEK_END))
	bailout();
    fsize = ftell(file);
    if (fsize == 0)
	bailout();
    filedata = safer_malloc(sizeof(char)*fsize, __LINE__);
    if (fseek(file, 0, SEEK_SET))
	bailout();
    tmp = safer_malloc(sizeof(char)*128, __LINE__);
    fgets(tmp, 128, file);
    if (fseek(file, strlen(tmp), SEEK_SET))
	bailout();
    fsize -= strlen(tmp);
    if (fsize < 0)
	bailout();
    free(tmp);
    if (fread(filedata, fsize, 1, file) != 1)
	bailout();
    fclose(file);

    /* end stupidity.  If the cal people ever fix thier model format, we can
       get rid of this */

    doc = xmlParseMemory(filedata, fsize);
    if (doc == NULL)
        bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        fprintf(stderr, "empty document\n");
        xmlFreeDoc(doc);
        bailout();
    }
    if (strcmp(cur->name, "MESH")) {
        fprintf(stderr, 
                "document of the wrong type, root node != MESH");
        xmlFreeDoc(doc);
        bailout();
    }

    pthread_mutex_lock(&obj->mutex);
    cur = cur->xmlChildrenNode;

    while (cur != NULL) {
 	if (!strcmp(cur->name, "SUBMESH")) {
	    int verts, cvert;

	    printf("parsing submesh\n");
	    tmp = xmlGetProp(cur, "NUMVERTICES");
	    verts = atoi(tmp);
	    free(tmp);
	    v = safer_malloc(sizeof(float)*verts*3, __LINE__);
	    coll = malloc(sizeof(int)*verts);
	    if (coll == NULL)
		bailout();
	    memset(coll, -1, sizeof(int)*verts);
	    cvert = 0;
	    ncur = cur->xmlChildrenNode;
	    while (ncur != NULL) {
		if (!strcmp(ncur->name, "VERTEX")) {
		    xcur = ncur->xmlChildrenNode;
		    while (xcur != NULL) {
			if (!strcmp(xcur->name, "POS")) {
			    parsetrifloat(doc, xcur, &v[cvert], &v[cvert+1],
					  &v[cvert+2]);
			    cvert+=3;
			}
			if (!strcmp(xcur->name, "COLLAPSEID")) {
			    parseint(doc, xcur, &coll[cvert/3]);
			}
			xcur = xcur->next;
		    } /* xcur */
		}
		if (!strcmp(ncur->name, "FACE")) {
		    tmp = xmlGetProp(ncur, "VERTEXID");
		    sscanf(tmp, "%d %d %d", &x, &y, &z);
		    free(tmp);
		    if (coll[x] != -1)
			x = coll[x];
		    if (coll[y] != -1)
			y = coll[y];
		    if (coll[z] != -1)
			z = coll[z];
		    dtBegin(DT_SIMPLEX);
		    dtVertexBase(v);
		    dtVertexIndex(x);
		    dtVertexIndex(y);
		    dtVertexIndex(z);
		    dtEnd();
		}
		ncur = ncur->next;
	    } /* ncur */
	    free(v);
	    free(coll);
	} /* SUBMESH */
	cur = cur->next;
    }/* cur */
    free(filedata);
    xmlFreeDoc(doc);
}
#endif

/**
   \brief parse a cal model xcm file
   \param filename full path filename to parse
   \param obj to set the shape for
   \note should mostly be called for arch objects

   parse a calmodel file, sort of. We approximate the shape by using a
   bounding box for each submesh.  This gives us decent resolution, without
   using up 6MB of ram per paladin model.
*/

void parse_calmodel(char *filename, object_t *obj)
{
    xmlDocPtr doc;
    xmlNodePtr cur, ncur;
#if 0
    char meshfile[1024], *tmp, *path=NULL;
#endif
    float pts[6], modscale, twist=0.0, tilt=0.0;
    int i;
    matrix_t m1, m2;

    doc = xmlParseFile(filename);
    if (doc == NULL)
        bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        fprintf(stderr, "empty document\n");
        xmlFreeDoc(doc);
        bailout();
    }
    if (strcmp(cur->name, "calmodel")) {
        fprintf(stderr, 
                "document of the wrong type, root node != calmodel");
        xmlFreeDoc(doc);
        bailout();
    }

    obj->shape = dtNewComplexShape();
    cur = cur->xmlChildrenNode;
    modscale = 1.0;
    obj->bmin.x = obj->bmin.y = obj->bmin.z = 999999.0;
    obj->bmax.x = obj->bmax.y = obj->bmax.z = -999999.0;

    while (cur != NULL) {
#if 0
	if (!strcmp(cur->name, "path")) {
	    path = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	}
	if (!strcmp(cur->name, "mesh")) {
	    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	    if (path == NULL)
		bailout();
	    sprintf(meshfile, "data/arch/%s%s", path, tmp);
	    free(tmp);
	    parse_calmesh(meshfile, obj);
	}
#endif
	if (!strcmp(cur->name, "scale"))
	    parsefloat(doc, cur, &modscale);
	if (!strcmp(cur->name, "tiltangle"))
	    parsefloat(doc, cur, &tilt);
	if (!strcmp(cur->name, "twistangle"))
	    parsefloat(doc, cur, &twist);
	if (!strcmp(cur->name, "bboxes")) {
	    ncur = cur->xmlChildrenNode;
	    while (ncur != NULL) {
		if (!strcmp(ncur->name, "min")) {
		    parsetrifloat(doc, ncur, &pts[0], &pts[1], &pts[2]);
		    if (tilt != 0.0) {
			prep_xmat(&m2, tilt);
			MatrixIdentity(&m1);
			m1.m[3][0] = pts[0];
			m1.m[3][1] = pts[1];
			m1.m[3][2] = pts[2];
			MatrixMultiply(&m1, &m2, &m1);
			pts[0] = m1.m[3][0];
			pts[1] = m1.m[3][1];
			pts[2] = m1.m[3][2];
		    }
		    /*twist += 45.0;*/
		    if (twist != 0.0) {
			prep_zmat(&m2, twist+90.0);
			MatrixIdentity(&m1);
			m1.m[3][0] = pts[0];
			m1.m[3][1] = pts[1];
			m1.m[3][2] = pts[2];
			MatrixMultiply(&m1, &m2, &m1);
			pts[0] = m1.m[3][0];
			pts[1] = m1.m[3][1];
			pts[2] = m1.m[3][2];
		    }
		}
		if (!strcmp(ncur->name, "max")) {
		    int j, k, l, v[8][3], m=0;
		    parsetrifloat(doc, ncur, &pts[3], &pts[4], &pts[5]);
		    if (tilt != 0.0) {
			prep_xmat(&m2, tilt);
			MatrixIdentity(&m1);
			m1.m[3][0] = pts[3];
			m1.m[3][1] = pts[4];
			m1.m[3][2] = pts[5];
			MatrixMultiply(&m1, &m2, &m1);
			pts[3] = m1.m[3][0];
			pts[4] = m1.m[3][1];
			pts[5] = m1.m[3][2];
		    }
		    if (twist != 0.0) {
			prep_zmat(&m2, twist+90.0);
			MatrixIdentity(&m1);
			m1.m[3][0] = pts[3];
			m1.m[3][1] = pts[4];
			m1.m[3][2] = pts[5];
			MatrixMultiply(&m1, &m2, &m1);
			pts[3] = m1.m[3][0];
			pts[4] = m1.m[3][1];
			pts[5] = m1.m[3][2];
		    }
		    for (i=0; i<6; i++)
			pts[i] = pts[i] * modscale * 0.5;
		    if (pts[0] > obj->bmax.x)
			obj->bmax.x = pts[0];
		    if (pts[3] > obj->bmax.x)
			obj->bmax.x = pts[3];
		    if (pts[0] < obj->bmin.x)
			obj->bmin.x = pts[0];
		    if (pts[3] < obj->bmin.x)
			obj->bmin.x = pts[3];
		    if (pts[1] > obj->bmax.y)
			obj->bmax.y = pts[1];
		    if (pts[4] > obj->bmax.y)
			obj->bmax.y = pts[4];
		    if (pts[1] < obj->bmin.y)
			obj->bmin.y = pts[1];
		    if (pts[4] < obj->bmin.y)
			obj->bmin.y = pts[4];
		    if (pts[2] > obj->bmax.z)
			obj->bmax.z = pts[2];
		    if (pts[5] > obj->bmax.z)
			obj->bmax.z = pts[5];
		    if (pts[2] < obj->bmin.z)
			obj->bmin.z = pts[2];
		    if (pts[5] < obj->bmin.z)
			obj->bmin.z = pts[5];

		    for (j=0; j<4; j+=3) {
			for (k=1; k<5; k+=3) {
			    for (l=2; l<6; l+=3) {
				v[m][0] = j;
				v[m][1] = k;
				v[m][2] = l;
				m++;
			    }
			}
		    }
		    for (j=0; j<8; j+=7) {
			dtBegin(DT_SIMPLEX);
			dtVertex(pts[v[j][0]], pts[v[j][1]], pts[v[j][2]]);
			dtVertex(pts[v[2][0]], pts[v[2][1]], pts[v[2][2]]);
			dtVertex(pts[v[6][0]], pts[v[6][1]], pts[v[6][2]]);
			dtEnd();
			dtBegin(DT_SIMPLEX);
			dtVertex(pts[v[j][0]], pts[v[j][1]], pts[v[j][2]]);
			dtVertex(pts[v[2][0]], pts[v[2][1]], pts[v[2][2]]);
			dtVertex(pts[v[3][0]], pts[v[3][1]], pts[v[3][2]]);
			dtEnd();
			dtBegin(DT_SIMPLEX);
			dtVertex(pts[v[j][0]], pts[v[j][1]], pts[v[j][2]]);
			dtVertex(pts[v[1][0]], pts[v[1][1]], pts[v[1][2]]);
			dtVertex(pts[v[5][0]], pts[v[5][1]], pts[v[5][2]]);
			dtEnd();
			dtBegin(DT_SIMPLEX);
			dtVertex(pts[v[j][0]], pts[v[j][1]], pts[v[j][2]]);
			dtVertex(pts[v[4][0]], pts[v[4][1]], pts[v[4][2]]);
			dtVertex(pts[v[6][0]], pts[v[6][1]], pts[v[6][2]]);
			dtEnd();
			dtBegin(DT_SIMPLEX);
			dtVertex(pts[v[j][0]], pts[v[j][1]], pts[v[j][2]]);
			dtVertex(pts[v[4][0]], pts[v[4][1]], pts[v[4][2]]);
			dtVertex(pts[v[5][0]], pts[v[5][1]], pts[v[5][2]]);
			dtEnd();
			dtBegin(DT_SIMPLEX);
			dtVertex(pts[v[j][0]], pts[v[j][1]], pts[v[j][2]]);
			dtVertex(pts[v[1][0]], pts[v[1][1]], pts[v[1][2]]);
			dtVertex(pts[v[3][0]], pts[v[3][1]], pts[v[3][2]]);
			dtEnd();
		    }

/*
		    dtBegin(DT_POLYGON);
		    dtVertex(pts[0], pts[1], pts[2]);
		    dtVertex(pts[0], pts[1], pts[5]);
		    dtVertex(pts[0], pts[4], pts[2]);
		    dtVertex(pts[0], pts[4], pts[5]);
		    dtVertex(pts[3], pts[1], pts[2]);
		    dtVertex(pts[3], pts[1], pts[5]);
		    dtVertex(pts[3], pts[4], pts[2]);
		    dtVertex(pts[3], pts[4], pts[5]);
		    dtEnd();
*/
		}
		ncur = ncur->next;
	    }
	}
	/* tilt and twist will eventually be required.  Come back here later */
	cur = cur->next;
    }
#if 0
    free(path);
#endif
    dtEndComplexShape();
    obj->bmax.x *= obj->cli.scale.x;
    obj->bmax.y *= obj->cli.scale.y;
    obj->bmax.z *= obj->cli.scale.z;
    obj->bmin.x *= obj->cli.scale.x;
    obj->bmin.y *= obj->cli.scale.y;
    obj->bmin.z *= obj->cli.scale.z;
    obj->coff.x = (obj->bmax.x+obj->bmin.x)/2.0;
    obj->coff.y = (obj->bmax.y+obj->bmin.y)/2.0;
    obj->coff.z = (obj->bmax.z+obj->bmin.z)/2.0;
    obj->center = obj->coff;
    xmlFreeDoc(doc);
}

/**
   \brief parse only the bounding box of an xgm model
   \param filename full path filename to parse
   \param obj object to fill data in for
*/

void parse_genmodel_boxonly(char *filename, object_t *obj)
{
    xmlDocPtr doc;
    xmlNodePtr cur;
    int cf=0;

    doc = xmlParseFile(filename);
    if (doc == NULL)
        bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        LOG(LOG_FATAL, "%s: empty document", filename);
        xmlFreeDoc(doc);
        bailout();
    }
    if (strcmp(cur->name, "genmodel")) {
        LOG(LOG_FATAL, "%s: root node != genmodel", filename);
        xmlFreeDoc(doc);
        bailout();
    }

    cur = cur->xmlChildrenNode;

    while (cur != NULL) {
	if (!strcmp(cur->name, "bmin")) {
	    parsetrifloat(doc, cur, &obj->bmin.x, &obj->bmin.y, &obj->bmin.z);
	}
	if (!strcmp(cur->name, "bmax")) {
	    parsetrifloat(doc, cur, &obj->bmax.x, &obj->bmax.y, &obj->bmax.z);
	}
	if (!strcmp(cur->name, "cog")) {
	    parsetrifloat(doc, cur, &obj->center.x, &obj->center.y,
			   &obj->center.z);
	    cf++;
	}
	cur = cur->next;
    }
    obj->coff.x = (obj->bmax.x+obj->bmin.x)/2.0;
    obj->coff.z = (obj->bmax.z+obj->bmin.z)/2.0;
    if (cf == 0) {
	obj->center.x = (obj->bmax.x+obj->bmin.x)/2.0;
	obj->center.y = (obj->bmax.y+obj->bmin.y)/2.0;
	obj->center.z = (obj->bmax.z+obj->bmin.z)/2.0;
    }
    xmlFreeDoc(doc);
}


/**
  \brief Parse a genmodel file (xgm)
  \param filename full path filename to parse
  \param obj object to fill in shape data for.  Should be called on arch objs
  \todo make the parser skip meshes marked optional
  \todo make the parser operate on the lowest level of detail if available
*/

void parse_genmodel(char *filename, object_t *obj)
{
    xmlDocPtr doc;
    xmlNodePtr cur, ncur, xcur, ycur;
    float *vx=NULL, *vy=NULL, *vz=NULL;
    int nrofpoints, stripcur=0, x, y, z, cf=0, done=0;
    char *tmp;

    doc = xmlParseFile(filename);
    if (doc == NULL)
        bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        LOG(LOG_FATAL, "%s: empty document", filename);
        xmlFreeDoc(doc);
        bailout();
    }
    if (strcmp(cur->name, "genmodel")) {
	LOG(LOG_FATAL, "%s: root node != genmodel", filename);
        xmlFreeDoc(doc);
        bailout();
    }

    obj->shape = dtNewComplexShape();
    cur = cur->xmlChildrenNode;

    while (cur != NULL && !done) {
	if (!strcmp(cur->name, "bmin")) {
	    parsetrifloat(doc, cur, &obj->bmin.x, &obj->bmin.y, &obj->bmin.z);
	}
	if (!strcmp(cur->name, "bmax")) {
	    parsetrifloat(doc, cur, &obj->bmax.x, &obj->bmax.y, &obj->bmax.z);
	}
	if (!strcmp(cur->name, "cog")) {
	    parsetrifloat(doc, cur, &obj->center.x, &obj->center.y,
			   &obj->center.z);
	    cf++;
	}
	if (!strcmp(cur->name, "solid")) {
	    ncur = cur->xmlChildrenNode;
	    while (ncur != NULL) {
		if (!strcmp(ncur->name, "boxonly")) {
		        dtBegin(DT_POLYHEDRON);
			dtVertex(obj->bmin.x,obj->bmin.y,obj->bmin.z);
			dtVertex(obj->bmax.x,obj->bmin.y,obj->bmin.z);
			dtVertex(obj->bmax.x,obj->bmax.y,obj->bmin.z);
			dtVertex(obj->bmin.x,obj->bmax.y,obj->bmin.z);
			dtVertex(obj->bmin.x,obj->bmin.y,obj->bmax.z);
			dtVertex(obj->bmax.x,obj->bmin.y,obj->bmax.z);
			dtVertex(obj->bmax.x,obj->bmax.y,obj->bmax.z);
			dtVertex(obj->bmin.x,obj->bmax.y,obj->bmax.z);
			dtEnd();
			done++;
		}
		ncur = ncur->next;
	    }
	}
 	if (!strcmp(cur->name, "surface")) {
	    ncur = cur->xmlChildrenNode;
	    while (ncur != NULL) {
		if (!strcmp(ncur->name, "vertexs")) {
		    xcur = ncur->xmlChildrenNode;
		    while (xcur != NULL) {
			int nrofverts;
			static int cvert;

			if (!strcmp(xcur->name, "verts")) {
			    parseint(doc, xcur, &nrofverts);
			    vx = safer_malloc(sizeof(float) * nrofverts);
			    vy = safer_malloc(sizeof(float) * nrofverts);
			    vz = safer_malloc(sizeof(float) * nrofverts);
			    cvert = 0;
			}
			if (!strcmp(xcur->name, "vertex")) {
			    ycur = xcur->xmlChildrenNode;
			    while (ycur != NULL) {
				if (!strcmp(ycur->name, "vertex_x"))
				    parsefloat(doc, ycur, &vx[cvert]);
				if (!strcmp(ycur->name, "vertex_y"))
				    parsefloat(doc, ycur, &vy[cvert]);
				if (!strcmp(ycur->name, "vertex_z"))
				    parsefloat(doc, ycur, &vz[cvert]);
				ycur = ycur->next;
			    }
			    cvert++;
			}
			xcur = xcur->next;
		    }
		}
		if (!strcmp(ncur->name, "triangle")) {
		    int type=0, vnum;

		    xcur = ncur->xmlChildrenNode;
		    while (xcur != NULL) {
			if (!strcmp(xcur->name, "points")) {
			    parseint(doc, xcur, &nrofpoints);
			}
			if (!strcmp(xcur->name, "type")) {
			    tmp = xmlNodeListGetString(doc,
                                xcur->xmlChildrenNode, 1);
			    if (!strcmp(tmp, "strip")) {
				type = M_TRI_STRIP;
				/*dtVertexBase(v);*/
				stripcur = 0;
			    }
			    else if (!strcmp(tmp, "block")) {
				type = M_TRI_BLOCK;
				/*dtVertexBase(v);*/
			    }
			    free(tmp);
			}
			if (!strcmp(xcur->name, "face")) {
			    if (type == M_TRI_STRIP) {
				parseint(doc, xcur, &vnum);
				stripcur++;
				if (stripcur > 3)
				    stripcur = 1;
				if (stripcur == 1)
				    dtBegin(DT_SIMPLEX);
				dtVertex(vx[vnum], vy[vnum], vz[vnum]);
				if (stripcur == 3)
				    dtEnd();
			    } else if (type == M_TRI_BLOCK) {
				parsetriint(doc, xcur, &x, &y, &z);
				dtBegin(DT_SIMPLEX);
				/*dtVertexIndex(x);
				dtVertexIndex(y);
				dtVertexIndex(z);*/
				dtVertex(vx[x], vy[x], vz[x]);
				dtVertex(vx[y], vy[y], vz[y]);
				dtVertex(vx[z], vy[z], vz[z]);
				dtEnd();
			    }
			}
			xcur = xcur->next;
		    } /* while triangle */
		    if (stripcur < 3)
			dtEnd();
		} /* triangle */
		ncur = ncur->next;
	    } /* while */
	} /* surface */
	free(vx);
	free(vy);
	free(vz);
	cur = cur->next;
    }
    obj->coff.x = (obj->bmax.x+obj->bmin.x)/2.0;
    /*obj->coff.y = (obj->bmax.y+obj->bmin.y)/2.0;*/
    obj->coff.z = (obj->bmax.z+obj->bmin.z)/2.0;
    if (cf == 0) {
	obj->center.x = (obj->bmax.x+obj->bmin.x)/2.0;
	obj->center.y = (obj->bmax.y+obj->bmin.y)/2.0;
	obj->center.z = (obj->bmax.z+obj->bmin.z)/2.0;
    }
    dtEndComplexShape();
    xmlFreeDoc(doc);

    obj->coff.x = 0;/*(obj->bmax.x+obj->bmin.x)/2.0;*/
    obj->coff.z = 0;/*(obj->bmax.z+obj->bmin.z)/2.0; */
}
