/* $Id: network.c,v 1.31 2004/04/02 02:39:00 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 1998-2002 David Watson and Able Data Systems
    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#if defined(__linux__)
#include "queue.h"
#include <sys/time.h>
#else
#include <sys/queue.h>
#endif

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "shstr.h"
#include "object.h"
#include "player.h"
#include "network.h"
#include "utils.h"
#include "timers.h"
#include "settings.h"
#include "netparser.h"
#include "commands.h"
#include "ninterface.h"
#include "servdata.h"
#include "sutils.h"

/**
   \file network.c
   \brief the main networking functions for the server
   \author Dave Watson and Tim Rightnour
*/

extern persist_data_t persist;
extern npersist_data_t nonpersist;
extern pthread_mutex_t dlist_mutex;

TAILQ_HEAD(, _descriptor_data_t) descriptor_list =
    TAILQ_HEAD_INITIALIZER(descriptor_list);
typedef void (*parsefunc_t)(descriptor_data_t *d, pargs_t *args);

parsefunc_t parsefunc[] = {
    parse_command_null,
    parse_command_adpl,
    parse_command_modpl,
    parse_command_rot,
    parse_command_null
};

extern settings_t settings;

/**
   \brief Create the listener socket
   \param port portnumber to listen on
   \return -1 for failure to create, or socket number
*/

int init_socket(int port)
{
    int s, sbuf, opt;
    struct sockaddr_in sa;
    struct linger ld;

    memset(&sa, 0, sizeof(struct sockaddr_in));
    sa.sin_family = AF_INET;
    sa.sin_port = htons(port);

    if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	perror("socket()");
	return(-1);
    }

    opt=1;

    /* Reuse address */
    if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
		   (char *)&opt, sizeof(opt)) < 0) {
	perror("setsockopt() Reuse address");
	return(-1);
    }

#ifdef NO_NAGLE
    /*
     * TCP_NODELAY: Turns off nagle algorithm if NO_NAGLE is defined.
     * The Nagle algorithm delays sending partial packets in hopes of
     * getting more data. The server will wait up to 200ms per packet
     * sent for a response from the client.
     *
     * If this is a server that sends small packets of data, or is a
     * server that has short lived connections, then the Nagle algorithm
     * will be undesirable as it will cause performance problems.
     * TCP_NODELAY should be enabled by defining NO_NAGLE.
     *
     * If this is a server that receives a lot of data from the client
     * and/or sends large packets of data, then the Nagle algorithm may
     * give you some performance gains. TCP_NODELAY should be disabled by
     * undefining NO_NAGLE.
     */

    if (setsockopt(s, IPPROTO_TCP, TCP_NODELAY,
		   (char *)&opt, sizeof(opt)) < 0) {
	perror("setsockopt() TCP_NODELAY");
	return(-1);
    }
#endif /* #ifdef NO_NAGLE */

    /* Linger */
    ld.l_onoff = 0;
    ld.l_linger = 1000; /* We don't care what this is */
    if (setsockopt(s, SOL_SOCKET, SO_LINGER, (char *)&ld, sizeof(ld)) < 0) {
	perror("setsockopt() Linger");
	return(-1);
    }

    /* 8K Send buffer */
    sbuf = 8192;

    if (setsockopt(s, SOL_SOCKET, SO_SNDBUF,
		   (char *)&sbuf, sizeof(sbuf)) < 0) {
	perror("setsockopt() Send buffer");
	return(-1);
    }

    /* Bind it */
    if (bind(s, (struct sockaddr *)&sa, sizeof(sa)) < 0) {
	perror("bind()");
	close(s);
	return(-1);
    }

    /* Set connection queue */
    if (listen(s, 5) < 0) {
	perror("listen()");
	return(-1);
    }

    return(s);
}

/**
   \brief Create a new descriptor_data structure
   \param s socket to initialize with
   \return 0
*/

int new_desc(int s)
{
    int size;
    struct sockaddr_in sock;
    struct hostent *from;
    descriptor_data_t *newd;
    char buf[256];

    newd = (descriptor_data_t *)smalloc(descriptor_data_t);
    init_mutex(&(newd->mutex));
    pthread_mutex_lock(&newd->mutex);
    /* Get host information of connecting host */
    size = sizeof(sock);
    if (getpeername(s, (struct sockaddr *) &sock, &size) < 0) {
	perror("getpeername()");
	*newd->host = '\0';
    }

    else if (!(from = gethostbyaddr((void *)&sock.sin_addr,
				    sizeof(struct in_addr), AF_INET))) {
#ifndef USE_INET_NTOA
	addr = sock.sin_addr.s_addr;
	sprintf(newd->host, "%d.%d.%d.%d",
		(addr & 0xFF000000) >> 24,
		(addr & 0x00FF0000) >> 16,
		(addr & 0x0000FF00) >> 8,
		(addr & 0x000000FF));
#else /* #ifndef USE_INET_NTOA */
	strcpy(newd->host, (char *)inet_ntoa(sock.sin_addr));
#endif /* #ifndef USE_INET_NTOA */
    } else {
	strncpy(newd->host, from->h_name, 49);
	*(newd->host + 49) = '\0';
    }

    /* Initialize values for descriptor_data structure */
    newd->descriptor = s;
    newd->idle = time(0);
    newd->in_buf = 0;
    newd->out_buf = 0;
    newd->outbuflen = 0;

    /* Add the structure to the list */
    pthread_mutex_lock(&dlist_mutex);
    TAILQ_INSERT_TAIL(&descriptor_list, newd, next);
    pthread_mutex_unlock(&dlist_mutex);
    pthread_mutex_unlock(&newd->mutex);


    /* Send out a welcome to the new connection */
    sprintf(buf, "ready\r"/*, newd->host*/);
    pthread_mutex_lock(&newd->mutex);
    do_write(newd, buf);
    pthread_mutex_unlock(&newd->mutex);
    LOG(LOG_DEBUG, "New connection: Desc %d, Host %s", newd->descriptor,
	newd->host);
    return(0);
}

/*
  \brief Create a new connection
  \param s socket to use
  \return -1 for busy, 0 for failure or descriptor
*/

int init_desc(int s)
{
    struct sockaddr_in isa;
    int i;
    int t;
    char buf[256];

    i = sizeof(isa);
    getsockname(s, (struct sockaddr *)&isa, &i);

    if ((t = accept(s, (struct sockaddr *)&isa, &i)) < 0) {
	perror("accept()");
	return(-1);
    }

#ifdef NONBLOCK
    /*
     * Turn on No Delay (or non-blocking) for descriptor. This is almost
     * always necessary in a multiplexing type server. If it is not
     * set, then the process will sleep while waiting for data on a
     * read() call. This causes the whole process to stop all operations
     * until it receives data. Undefine NONBLOCK if you want non-blocking
     * disabled
     */
    if (fcntl(t, F_SETFL, O_NDELAY) < 0) {
	perror("fcntl() non-block");
	return(-1);
    }
#endif /* #ifdef NONBLOCK */

    if (nonpersist.maxdesc + 1 >= nonpersist.avail_descs ||
	nonpersist.numusers + 1 > settings.maxusers) {
	strcpy(buf, "srvfull\r");
	write(t, buf, strlen(buf));
	close(t);
	return(0);
    }

    if (t > nonpersist.maxdesc)
	nonpersist.maxdesc = t;

    new_desc(t);
    return(t);
}

/**
   \brief Close a descriptor
   \param d descriptor to close
   \note d should be locked
*/

void close_desc(int d)
{
    close(d);

    if (d == nonpersist.maxdesc)
	--nonpersist.maxdesc;

    LOG(LOG_DEBUG, "Closed connection: Desc %d\n", d);
}

/**
   \brief Free a descriptor.
   \param d descriptor to free
   \note d should be locked.  Assumes dlist is locked
*/
void free_desc(descriptor_data_t *d)
{
    event_t *event;

    TAILQ_REMOVE(&descriptor_list, d, next);

    if (d->pl != NULL && d->pl->obj != NULL) {
	event = new_obj_event(Q_PLAYER, ETYPE_NONE, "remove_player_game",
			      d->pl->obj);
	queue_obj_event(event);
    }

    /* If there is anything in the output buffer, clear it and free it */
    if (d->out_buf) {
	free(d->out_buf);
	d->out_buf=0;
    }

    /* If there is anything in the input buffer, clear it and free it */
    if (d->in_buf) {
	free(d->in_buf);
	d->in_buf=0;
    }
    /* Free the structure */
    pthread_mutex_unlock(&d->mutex);
    pthread_mutex_destroy(&d->mutex);
    free(d);
}

/**
   \brief queue data to be sent by network thread
   \param d descriptor to queue against
   \param buf a c string to queue
   \note Assumes descriptor is locked.

   A function you may wish to use to put data in the descriptor structure's
   output buffer.
*/

void do_write(descriptor_data_t *d, char *buf)
{
    char *p;

    if (!d->out_buf) {
	d->out_buf = (char *)malloc(strlen(buf)+1);
	strcpy(d->out_buf, buf);
	d->outbuflen = strlen(buf);
    } else {
	d->out_buf = (char *)realloc(d->out_buf, (d->outbuflen+strlen(buf)+2));
	p = d->out_buf;
	p += d->outbuflen;
        memcpy(p, buf, strlen(buf));
	d->outbuflen += strlen(buf);
    }
}
/**
   \brief queue data to be sent by network thread (locks descriptor)
   \param d descriptor to queue against
   \param buf a c string to queue
   \note locks the descriptor
*/

void lock_do_write(descriptor_data_t *d, char *buf)
{
    pthread_mutex_lock(&d->mutex);
    do_write(d, buf);
    pthread_mutex_unlock(&d->mutex);
}

/**
   \brief Read data from the incoming network and place it on the d buffer
   \param d the descriptor to work on
   \note assumes d is locked on entry
   \return -1 on read failure, or number of bytes read.

   * Read data from client. If the descriptor is non-delayed, we are
   * polling for input. If read() returns 0, then the client
   * disconnected and the server needs to know so it can close the
   * descriptor and free the descriptor_data struct associated with it.
   * If the descriptor is delayed, read() is only going to return -1
   * for an error, or the number of bytes read from the client. This is
   * why we will always return -1 if < 1 byte is read.
*/

int process_input(descriptor_data_t *d)
{
    char buf[8193];
    int i;

    *buf='\0';

    if ((i=read(d->descriptor, buf, 8192)) < 1)
	return(-1);

    buf[i] = '\0';

    d->idle = time(0);

    if (!d->in_buf) {
	d->in_buf = (char *)malloc(strlen(buf)+1);
	strcpy(d->in_buf, buf);
    } else {
	d->in_buf = (char *)realloc(d->in_buf, (strlen(d->in_buf)+strlen(buf))+2);
	strcat(d->in_buf, buf);
    }

    return(i);
}

/**
   \brief If there is any data waiting for output, send it
   \param d descriptor to work on
   \return new buffer position, -1 for failure
   \note assumes d is locked on entry
   \note this function works with binary data
*/

int process_output(descriptor_data_t *d)
{
    int i=0, cpos;
    char *p;

    if (d->outbuflen > 0) {
	p = d->out_buf;
	cpos = 0;
	while (cpos < d->outbuflen) {
	    i = write(d->descriptor, p, d->outbuflen-cpos);
	    if (i < 0) {
		/* don't sit around waiting for slow clients all day */
		if (errno == EAGAIN) {
		    if (cpos > 0) {
			memmove(d->out_buf, p, d->outbuflen-cpos);
			d->outbuflen -= cpos;
			return(cpos);
		    } else {
			return(0);
		    }
		} else
		    return(-1);
	    } else {
		cpos += i;
		p += i;
	    }
	}
	free(d->out_buf);
	d->out_buf = 0;
	d->outbuflen = 0;
	return(cpos);
    } else
	return(0);
}

/**
   \brief Process commands
   \param d descriptor to process
   \note Assumes d is locked on entry.

   This is where you will take commands from the
   descriptor structure's input buffer and do a command with it.
*/

void process_command(descriptor_data_t *d)
{
    char **words, *buf, *p, *pp, *cmdword = NULL;
    int numwords, command, i;
    pargs_t *args=NULL;

    /*printf("Server Got: %s :END\n", d->in_buf);*/

    p = d->in_buf;
    pp = NULL;
    for (buf=d->in_buf; buf != NULL && strlen(buf) > 0; ) {
	pp = p;
	p = strchr(buf, '\r');
	if (p == NULL)
	    goto next;
	*p = '\0';
	p++;
	words = parse_netcommand(buf, &numwords);
	if (words == NULL)
	    goto next;
	cmdword= strdup(words[0]);
	args = parse_command(words, &command, numwords);
	
	if (command < 1 || command > COMMAND_MAX)
	    parsed_command(d, cmdword, args);
	else
	    parsefunc[command](d, args);
	free(cmdword);
    next:
	if (args) {
	    for (i=0; args[i].cword != -1; i++)
		if (args[i].type == PTCHAR)
		    free_string(args[i].arg.c);
	    free(args);
	    args = NULL;
	}
	buf = p;
    }
    if (p == NULL && pp != NULL) {
        buf = strdup(pp);
        strcpy(d->in_buf, buf);
        free(buf);
    } else {
	free(d->in_buf);
	d->in_buf=0;
    }
}

/**
   \brief initialize the network
*/

void init_network(void)
{
    /* Ignore SIGPIPE */
    signal(SIGPIPE, SIG_IGN);

    /* Initialize the socket */
    if ((nonpersist.mainsocket=init_socket(settings.bindport)) < 0)
	bailout();

    nonpersist.maxdesc = nonpersist.mainsocket;
    nonpersist.avail_descs = getdtablesize() - 2;
}

/**
   \brief the main network processing thread
   \param nada ignored
*/

void *network_loop(void *nada)
{
    struct timeval lastt, curt, tspent, tout, nullt, optt;
    fd_set input, output, errfds;
    descriptor_data_t *d, *dnext;

    /* Main loop */
    for (;;) {
	FD_ZERO(&input);
	FD_ZERO(&output);
	FD_ZERO(&errfds);
	FD_SET(nonpersist.mainsocket, &input);

	pthread_mutex_lock(&dlist_mutex);
	TAILQ_FOREACH(d, &descriptor_list, next) {
	    pthread_mutex_lock(&d->mutex);
#ifdef NONBLOCK
	    /*
	     * We don't need to set these file descriptor sets if
	     * the descriptors are delayed.
	     */
	    FD_SET(d->descriptor, &input);
	    FD_SET(d->descriptor, &output);
#endif /* #ifdef NONBLOCK */
	    FD_SET(d->descriptor, &errfds);
	    pthread_mutex_unlock(&d->mutex);
	}
	pthread_mutex_unlock(&dlist_mutex);

	nullt.tv_sec = 0;
	nullt.tv_usec = 0;

	optt.tv_usec = USEC_TIMEOUT;
	optt.tv_sec = 0;

	gettimeofday(&curt, (struct timezone *)0);
	tspent = tvaldiff(&curt, &lastt);
	tout = tvaldiff(&optt, &tspent);
	lastt.tv_sec = curt.tv_sec + tout.tv_sec;
	lastt.tv_usec = curt.tv_usec + tout.tv_usec;
	if (lastt.tv_usec >= 1000000) {
	    lastt.tv_usec -= 1000000;
	    lastt.tv_sec++;
	}
	if (select(nonpersist.maxdesc + 1, &input, &output, &errfds, &nullt) < 0)
	    bailout();

	if (select(0, 0, 0, 0, &tout) < 0)
	    perror("select() sleep");
	/* Is there a new connection? */
	if (FD_ISSET(nonpersist.mainsocket, &input))
	    if (init_desc(nonpersist.mainsocket) < 0)
		perror("init_desc()");

	/* Is there any input, output, or errors for connected
	   descriptors out there? */
	pthread_mutex_lock(&dlist_mutex);
	for (d=TAILQ_FIRST(&descriptor_list); d != NULL; d = dnext) {
	    pthread_mutex_lock(&d->mutex);
	    dnext = TAILQ_NEXT(d, next);
	    /* Connection idled too long */
	    if (d && d->idle && (time(0) - d->idle >= IDLE_TIMEOUT)) {
		LOG(LOG_DEBUG, "Idle Timeout: Desc %d", d->descriptor);
		close_desc(d->descriptor);
		free_desc(d);
		d = 0;
		continue;
	    } else if (d && FD_ISSET(d->descriptor, &errfds)) {
	    /* We have an error on the descriptor */
		FD_CLR(d->descriptor, &input);
		FD_CLR(d->descriptor, &output);
		close_desc(d->descriptor);
		free_desc(d);
		d = 0;
		continue;
	    }

#ifdef NONBLOCK
	    /*
	     * Only check FD_ISSET if this is a non-delayed
	     * server. If this is a delayed server, you may need
	     * to swap the process_output and process_input if
	     * the server waits for input from the client before
	     * doing any output, instead of doing output and then
	     * waiting for input.
	     */

	    /* Is there any input or output pending? */
	    else if ( FD_ISSET(d->descriptor, &input) ||
		      FD_ISSET(d->descriptor, &output) ) {
		if (d && FD_ISSET(d->descriptor, &output))
#endif /* #ifdef NONBLOCK */
		    if (process_output(d) < 0) {
			close_desc(d->descriptor);
			free_desc(d);
			d = 0;
			continue;
		    }
#ifdef NONBLOCK
		if (d && FD_ISSET(d->descriptor, &input))
#endif /* #ifdef NONBLOCK */
		    if (process_input(d) < 0) {
			close_desc(d->descriptor);
			free_desc(d);
			d = 0;
			continue;
		    }
#ifdef NONBLOCK
	    }
#endif /* #ifdef NONBLOCK */

	    /* Process command */
	    if (d && d->in_buf)
		process_command(d);
	    pthread_mutex_unlock(&d->mutex);
	}
	pthread_mutex_unlock(&dlist_mutex);
    }
}
