/* $Id: object.c,v 1.53 2004/04/01 07:29:54 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif
#include <unistd.h>
#include <sys/param.h>
#include <math.h>

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "object.h"
#include "map.h"
#include "shstr.h"
#include "utils.h"
#include "models.h"
#include "collision.h"
#include "tod.h"
#include "xmlparse.h"
#include "matrix.h"
#include "event.h"
#include "physics.h"
#include "player.h"
#include "settings.h"
#include "sutils.h"
#include "servdata.h"

/**
   \file object.c
   \brief Object related functions
   \author Tim Rightnour
*/

TAILQ_HEAD(, _object_t) arch = TAILQ_HEAD_INITIALIZER(arch);

extern pthread_mutex_t arch_mutex;
extern maphead_t *activemap;
extern pthread_mutex_t maps_mutex;
extern pthread_mutex_t collision_mutex;
extern TAILQ_HEAD(, _mapgroup_t) maps;
extern settings_t settings;
extern pthread_mutex_t objtable_mutex;
extern persist_data_t persist;
extern eflag_t evflags[];
extern const size_t evflag_size;
extern npersist_data_t nonpersist;

eflag_t eflags[] = {
    {"head", EQ_HEAD},
    {"neck", EQ_NECK},
    {"body", EQ_BODY},
    {"rshoulder", EQ_RSHOULDER},
    {"rarm", EQ_RARM},
    {"rfarm", EQ_RFARM},
    {"rhand", EQ_RHAND},
    {"wrhand", EQ_WRHAND},
    {"rring", EQ_RRING},
    {"lshoulder", EQ_LSHOULDER},
    {"larm", EQ_LARM},
    {"lfarm", EQ_LFARM},
    {"lhand", EQ_LHAND},
    {"wlhand", EQ_WLHAND},
    {"lring", EQ_LRING},
    {"waist", EQ_WAIST},
    {"ulegs", EQ_ULEGS},
    {"llegs", EQ_LLEGS},
    {"feet", EQ_FEET}
};

eflag_t oflags[] = {
    {"alive", FLAG_ALIVE},
    {"no_pick", FLAG_NO_PICK},
    {"aggr", FLAG_AGGR},
    {"metaaggr", FLAG_META_AGGR},
    {"cursed", FLAG_CURSED},
    {"blessed", FLAG_BLESSED},
    {"blind", FLAG_BLIND},
    {"immobile", FLAG_IMMOBILE},
    {"nocollide", FLAG_NOCOLLIDE},
    {"notip", FLAG_NOTIP},
    {"roll", FLAG_ROLL},
    {"rollx", FLAG_ROLLX},
    {"rollz", FLAG_ROLLZ},
    {"anchor", FLAG_ANCHOR},
    {"antigrav", FLAG_ANTIGRAV},
    {"internal1", FLAG_STABLE_ROLL},
    {"internal2", FLAG_STABLE_TIP},
    {"internal3", FLAG_STABLE_SLIDE},
    {"internal4", FLAG_DESTROYED}
};

const size_t eflag_size = sizeof(eflags) / sizeof(eflag_t);
const size_t oflag_size = sizeof(oflags) / sizeof(eflag_t);

/*=========   uid functions ===================== */
static pthread_mutex_t uid_mutex = PTHREAD_MUTEX_INITIALIZER;

/** generate a new uid */
void uid_new(uniqid_t *id) {
    pthread_mutex_lock(&uid_mutex);
    uid_inc(&(persist.uid_seed));
    uid_cpy(id, &(persist.uid_seed));
    pthread_mutex_unlock(&uid_mutex);
}

/*===== new obj_table with hash =========*/
#define UID_HASH_SIZE 512 
#define UID_HASH(o) ((o)[3] &511)
TAILQ_HEAD(uid_hash_t, _object_t) _uid_objtable[UID_HASH_SIZE];

void init_uid_obj_table() {
    int i;
    for(i=0;i<UID_HASH_SIZE;i++) TAILQ_INIT(&_uid_objtable[i]);
};

void add_uid_obj_table(object_t *obj) {
    int h;
    pthread_mutex_lock(&objtable_mutex);
    /* set a new uid if we don't have one */
    if (uid_isnull(&(obj->cli.uid))) uid_new(&(obj->cli.uid));
    h = UID_HASH(obj->cli.uid);
    TAILQ_INSERT_HEAD(&_uid_objtable[h],obj,hash_next);
    pthread_mutex_unlock(&objtable_mutex);
};

void rem_uid_obj_table(object_t *obj) {
    int h;
    pthread_mutex_lock(&objtable_mutex);
    h = UID_HASH(obj->cli.uid);
    TAILQ_REMOVE(&_uid_objtable[h],obj,hash_next);
    pthread_mutex_unlock(&objtable_mutex);
};
/** \brief search an object from uid 
\note it increment refcount. Don't forget to call obj_decref
*/
struct _object_t * get_obj_from_uid(uniqid_t*id ) {
    struct _object_t * obj;
    int h;
    pthread_mutex_lock(&objtable_mutex);
    h = UID_HASH(*id);
    TAILQ_FOREACH(obj,&_uid_objtable[h],hash_next) 
	if(uid_eq(id,&obj->cli.uid)) {
	    obj->refcount++;
	    pthread_mutex_unlock(&objtable_mutex);
	    return obj;
	}
    pthread_mutex_unlock(&objtable_mutex);
    return (struct _object_t *)NULL;
};




/**
   \brief automove an object along it's current path
   \param obj object to move
   \note This routine needs to become much more complex.  Only works on
   modeled objects
*/

void do_automove_obj(object_t *obj)
{
    event_t *event;
    maphead_t *maph;

    if (obj->cli.model != NULL) {
	if (attempt_move_obj(obj)) {
	    maph = obj_on_map(obj);
	    if (maph != obj->maphead)
		obj->maphead = maph;
	    event = new_obj_event(Q_PLAYER, 0, "object_teach_players_self",
				  obj);
	    queue_obj_event(event);
	    return;
	}
    }
}

/**
   \brief find radius of a given object
   \param obj object to find radius of
   \return radius
*/

float obj_radius(object_t *obj)
{
    float r1;

    r1 = fabs(obj->bmin.x*obj->cli.scale.x);
    if (fabs(obj->bmax.x*obj->cli.scale.x) > r1)
	r1 = fabs(obj->bmax.x*obj->cli.scale.x);

    if (fabs(obj->bmin.z*obj->cli.scale.z) > r1)
	r1 = fabs(obj->bmin.z*obj->cli.scale.z);
    if (fabs(obj->bmax.z*obj->cli.scale.z) > r1)
	r1 = fabs(obj->bmax.z*obj->cli.scale.z);

    if (fabs(obj->bmin.y*obj->cli.scale.y) > r1)
	r1 = fabs(obj->bmin.y*obj->cli.scale.y);
    if (fabs(obj->bmax.y*obj->cli.scale.y) > r1)
	r1 = fabs(obj->bmax.y*obj->cli.scale.y);
    return(r1);
}

/**
   \brief instantiate a 3d model in the collision datastream
   \param obj object to collisionify
   \note takes the arch lock
*/

void put_obj_in_space(object_t *obj)
{
    int found=0;
    char file[MAX_PATH];

    /* it is possible to have an object with no model data, such as an
       invisible force.  It requires no SOLID. */
    if (obj->cli.model == NULL)
	return;

    if (QUERY_FLAG(obj, FLAG_NOCOLLIDE) && obj->archobj->cli.model != NULL) {
	sprintf(file, "%s/%s", settings.archdir, obj->archobj->cli.model);
	if (strncmp(strrchr(obj->archobj->cli.model, '.'), ".xgm", 4) == 0) {
	    pthread_mutex_lock(&arch_mutex);
	    parse_genmodel_boxonly(file, obj->archobj);
	    parse_genmodel_boxonly(file, obj);
	    obj->center = obj->cli.coord;
	    obj->bmin.x *= obj->cli.scale.x;
	    obj->bmin.y *= obj->cli.scale.y;
	    obj->bmin.z *= obj->cli.scale.z;
	    obj->bmax.x *= obj->cli.scale.x;
	    obj->bmax.y *= obj->cli.scale.y;
	    obj->bmax.z *= obj->cli.scale.z;
	    obj->coff.x *= obj->cli.scale.x;
	    obj->coff.y *= obj->cli.scale.y;
	    obj->coff.z *= obj->cli.scale.z;
	    obj->center.y += (obj->bmin.y+obj->bmax.y)/2.0;
	    obj->radius = obj_radius(obj);
	    pthread_mutex_unlock(&arch_mutex);
	}
	return;
    }
    if (obj->shape != NULL)
	found++;

    pthread_mutex_lock(&arch_mutex);

    if (obj->archobj != NULL && obj->archobj->shaperef == 0 &&
	obj->archobj->cli.model != NULL && !found) {
	sprintf(file, "%s/%s", settings.archdir, obj->archobj->cli.model);
	if (strncmp(strrchr(obj->archobj->cli.model, '.'), ".xgm", 4) == 0) {
	    parse_genmodel(file, obj->archobj);
	    obj->shape = obj->archobj->shape;
	    obj->archobj->shaperef += 2;
	    found++;
	} else if (strncmp(strrchr(obj->archobj->cli.model, '.'),
			   ".xcm", 4) == 0) {
	    parse_calmodel(file, obj->archobj);
	    obj->shape = obj->archobj->shape;
	    obj->archobj->shaperef += 2;
	    found++;
	}
    } else if (obj->archobj != NULL && obj->archobj->shaperef > 0 &&
	!found) {
	obj->shape = obj->archobj->shape;
	obj->archobj->shaperef++;
	found++;
    }
    if (found) {
	if (obj->archobj != NULL) {
	    obj->bmin = obj->archobj->bmin;
	    obj->bmax = obj->archobj->bmax;
	    obj->coff = obj->archobj->coff;
	    obj->center = obj->archobj->center;
	    obj->bmin.x *= obj->cli.scale.x;
	    obj->bmin.y *= obj->cli.scale.y;
	    obj->bmin.z *= obj->cli.scale.z;
	    obj->bmax.x *= obj->cli.scale.x;
	    obj->bmax.y *= obj->cli.scale.y;
	    obj->bmax.z *= obj->cli.scale.z;
	    obj->coff.x *= obj->cli.scale.x;
	    obj->coff.y *= obj->cli.scale.y;
	    obj->coff.z *= obj->cli.scale.z;
	    obj->center.x *= obj->cli.scale.x;
	    obj->center.y *= obj->cli.scale.y;
	    obj->center.z *= obj->cli.scale.z;
	    obj->radius = obj_radius(obj);
	}
	printf("Put OBJ at %f %f %f\n", obj->cli.coord.x, obj->cli.coord.y,
	       obj->cli.coord.z);
    }
    pthread_mutex_unlock(&arch_mutex);
}

/**
   \brief collision map an object
   \param obj object to map
   \param coord coord to place object at.
   \note expects collision lock to be held.
*/

void collision_map_obj(object_t *obj, vertex_t coord)
{
    float quat4[4];

    if (obj->shape == NULL || QUERY_FLAG(obj, FLAG_NOCOLLIDE) ||
	obj->cli.model == NULL)
	return;

    dtCreateObject(obj, obj->shape);
    dtSelectObject(obj);
    dtLoadIdentity();
    dtTranslate((coord.x-obj->coff.x), coord.y, (coord.z-obj->coff.z));
    dtScale(obj->cli.scale.x, obj->cli.scale.y, obj->cli.scale.z);
    euler_to_quat4(obj->cli.facing.x, obj->cli.facing.y,
		   obj->cli.facing.z, quat4);
    dtRotate(quat4[0], quat4[1], quat4[2], quat4[3]);
    dtSetObjectResponse(obj, collide_count, DT_SIMPLE_RESPONSE, (void *)obj);
}

/**
   \brief place an object on the collision plane, and test it
   \param obj object to place and test against
   \param coord coord to place object at.
   \return 0 no collision, 1 collision
*/

int place_obj_and_test(object_t *obj, vertex_t coord)
{
    int ret;

    if (QUERY_FLAG(obj, FLAG_NOCOLLIDE))
	return(0);

    pthread_mutex_lock(&collision_mutex);
    collision_map_obj(obj, coord);
    ret = perform_collision_test(obj, coord);
    pthread_mutex_unlock(&collision_mutex);
    return(ret);
}

/**
   \brief remove object from the collision datastream
   \param obj obj to de-collisionify
   \note Don't ever call this on an arch obj. Takes the arch lock.
*/

void remove_obj_from_space(object_t *obj)
{
    if (obj->cli.model == NULL || QUERY_FLAG(obj, FLAG_NOCOLLIDE))
	return;
    pthread_mutex_lock(&arch_mutex);
    if (obj->archobj != NULL) {
        obj->archobj->shaperef--;
    }
    obj->shape = NULL;
    pthread_mutex_unlock(&arch_mutex);
}

/**
   \brief unload an object's space from memory
   \param obj obj to de-collisionify
   \note Use of this function should be very rare, and should ONLY occur on the
   arch object. Takes the arch lock
*/

void unload_obj_space(object_t *obj)
{
    pthread_mutex_lock(&arch_mutex);
    if (obj->shaperef == 1) {
	dtDeleteShape(obj->shape);
	obj->shaperef=0;
    } else {
	LOG(LOG_ERROR, "unload_obj_shape called with shaperef != 1");
    }
    pthread_mutex_unlock(&arch_mutex);
}

/**
   \brief put an object inside another object.  
   \param obj object to place
   \param cont containter to place object into
   \param slot bucket to put object into
   \note places obj into cont, in slot.  See BUCKET_*
*/

void put_obj_in_obj(object_t *obj, object_t *cont, int slot)
{
    TAILQ_HEAD(foo, _object_t) *bucket = NULL;

    switch (slot) {
    case BUCKET_SPELL:
	bucket = (struct foo *)&cont->bucket.spells;
	break;
    case BUCKET_JOB:
	bucket = (struct foo *)&cont->bucket.jobs;
	break;
    case BUCKET_INV:
	bucket = (struct foo *)&cont->bucket.inv;
	break;
    case BUCKET_SKILL:
	bucket = (struct foo *)&cont->bucket.skills;
	break;
    case BUCKET_KEY:
	bucket = (struct foo *)&cont->bucket.keys;
	break;
    case BUCKET_EQUIP:
	bucket = (struct foo *)&cont->bucket.equip;
	break;
    default:
	LOG(LOG_ERROR, "put_obj_in_obj called with bad slot:%d", slot);
	return;
    }
    if (TAILQ_EMPTY(bucket))
	TAILQ_INIT(bucket);
    TAILQ_INSERT_TAIL(bucket, obj, stack);
    obj->container = cont;
    obj_incref(cont);
    /* the cont references the obj */
    obj_incref(obj);
}

/**
   \brief remove an object from another object
   \param obj object to remove
   \param cont container to remove from
   \param slot slot to remove object from
*/

void remove_obj_from_obj(object_t *obj, object_t *cont, int slot)
{
    TAILQ_HEAD(foo, _object_t) *bucket = NULL;

    switch (slot) {
    case BUCKET_SPELL:
	bucket = (struct foo *)&cont->bucket.spells;
	break;
    case BUCKET_JOB:
	bucket = (struct foo *)&cont->bucket.jobs;
	break;
    case BUCKET_INV:
	bucket = (struct foo *)&cont->bucket.inv;
	break;
    case BUCKET_SKILL:
	bucket = (struct foo *)&cont->bucket.skills;
	break;
    case BUCKET_KEY:
	bucket = (struct foo *)&cont->bucket.keys;
	break;
    case BUCKET_EQUIP:
	bucket = (struct foo *)&cont->bucket.equip;
	break;
    default:
	LOG(LOG_ERROR, "put_obj_in_obj called with bad slot:%d", slot);
	return;
    }
    TAILQ_REMOVE(bucket, obj, stack);
    obj_decref(obj);
    obj_decref(obj->container);
    obj->container = NULL;
}

/**
   \brief check that object is in a specific container
   \param obj obj to check for
   \param cont container to check in
   \param slot slot to look in
   \return true/false
   \todo make recursive
*/

int verify_obj_in_bucket(object_t *obj, object_t *cont, int slot)
{
    TAILQ_HEAD(foo, _object_t) *bucket = NULL;
    object_t *tmp;
    int found=0;

    switch (slot) {
    case BUCKET_SPELL:
	bucket = (struct foo *)&cont->bucket.spells;
	break;
    case BUCKET_JOB:
	bucket = (struct foo *)&cont->bucket.jobs;
	break;
    case BUCKET_INV:
	bucket = (struct foo *)&cont->bucket.inv;
	break;
    case BUCKET_SKILL:
	bucket = (struct foo *)&cont->bucket.skills;
	break;
    case BUCKET_KEY:
	bucket = (struct foo *)&cont->bucket.keys;
	break;
    case BUCKET_EQUIP:
	bucket = (struct foo *)&cont->bucket.equip;
	break;
    default:
	LOG(LOG_ERROR, "verify_obj_in_bucket called with bad slot:%d", slot);
	return(0);
    }
    TAILQ_FOREACH(tmp, bucket, stack) {
	if (obj == tmp)
	    found++;
    }
    return(found);
}

/**
   \brief Verify that an object is within the inventory of another object
   \param obj object to look within
   \param target object we wish to verify
   \return true/false
   \note upwards recursive
*/

int verify_obj_in_obj(object_t *obj, object_t *target)
{
    object_t *tmp;

    tmp = target;

    while ((tmp = tmp->container) != NULL)
	if (tmp == obj)
	    return(1);
    return(0);
}
    
/**
   \brief create an object from archname
   \param arch arch name
   \param recursive do we create it's inventory?
   \see duplicate_object
   \note register object in obj_table and create uid
*/

struct _object_t * new_obj_from_arch(char * arch,  int recursive){
    object_t *archobj,*obj;
    archobj = find_arch_byname(arch); /* XXX */
    if (archobj == NULL) {
	LOG(LOG_FATAL, "arch %s does not exist!", arch);
	bailout();
    }
    obj = new_obj();
    duplicate_object(archobj, obj, recursive);
    nonpersist.objectcount++;
    add_uid_obj_table(obj);
   return obj;
};
/**
   \brief duplicate an object
   \param from object to clone
   \param to object to change
   \param recursive do we create it's inventory?
   \todo recursive doesn't work yet

   Duplicate an object into a pre-allocated object structure.  Recursive, see
   DO_*, do we duplicate inventory bits?  If so, we allocate those, rather
   than assuming the caller knew the structure.
*/

void duplicate_object(object_t *from, object_t *to, int recursive)
{
    int i;
    lore_t *nlore, *tlore;
    dialog_t *ndialog, *tdialog;
    dkeyword_t *nkey, *tkey;

    to->arch = add_string(from->arch);
    if (recursive)
	/* next, stack ignored, but bucket_t should be parsed */
	printf("Not recursive in duplicate_object yet.\n");

    /* client structure */
    /* this works, who knew */
    to->cli = from->cli;
    if (from->cli.name)
	to->cli.name = add_string(from->cli.name);
    if (from->cli.name_pl)
	to->cli.name_pl = add_string(from->cli.name_pl);
    if (from->cli.model)
	to->cli.model = add_string(from->cli.model);

    for (i=0; i < FLAGARRAYSZ; i++)
	to->flags[i] = from->flags[i];
    to->msg = add_string(from->msg);
    to->bmin = from->bmin;
    to->bmax = from->bmax;
    to->coff = from->coff;

    /* set the archobj */
    if (from->archobj == NULL)
	to->archobj = from;

    /* Yuck, now we have to duplicate the dialog and lore */
    TAILQ_FOREACH(tlore, &from->lore, next) {
	if (TAILQ_EMPTY(&to->lore))
	    TAILQ_INIT(&to->lore);
	nlore = smalloc(lore_t);
	nlore->chance = tlore->chance;
	nlore->text = add_string(tlore->text);
	/* lore is allready sorted.. no reason to resort */
	TAILQ_INSERT_TAIL(&to->lore, nlore, next);
    }
    TAILQ_FOREACH(tdialog, &from->dialog, next) {
	if (TAILQ_EMPTY(&to->dialog))
	    TAILQ_INIT(&to->dialog);
	ndialog = smalloc(dialog_t);
	ndialog->text = add_string(tdialog->text);
	TAILQ_FOREACH(tkey, &tdialog->keywords, next) {
	    if (TAILQ_EMPTY(&ndialog->keywords))
		TAILQ_INIT(&ndialog->keywords);
	    nkey = smalloc(dkeyword_t);
	    nkey->key = add_string(tkey->key);
	    TAILQ_INSERT_TAIL(&ndialog->keywords, nkey, next);
	}
	TAILQ_INSERT_TAIL(&to->dialog, ndialog, next);
    }
}

/**
   \brief search the arch TAILQ for an arch matching name
   \param name name to look for
   \return object_t * if found, NULL if not

   search the arch TAILQ for an arch matching name. Returns NULL if not found.
   XXX: For now, this is a linear serch.  It should be possible to make
   a binary search by cheating, and creating an array of pointers into
   the arch TAILQ.  There will have to be an update_arch_table() routine
   to go through the TAILQ and create the table, this will have to be run
   after ANY insertions into the queue.  So if we implement additive loading
   of arches, the table will have to be regenerated.  Normally it would only
   have to be generated once at boot.  We would trade memory for speed.
*/

object_t *find_arch_byname(char *name)
{
    object_t *obj;
    size_t len;

    len = strlen(name);
    TAILQ_FOREACH(obj, &arch, next) {
	if (!strncmp(obj->arch, name, len)) {
	    return(obj);
	}
    }
    return(NULL);
}

/** \brief compare two flags */

static int compare_flag(const void *a, const void *b)
{
    return strcmp(((eflag_t *)a)->flag, ((eflag_t *)b)->flag);
}

/**
   \brief initialize the equipment flags and object flags
*/

void init_flags(void)
{
    qsort((char *)eflags, eflag_size, sizeof(eflag_t),
          compare_flag);
    qsort((char *)oflags, oflag_size, sizeof(eflag_t),
          compare_flag);
    qsort((char *)evflags, evflag_size, sizeof(eflag_t),
          compare_flag);
}

/**
   \brief parse equipment flags off an object
   \param doc xmlDocPtr
   \param cur current location in file
   \param obj object to parse
*/

void parse_equipflags(xmlDocPtr doc, xmlNodePtr cur, object_t *obj)
{
    xmlNodePtr child;
    eflag_t *asp, dummy;

    child = cur->xmlChildrenNode;
    while (child != NULL) {
	dummy.flag = (char *)child->name;
	asp = (eflag_t *)bsearch((void *)&dummy, (void *)eflags, eflag_size,
				 sizeof(eflag_t), compare_flag);
	if (asp)
	    SET_EFLAG(obj->cli.equip, asp->num);
	child = child->next;
    }
}

/**
   \brief parse flags off an object
   \param doc xmlDocPtr
   \param cur current location in file
   \param obj object to parse
*/

void parse_flags(xmlDocPtr doc, xmlNodePtr cur, object_t *obj)
{
    xmlNodePtr child;
    eflag_t *asp, dummy;

    child = cur->xmlChildrenNode;
    while (child != NULL) {
	dummy.flag = (char *)child->name;
	asp = (eflag_t *)bsearch((void *)&dummy, (void *)oflags, oflag_size,
				 sizeof(eflag_t), compare_flag);
	if (asp)
	    SET_FLAG(obj, asp->num);
	child = child->next;
    }
}

/**
   \brief parse a lore_t from the XML stuff
   \param doc xmlDocPtr
   \param cur current location in file
   \param obj object to parse
   \note there can only be one lore_t per object with this code
*/

void parse_lore(xmlDocPtr doc, xmlNodePtr cur, object_t *obj)
{
    xmlNodePtr child;
    char *tmp, *ch;
    lore_t *nlore, *tlore;
    int placed;

    child = cur->xmlChildrenNode;
    while (child != NULL) {
	if (!strcmp(child->name, "loredata")) {
	    if (TAILQ_EMPTY(&obj->lore))
		TAILQ_INIT(&obj->lore);
	    nlore = smalloc(lore_t);
	    tmp = xmlNodeListGetString(doc, child->xmlChildrenNode, 1);
	    if (xmlHasProp(child, "chance")) {
		ch = xmlGetProp(child, "chance");
		nlore->chance = atoi(ch);
		free(ch);
	    } else
		nlore->chance = 1;
	    nlore->text = add_string(tmp);
	    free(tmp);
	    placed = 0; /* keep sorted by chance */
	    TAILQ_FOREACH(tlore, &obj->lore, next) {
		if (nlore->chance < tlore->chance) {
		    TAILQ_INSERT_BEFORE(tlore, nlore, next);
		    placed++;
		    break;
		}
	    }
	    if (!placed)
		TAILQ_INSERT_TAIL(&obj->lore, nlore, next);
	    child = child->next;
	}
    }
}

/**
   \brief parse a dialog_t from the XML stuff
   \param doc xmlDocPtr
   \param cur current location in file
   \param obj object to parse
   \note there can only be one dialog_t per object with this code
*/

void parse_dialog(xmlDocPtr doc, xmlNodePtr cur, object_t *obj)
{
    xmlNodePtr child, grandchild;
    char *tmp;
    dialog_t *dialog;
    dkeyword_t *key, *tmpkey;
    int placed;

    child = cur->xmlChildrenNode;
    while (child != NULL) {
	if (!strcmp(child->name, "dentry")) {
	    if (TAILQ_EMPTY(&obj->dialog))
		TAILQ_INIT(&obj->dialog);
	    dialog = smalloc(dialog_t);
	    grandchild = child->xmlChildrenNode;

	    while (grandchild != NULL) {
		if (!strcmp(grandchild->name, "dtext")) {
		    tmp = xmlNodeListGetString(doc, grandchild->xmlChildrenNode, 1);
		    dialog->text = add_string(tmp);
		    free(tmp);
		}
		if (!strcmp(grandchild->name, "dkey")) {
		    if (TAILQ_EMPTY(&dialog->keywords))
			TAILQ_INIT(&dialog->keywords);
		    key = smalloc(dkeyword_t);
		    tmp = xmlNodeListGetString(doc, grandchild->xmlChildrenNode, 1);
		    key->key = add_string(tmp);
		    free(tmp);
		    /* sort the keywords lexographically */
		    placed = 0;
		    TAILQ_FOREACH(tmpkey, &dialog->keywords, next) {
			if (strcmp(key->key, tmpkey->key) <= 0) {
			    TAILQ_INSERT_BEFORE(tmpkey, key, next);
			    placed++;
			    break;
			}
		    }
		    if (!placed)
			TAILQ_INSERT_TAIL(&dialog->keywords, key, next);
		}
		grandchild = grandchild->next;
	    }
	    /* keep in same order read from file */
	    TAILQ_INSERT_TAIL(&obj->dialog, dialog, next);
	}
	child = child->next;
    }
}

/**
   \brief allocate space for a new object
   \return returns a pointer to a new object
   The smalloc performs a bzero, and we set the scale to 1.0 (default)
	refcount is set to 1
*/

object_t *new_obj(void)
{
    object_t *obj;

    obj = smalloc(object_t);
    obj->cli.scale.x = 1.0;
    obj->cli.scale.y = 1.0;
    obj->cli.scale.z = 1.0;
    obj->refcount = 1;
    return(obj);
}

/**
   \brief remove an object from the game
   \param obj object to remove
   \todo things that call this should probably be marking the object as dead
   and a cleanup event should do this actual work.  Much more needs be done
   here to make this clean. 

   this is to be used on NON ARCHETYPE objects only.  removing an archetype
   is major bad karma.
*/

void remove_obj(object_t *obj)
{
    dialog_t *dia;
    object_t *tmp;

    /* mark object as destructed */
    SET_FLAG(obj, FLAG_DESTROYED);
    
    rem_uid_obj_table(obj); 
    /* recurse into the buckets */
    while ((tmp = TAILQ_FIRST(&obj->bucket.spells)) != NULL) {
	remove_obj_from_obj(tmp, obj, BUCKET_SPELL);
	remove_obj(tmp);
	TAILQ_REMOVE(&obj->bucket.spells, tmp, stack);
    }
    while ((tmp = TAILQ_FIRST(&obj->bucket.jobs)) != NULL) {
	remove_obj_from_obj(tmp, obj, BUCKET_JOB);
	remove_obj(tmp);
	TAILQ_REMOVE(&obj->bucket.jobs, tmp, stack);
    }
    while ((tmp = TAILQ_FIRST(&obj->bucket.inv)) != NULL) {
	remove_obj_from_obj(tmp, obj, BUCKET_INV);
	remove_obj(tmp);
	TAILQ_REMOVE(&obj->bucket.inv, tmp, stack);
    }
    while ((tmp = TAILQ_FIRST(&obj->bucket.skills)) != NULL) {
	remove_obj_from_obj(tmp, obj, BUCKET_SKILL);
	remove_obj(tmp);
	TAILQ_REMOVE(&obj->bucket.skills, tmp, stack);
    }
    while ((tmp = TAILQ_FIRST(&obj->bucket.keys)) != NULL) {
	remove_obj_from_obj(tmp, obj, BUCKET_KEY);
	remove_obj(tmp);
	TAILQ_REMOVE(&obj->bucket.keys, tmp, stack);
    }
    while ((tmp = TAILQ_FIRST(&obj->bucket.equip)) != NULL) {
	remove_obj_from_obj(tmp, obj, BUCKET_INV);
	remove_obj(tmp);
	TAILQ_REMOVE(&obj->bucket.equip, tmp, stack);
    }
    /* recurse into the lore */
    while (TAILQ_FIRST(&obj->lore) != NULL) {
	free_string(TAILQ_FIRST(&obj->lore)->text);
	TAILQ_REMOVE(&obj->lore, TAILQ_FIRST(&obj->lore), next);
    }
    /* recurse into the dialog */
    while (TAILQ_FIRST(&obj->dialog) != NULL) {
	dia = TAILQ_FIRST(&obj->dialog);
	free_string(dia->text);
	while (TAILQ_FIRST(&dia->keywords) != NULL) {
	    free_string(TAILQ_FIRST(&dia->keywords)->key);
	    TAILQ_REMOVE(&dia->keywords, TAILQ_FIRST(&dia->keywords), next);
	}
	TAILQ_REMOVE(&obj->dialog, TAILQ_FIRST(&obj->dialog), next);
    }
    /* unqueue and free any events owned by this obj */
    delete_all_obj_event(obj);
    /* decrement ref count on this object */
    obj_decref(obj);
}


/**
   \brief parse an object
   \param doc pointer to xml doc file
   \param cur expected to be at the level of the Object tag on entry.
   \param obj pointer to an object_t, will be filled out by this function.
   \param mode load arch object, or overlay load for a map? (PO_*)
*/

void parse_object(xmlDocPtr doc, xmlNodePtr cur, object_t *obj, int mode)
{
    object_t *tmp;
    char *c;

    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	/* special cases */
	/* these are the inventory buckets */
	if (!strcmp(cur->name, "Object")) {
	    c = xmlGetProp(cur, "bucket");
	    if (c == NULL)
		c = strdup("inv");
	    if (!strcmp(c, "spells")) {
		tmp = new_obj();
		parse_object(doc, cur, tmp, mode);
		put_obj_in_obj(tmp, obj, BUCKET_SPELL);
	    } else if (!strcmp(c, "jobs")) {
		tmp = new_obj();
		parse_object(doc, cur, tmp, mode);
		put_obj_in_obj(tmp, obj, BUCKET_JOB);
	    } else if (!strcmp(c, "inv")) {
		tmp = new_obj();
		parse_object(doc, cur, tmp, mode);
		put_obj_in_obj(tmp, obj, BUCKET_INV);
	    } else if (!strcmp(c, "skills")) {
		tmp = new_obj();
		parse_object(doc, cur, tmp, mode);
		put_obj_in_obj(tmp, obj, BUCKET_SKILL);
	    } else if (!strcmp(c, "keys")) {
		tmp = new_obj();
		parse_object(doc, cur, tmp, mode);
		put_obj_in_obj(tmp, obj, BUCKET_KEY);
       	    } else if (!strcmp(c, "equip")) {
		tmp = new_obj();
		parse_object(doc, cur, tmp, mode);
		put_obj_in_obj(tmp, obj, BUCKET_EQUIP);
	    }
	    if (c)
		free(c);
	}
	/* implement connected here, somehow */

	/* standard elements */
	if (!strcmp(cur->name, "arch"))
	    parsestring(doc, cur, &obj->arch);
	if (!strcmp(cur->name, "name"))
	    parsestring(doc, cur, &obj->cli.name);
	if (!strcmp(cur->name, "name_pl"))
	    parsestring(doc, cur, &obj->cli.name_pl);
	if (!strcmp(cur->name, "model"))
	    parsestring(doc, cur, &obj->cli.model);
	if (!strcmp(cur->name, "msg"))
	    parsestring(doc, cur, &obj->msg);
	if (!strcmp(cur->name, "lore"))
	    parse_lore(doc, cur, obj);
	if (!strcmp(cur->name, "dialog"))
	    parse_dialog(doc, cur, obj);
	if (!strcmp(cur->name, "facing"))
	    parsevector(doc, cur, &obj->cli.facing);
	if (!strcmp(cur->name, "direction"))
	    parsevector(doc, cur, &obj->cli.direction);
	if (!strcmp(cur->name, "coord"))
	    parsevector(doc, cur, &obj->cli.coord);
	if (!strcmp(cur->name, "scale"))
	    parsevector(doc, cur, &obj->cli.scale);
	if (!strcmp(cur->name, "velocity"))
	    parsefloat(doc, cur, &obj->cli.velocity);
	if (!strcmp(cur->name, "sex"))
	    parseint(doc, cur, &obj->cli.stats.sex);
	if (!strcmp(cur->name, "type"))
	    parseint(doc, cur, &obj->cli.stats.type);
	if (!strcmp(cur->name, "subtype"))
	    parseint(doc, cur, &obj->cli.stats.subtype);
	if (!strcmp(cur->name, "weight"))
	    parseint(doc, cur, &obj->cli.stats.weight);
	if (!strcmp(cur->name, "level"))
	    parseint(doc, cur, &obj->cli.stats.level);
	if (!strcmp(cur->name, "exp"))
	    parseint(doc, cur, &obj->cli.stats.exp);
	if (!strcmp(cur->name, "material"))
	    parseint(doc, cur, &obj->cli.stats.material);
	if (!strcmp(cur->name, "value"))
	    parseint(doc, cur, &obj->cli.stats.value);
	if (!strcmp(cur->name, "str"))
	    parsestat(doc, cur, &obj->cli.stats.Str);
	if (!strcmp(cur->name, "dex"))
	    parsestat(doc, cur, &obj->cli.stats.Dex);
	if (!strcmp(cur->name, "con"))
	    parsestat(doc, cur, &obj->cli.stats.Con);
	if (!strcmp(cur->name, "wis"))
	    parsestat(doc, cur, &obj->cli.stats.Wis);
	if (!strcmp(cur->name, "int"))
	    parsestat(doc, cur, &obj->cli.stats.Int);
	if (!strcmp(cur->name, "per"))
	    parsestat(doc, cur, &obj->cli.stats.Per);
	if (!strcmp(cur->name, "fame"))
	    parseint(doc, cur, &obj->cli.stats.fame);
	if (!strcmp(cur->name, "appr"))
	    parseint(doc, cur, &obj->cli.stats.appr);
	if (!strcmp(cur->name, "hp"))
	    parseint(doc, cur, &obj->cli.stats.hp);
	if (!strcmp(cur->name, "maxhp"))
	    parseint(doc, cur, &obj->cli.stats.maxhp);
	if (!strcmp(cur->name, "sp"))
	    parseint(doc, cur, &obj->cli.stats.sp);
	if (!strcmp(cur->name, "maxsp"))
	    parseint(doc, cur, &obj->cli.stats.maxsp);
	if (!strcmp(cur->name, "stm"))
	    parseint(doc, cur, &obj->cli.stats.stm);
	if (!strcmp(cur->name, "maxstm"))
	    parseint(doc, cur, &obj->cli.stats.maxstm);
	if (!strcmp(cur->name, "food"))
	    parseint(doc, cur, &obj->cli.stats.food);
	if (!strcmp(cur->name, "water"))
	    parseint(doc, cur, &obj->cli.stats.water);
	if (!strcmp(cur->name, "luck"))
	    parseint(doc, cur, &obj->cli.stats.luck);
	if (!strcmp(cur->name, "ac"))
	    parseint(doc, cur, &obj->cli.stats.ac);
	if (!strcmp(cur->name, "speed"))
	    parseint(doc, cur, &obj->cli.stats.speed);
	if (!strcmp(cur->name, "flags"))
	    parse_flags(doc, cur, obj);
	if (!strcmp(cur->name, "equip"))
	    parse_equipflags(doc, cur, obj);
	if (!strcmp(cur->name, "uid"))
	    parseuid(doc, cur, &(obj->cli.uid));

	cur = cur->next;
    }
    /* take care of the uid */
    if (mode == PO_ARCH) uid_clear(&(obj->cli.uid));
    /* no uid => generate a new one */
    else if (uid_isnull(&(obj->cli.uid))) uid_new(&(obj->cli.uid));
    
}

/**
   \brief parse the archetypes file
   \param filename filename to parse
   \param path path to find file
   \note takes the arch mutex
*/

void parse_archfile(char *filename, char *path)
{
    xmlDocPtr doc;
    xmlNodePtr cur;
    object_t *obj, *tmp;
    int placed;
    char file[MAX_PATH];

    sprintf(file, "%s/%s", path, filename);
    doc = xmlParseFile(file);
    if (doc == NULL)
	bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
	LOG(LOG_FATAL, "%s: empty document", file);
	xmlFreeDoc(doc);
	bailout();
    }
    if (strcmp(cur->name, "Archetypes")) {
	LOG(LOG_FATAL, "%s: root node != Archetypes", file);
	xmlFreeDoc(doc);
	bailout();
    }

    TAILQ_INIT(&arch);

    /* Second level is a list of obj, but be laxist */
    cur = cur->xmlChildrenNode;

    pthread_mutex_lock(&arch_mutex);
    while (cur != NULL) {
	if (!strcmp(cur->name, "Object")) {
	    obj = new_obj();
	    parse_object(doc, cur, obj, PO_ARCH);
	    /* place the objects in the arch list in sorted order */
	    placed = 0;
	    TAILQ_FOREACH(tmp, &arch, next) {
		if (strcmp(obj->arch, tmp->arch) <= 0) {
		    TAILQ_INSERT_BEFORE(tmp, obj, next);
		    placed++;
		    break;
		}
	    }
	    if (!placed) {
		TAILQ_INSERT_TAIL(&arch, obj, next);
	    }
	}
	cur = cur->next;
    }
    pthread_mutex_unlock(&arch_mutex);
    xmlFreeDoc(doc);
}
/* end of object loading routines */


/*================ ref count =================*/

/**
   \brief release memory used by obj
   \param obj object to free
*/
void free_obj(object_t *obj)
{
    if (obj->cli.name)
	free_string(obj->cli.name);
    if (obj->cli.name_pl)
	free_string(obj->cli.name_pl);
    if (obj->cli.model)
	free_string(obj->cli.model);
    if (obj->arch)
	free_string(obj->arch);
    if (obj->msg)
	free_string(obj->msg);
    free(obj);
}

void obj_incref(object_t *obj)
{
    pthread_mutex_lock(&objtable_mutex);
    obj->refcount++;
    pthread_mutex_unlock(&objtable_mutex);
}

void obj_decref(object_t *obj) 
{
    int n;

    pthread_mutex_lock(&objtable_mutex);
    n = --(obj->refcount);
    if (n<=0)
	free_obj(obj);
    pthread_mutex_unlock(&objtable_mutex);
}

/* test if object is destructed */
int obj_exists(object_t *obj)
{
    return !QUERY_FLAG(obj, FLAG_DESTROYED);
};

/* test if an object is an arch = uid is null*/
int obj_isarch(object_t *obj)
{
    return obj && uid_isnull(&(obj->cli.uid));
}

