/* $Id: outnet.c,v 1.8 2004/04/02 02:39:00 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 1998-2002 David Watson and Able Data Systems
    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <ctype.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <fcntl.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "shstr.h"
#include "utils.h"
#include "network.h"
#include "event.h"
#include "outnet.h"
#include "netparser.h"
#include "sutils.h"

/**
   \file outnet.c
   \brief Outgoing network functions
   \author Tim Rightnour

   These functions allow the server to make outgoing connections to other
   servers, or the metaserver.
*/

TAILQ_HEAD(, _servout) servout = TAILQ_HEAD_INITIALIZER(servout);
typedef void (*out_parsefunc_t)(servout_t *out, pargs_t *args);
out_parsefunc_t outparsefunc[] = {
    parse_outcommand_null
};

/**
   \brief process an outgoing network event
   \param event the event to process
   \return event status
*/

int process_outgoing_queue(event_t *event)
{
    if (event->out == NULL)
	return(1);

    return(outgoing_networker((servout_t *)event->out));
}

/**
   \brief generic null parse command
   \param out outgoing descriptor
   \param args command arguments
*/

void parse_outcommand_null(servout_t *out, pargs_t *args)
{
    return;
}

/**
   \brief make an outgoing network connection
   \param host hostname to connect to
   \param port port to connect to
*/

int connect_to(char *host, int port)
{
    struct sockaddr_in sa;
    struct hostent *hp;
    int s, rbuf;

    /* Resolve the host */
    bzero(&sa, sizeof(sa));

    if ((hp = gethostbyname(host)))
	bcopy(hp->h_addr, &sa.sin_addr, hp->h_length);
    else if (isdigit(*host))
	sa.sin_addr.s_addr = htonl(inet_network(host));

    sa.sin_family = AF_INET;
    sa.sin_port = htons(port);

    /* Create the socket */
    if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	perror("socket");
	return(-1);
    }

    /* Make the connection */
    if (connect(s, (struct sockaddr *)&sa, sizeof(sa)) < 0) {
	perror("connect");
	return(-1);
    }

    /* Set receive buffer to 8K */
    rbuf = 8192;

    if (setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char *)&rbuf, sizeof(rbuf)) < 0){
	perror("setsockopt()");
	return(-1);
    }

#ifdef NONBLOCK
    /* Turn on No Delay (or non-blocking) for descriptor. If it is not
     * set, then the process will sleep while waiting for data on a
     * read() call. This causes the whole process to stop all operations
     * until it receives data. Depending on what your client does, you
     * may or may not want the descriptor to be non-delayed. If the
     * client is an interactive client where data may be sent and
     * received at the same time, you may want it non-delayed. If the
     * client is more of a step-by-step model, then turning off No Delay
     * may be more desireable.
     */
    if (fcntl(s, F_SETFL, O_NDELAY) < 0) {
	perror("fcntl() non-block");
	return(-1);
    }
#endif /* #ifdef NONBLOCK */

    return(s);
}

/**
   \brief make a new outgoing connection
   \param host hostname to connect to
   \param port portnum to connect to
   \return a servout_t detailing the descriptor
   \note initializes the servout mutex.
*/
servout_t *new_outgoing_connection(char *host, int port)
{
    int s;
    servout_t *out;

    s = connect_to(host, port);
    if (s > 0) {
	if (TAILQ_EMPTY(&servout))
	    TAILQ_INIT(&servout);
	out = smalloc(servout_t);
	out->socket = s;
	init_mutex(&(out->mutex));
	return(out);
    }
    return(NULL);
}

/**
   \brief Read data from descriptor s and store it in in_buf
   \param out the outgoing descriptor to act upon
   \return number of bytes read, -1 for disconnect
*/

int process_out_input(servout_t *out)
{
    char buf[8193];
    int i;

    if ((i=read(out->socket, buf, 8192)) < 1)
	return(-1);

    buf[i] = '\0';

    if (!out->in_buf) {
	out->in_buf = (char *)malloc(strlen(buf)+1);
	strcpy(out->in_buf, buf);
    } else {
	out->in_buf = (char *)realloc(out->in_buf,
	    (strlen(out->in_buf)+strlen(buf))+2);
	strcat(out->in_buf, buf);
    }
    return(i);
}

/**
   \brief queue data to be transmitted to the outgoing socket
   \param out outgoing socket
   \param buf a C string to be sent

   You may want to use this function to store data in out_buf to be sent to
   the server.
*/

void do_write_out(servout_t *out, char *buf)
{
    if (!out->out_buf) {
	out->out_buf = (char *)malloc(strlen(buf)+1);
	strcpy(out->out_buf, buf);
    } else {
	out->out_buf = (char *)realloc(out->out_buf,
	    (strlen(out->out_buf)+strlen(buf))+2);
	strcat(out->out_buf, buf);
    }
}

/**
   \brief Send data that is stored in out_buf to server
   \param out outgoing socket
   \return how much data did we actually send. -1 for disconnect.
*/

int process_out_output(servout_t *out)
{
    int i=0, buflen, cpos;
    char *p;

    if (out->out_buf) {
        p = out->out_buf;
        buflen = strlen(p);
        cpos = 0;
        while (cpos < buflen) {
            i = write(out->socket, p, buflen-cpos);
            if (i < 0) {
		if (errno == EAGAIN)
		    return(0);
		else
		    return(-1);
            } else {
                cpos += i;
                p += i;
            }
        }
        free(out->out_buf);
        out->out_buf=0;
        return(cpos);
    } else
        return(0);
}

/**
   \brief This is where the commands will be parsed
   \param out outgoing connection
*/

void out_process_command(servout_t *out)
{
    char **words, *buf, *p, *pp;
    int numwords, command, i;
    pargs_t *args=NULL;

    /*printf("Client Got: \n::%s::\n :END\n", in_buf);*/

    p = out->in_buf;
    pp = NULL;
    for (buf=out->in_buf; buf != NULL && strlen(buf) > 0; ) {
	pp = p;
	p = strchr(buf, '\r');
	if (p == NULL)
	    goto next;
	*p = '\0';
	p++;
	words = parse_netcommand(buf, &numwords);
	if (words == NULL)
	    goto next;
	args = parse_command(words, &command, numwords);
	if (command < 1 || command > OUTCOMMAND_MAX)
	    goto next;
	outparsefunc[command](out, args);
    next:
	if (args) {
	    for (i=0; args[i].cword != -1; i++)
		if (args[i].type == PTCHAR)
		    free_string(args[i].arg.c);
	    free(args);
	    args = NULL;
	}
	buf = p;
    }
    if (p == NULL && pp != NULL) {
	buf = strdup(pp);
	strcpy(out->in_buf, buf);
	free(buf);
    } else {
	free(out->in_buf);
	out->in_buf=0;
    }
}

/**
   \brief process the outgoing network connection
   \param out outgoing connection
   \todo perhaps this should be a thread?
*/

int outgoing_networker(servout_t *out)
{
    fd_set input, output, errfds;
    struct timeval tval;

    FD_ZERO(&input);
    FD_ZERO(&output);
    FD_ZERO(&errfds);

#ifdef NONBLOCK
    /*
     * We will only be using input and output file descriptor sets
     * in a non-delayed type client.
     */
    FD_SET(out->socket, &input);
    FD_SET(out->socket, &output);
#endif /* #ifdef NONBLOCK */
    FD_SET(out->socket, &errfds);

    tval.tv_sec = 0;
    tval.tv_usec = 1000;

    if (select(out->socket + 1, &input, &output, &errfds, &tval) < 0) {
	perror("select()");
	return(1);
    }

    if (FD_ISSET(out->socket, &errfds)) {
	LOG(LOG_ERROR, "Error with connection %d.", out->socket);
	close(out->socket);
	return(1);
    }

#ifdef NONBLOCK
    /*
     * We only check if FD_ISSET on a non-delayed type client. You
     * may need to switch the input and output around if the server
     * expects something from the client before it sends anything,
     * rather than sending something before it waits for input.
     */
    if (FD_ISSET(out->socket, &input))
#endif /* #ifdef NONBLOCK */
	if (process_out_input(out) < 0) {
	    close(out->socket);
	    return(1);
	}

#ifdef NONBLOCK
    if (FD_ISSET(out->socket, &output))
#endif /* #ifdef NONBLOCK */
	if (process_out_output(out) < 0) {
	    close(out->socket);
	    return(1);
	}

    if (out->in_buf)
	out_process_command(out);
    return(0);
}
