/* $Id: physics.c,v 1.26 2004/04/02 02:39:00 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/time.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "object.h"
#include "map.h"
#include "shstr.h"
#include "utils.h"
#include "models.h"
#include "event.h"
#include "npc.h"
#include "player.h"
#include "timers.h"
#include "matrix.h"
#include "collision.h"
#include "physics.h"
#include "sutils.h"

/**
   \file physics.c
   \brief Physics routines
   \author Tim Rightnour
*/

extern pthread_mutex_t maps_mutex;
extern TAILQ_HEAD(, _mapgroup_t) maps;

/**
   \brief calculate the walking speed of an obj
   \param obj object to calculate
   \return the speed in points
*/

float calc_obj_velocity(object_t *obj)
{
    if (obj->cli.state < STATE_WALK)
	return(0.0);
    if (obj->cli.state == STATE_WALK)
	return((obj->cli.stats.speed/(float)MAXSPEED) * 0.3);
    if (obj->cli.state == STATE_RUN)
	return((obj->cli.stats.speed/(float)MAXSPEED) * 0.6);
    if (obj->cli.state == STATE_SPRINT)
	return((obj->cli.stats.speed/(float)MAXSPEED) * 0.9);
    /* XXX */
    return(0.1);
}

/**
   \brief calculate how far an object will rotate when tipping
   \param s1 first box corner on tip fulcrum
   \param s2 second box corner on tip fulcrum
   \param radius radius of the circle we create to map the falling arc
   \param dist how far do we want to fall
   \param obj object to fall
   \param box the bounding box for obj
   \param xorz which direction to fall? xorz == 1 for x, 0 for z.
   \return the angle of the fall
*/

static float calc_fall_rotation(int s1, int s2, float radius, float dist,
				 object_t *obj, bbox_t *box, int xorz)
{
    float m, neg, *lx, *ly, side1, side2, facing;
    int nrof, j;

    if (xorz) {
	side1 = box->side[s1].x;
	side2 = box->side[s2].x;
	facing = obj->cli.facing.x;
    } else {
	side1 = box->side[s1].z;
	side2 = box->side[s2].z;
	facing = obj->cli.facing.z;
    }

    if (side1 < side2 || dist < 0.0) {
	m = side1 - radius;
	neg = -SCALE_XZ;
    } else {
	m = side1 + radius;
	neg = SCALE_XZ;
    }
    nrof = (int)ceil(radius/SCALE_XZ) + 2;
    lx = malloc(sizeof(float) * nrof);
    ly = malloc(sizeof(float) * nrof);
    lx[0] = side1;
    ly[0] = box->side[s1].y;
    for (j=1; j < nrof; j++) {
	lx[j] = side1 + neg * j;
	lx[j] = rint(lx[j] / SCALE_XZ) * neg;
	if (((lx[j] > m) && (neg > 0.0)) ||
	    ((lx[j] < m) && (neg < 0.0)))
	    lx[j] = m;
	if (xorz) {
	    ly[j] =groundheight_at_mapxz(obj->maphead, lx[j], box->side[s1].z);
	} else {
	    ly[j] =groundheight_at_mapxz(obj->maphead, box->side[s1].x, lx[j]);
	}
    }
    m = farthest_fall_rotation(nrof, lx, ly, radius);
    if (m < 180.0 && ly[0] > ly[nrof-1])
	m = 360.0 - m;
    if (dist < 0.0 && facing > 180.0) {
	if (facing - fabs(dist) < m) {
	    dist = m;
	} else {
	    dist = 0.0 - add_euler(facing, dist);
	}
    } else {
	if (dist > 360.0)
	    dist = m;
	else if (dist < 0.0 && add_euler(facing, dist) < m)
	    dist = -m;
	else if (dist > 0.0 && add_euler(facing, dist) > m)
	    dist = m;
	else if (dist > 0.0)
	    dist = add_euler(facing, dist);
	else
	    dist = 0.0 - add_euler(facing, dist);
    }
    free(lx);
    free(ly);
    return(dist);
}

/**
   \brief is the object unstable in it's current orientation?
   \param sog sides on ground
   \param side the 4 points that make up the side closest to the ground
   \param box the bounding box
   \param obj the object to calculate
   \return An instability factor.

   If the object's center of gravity is beyond the X and Z planes of the side
   that is on the ground, then the object is not stable, and returns at least
   a 4.  If the object's height/x ratio is more than 10:1, then we return
   an additional 1.  if the height/z ratio is more than 10:1, then we return
   an additional 2.

   A 0 means the object is stable.  A 5 or 1 result means unstable in the Z
   axis and a 6 is unstable in X.  A 4 or 7 is unstable in all directions.
*/

static int calc_tip_unstability(int sog, int side[4], bbox_t *box,
				object_t *obj)
{
    int i, ii, jj, nside[4], unstable;
    float yy[4], xx[4], xr, zr, yr;

    unstable = 0;

    /* sort the sides, and flip the last two.  don't ask. */
    for (i = 0; i < 4; i++)
	nside[i] = side[i];
    for (ii = 0; ii < 4; ii++)
	for (i = ii; i < 4; i++)
	    if (nside[ii] > nside[i]) {
		jj = nside[ii];
		nside[ii] = nside[i];
		nside[i] = jj;
	    }
    jj = nside[3];
    nside[3] = nside[2];
    nside[2] = jj;

    for (i = 0; i < 4; i++) {
	xx[i] = box->side[nside[i]].x;
	yy[i] = box->side[nside[i]].z;
    }
    if (!pnpoly(4, xx, yy, box->cog.x, box->cog.z))
	unstable += 4;

    if ((obj->cli.facing.x/90.0 == 0.0 && obj->cli.facing.z/90.0 == 0.0 &&
	 unstable == 0) || unstable == 0) {
	xr = fabs(box->side[side[0]].x - box->side[side[1]].x);
	if (xr == 0.0)
	    xr = fabs(box->side[side[0]].x - box->side[side[2]].x);
	zr = fabs(box->side[side[0]].z - box->side[side[1]].z);
	if (zr == 0.0)
	    zr = fabs(box->side[side[0]].z - box->side[side[2]].z);
	yr = fabs(box->cog.y - box->side[i].y);

	if (yr/xr > 10.0)
	    unstable += 1;
	if (yr/zr > 10.0)
	    unstable += 2;
    }
    return(unstable);
}

/**
   \brief calculate the trajectory of an object based on physics
   \param obj the object to calculate
   \return velocity
   \todo a walking player will not transition to state falling until he stops
   \todo rolling in only one axis is not yet implemented

   Calculates the facing and direction of the object, and sets those in the
   object as well.
*/

float calc_obj_trajectory(object_t *obj)
{
    bbox_t *box;
    triangle_t *tri;
    vertex_t newp, angle;
    int i, t, ss;
    float diff=1.0, tlp, tresist;


    if (QUERY_FLAG(obj, FLAG_ANCHOR))
	return(0.0);

/*
    printf("dxyz %f %f %f\n", obj->cli.direction.x, obj->cli.direction.y,
	   obj->cli.direction.z);
    printf("fxyz %f %f %f\n", obj->cli.facing.x, obj->cli.facing.y,
	   obj->cli.facing.z);
    printf("cxyz %f %f %f\n", obj->cli.coord.x, obj->cli.coord.y,
	   obj->cli.coord.z);
*/
    /* quick out the players making normal movement */
    if (obj->cli.state >= STATE_WALK && obj->cli.state <= STATE_SPRINT) {
	obj->cli.direction.x = 0.0;
	if (obj->cli.state == STATE_WALK)
	    return((obj->cli.stats.speed/(float)MAXSPEED) * 3.0);
	if (obj->cli.state == STATE_RUN)
	    return((obj->cli.stats.speed/(float)MAXSPEED) * 6.0);
	if (obj->cli.state == STATE_SPRINT)
	    return((obj->cli.stats.speed/(float)MAXSPEED) * 9.0);
    }

    tlp = groundheight_at_mapcoord(obj->maphead, obj->cli.coord);

    box = create_bbox(obj->bmin, obj->bmax, obj->center, obj->cli.coord);
    newp.x = obj->cli.coord.x + obj->coff.x;
    newp.y = obj->cli.coord.y + obj->coff.y;
    newp.z = obj->cli.coord.z + obj->coff.z;
    rotate_bbox(box, obj->cli.facing, newp);
    i = find_lowest_bbox_point(box);
    diff = bbox_groundheight_diff(box, obj->maphead, obj->cli.coord);

    /* has the object fallen too far? correct */
    if (diff < 0.0 && !QUERY_FLAG(obj, FLAG_ANCHOR) &&
	!(obj->cli.state >= STATE_WALK && obj->cli.state <= STATE_SPRINT)) {
	obj->cli.coord.y -= diff;
	obj->cli.state = STATE_IDLE;
	diff = 0.0;
	free(box);
	/* recalculate the box, otherwise wierd things happen */
	box = create_bbox(obj->bmin, obj->bmax, obj->center, obj->cli.coord);
	newp.x = obj->cli.coord.x + obj->coff.x;
	newp.y = obj->cli.coord.y + obj->coff.y;
	newp.z = obj->cli.coord.z + obj->coff.z;
	rotate_bbox(box, obj->cli.facing, newp);
	i = find_lowest_bbox_point(box);
    }

    if (diff <= 0.01 && obj->cli.state == STATE_FALLING) {
	obj->cli.state = STATE_IDLE;
	if (obj->cli.velocity > 0.0 && obj->cli.direction.x != 0.0) {
	    /* we hit the ground */
	    obj->cli.velocity = 0.0;
	    obj->cli.direction.x = 0.0;
	}
    }

    if (diff > 0.2) {
	float carea, hcomp, vcomp, xresist, yresist, xacc, yacc, dist;
	/* we must calculate the trajectory and the velocity */

	hcomp = (cos(DEGREES_TO_RADIANS(obj->cli.direction.x)) *
		 obj->cli.velocity);
	vcomp = (sin(DEGREES_TO_RADIANS(obj->cli.direction.x)) *
		 obj->cli.velocity);

	carea = dist_between(obj->bmin, obj->bmax) / POINTS_PER_METER;
	carea *= carea;
	tresist = 0.5 * DRAG_COEF * AIR_DENSE * carea *
	    (obj->cli.velocity/POINTS_PER_METER) *
	    (obj->cli.velocity/POINTS_PER_METER);
	if (tresist != 0.0) {
	    xresist = tresist * (-hcomp / obj->cli.velocity);
	    yresist = tresist * (-vcomp / obj->cli.velocity);

	    xacc = xresist / (float)obj->cli.stats.weight / 1000.0;
	    yacc = GRAVITY +
		(yresist / (float)obj->cli.stats.weight / 1000.0);
	} else {
	    xacc = 0.0;
	    yacc = GRAVITY;
	}
	hcomp = hcomp + (xacc*TIMESLICE);
	vcomp = vcomp + (yacc*TIMESLICE);
	if (vcomp < 0.0 && (fabs(vcomp)/TIMESLICE) > diff)
	    vcomp = -diff/TIMESLICE;
	angle.x = angle.z = 0.0;
	angle.y = obj->cli.direction.y;
	new_coord(angle, hcomp*TIMESLICE, obj->cli.coord, &newp);
	newp.y = obj->cli.coord.y + vcomp*TIMESLICE;
	newp.x += obj->cli.coord.x;
	newp.z += obj->cli.coord.z;
	dist = dist_between(newp, obj->cli.coord);
	angle.x = RADIANS_TO_DEGREES(atan2(newp.y-obj->cli.coord.y,
		    newp.x-obj->cli.coord.x + newp.z-obj->cli.coord.z));
	obj->cli.direction.x = add_euler(0.0, angle.x);
/*	obj->cli.direction.x = add_euler(angle.x, -90.0);
	obj->cli.direction.z = add_euler(angle.z, -90.0);*/
	obj->cli.velocity = sqrt((hcomp * hcomp) + (vcomp * vcomp));
	obj->cli.state = STATE_FALLING;
	free(box);
	return(dist);
    }

    tri = find_closest_hf_triangle(obj->maphead->map->elevation,
				       obj->cli.coord);
    t = find_lowest_point_tri(tri);

    /* this gives us the angle of the ground at this location */
    newp = obj->cli.coord;
    /* if the three sides are the same, the ground is level, adjust tlp. */
    if (tri->side[0].y == tri->side[1].y && tri->side[0].y == tri->side[2].y)
	tlp = tri->side[0].y;
    if (tlp < newp.y)
	newp.y = tlp;
    ss = 0;

    if (floor(newp.x * 10.0) == floor(tri->side[t].x * 10.0))
	ss++;
    if (floor(newp.z * 10.0) == floor(tri->side[t].z * 10.0))
	ss++;
    if (floor(newp.y * 10.0) == floor(tri->side[t].y * 10.0))
	ss++;
    if (ss > 1) {
	find_center_tri(tri, &newp);
	newp.y = groundheight_at_mapcoord(obj->maphead, newp);
    }

    angle_between(newp, tri->side[t], &angle);
/*    if (angle.x > 180.0)
      angle_between(tri->side[t], newp, &angle, 0); */

    /* swap the angles if x was zero and z wasn't */
    if (newp.x - tri->side[t].x == 0.0 &&
	newp.z - tri->side[t].z != 0.0) {
	float a;

	a = angle.x;
	angle.x = angle.z;
	angle.z = a;
    }
/*    if (angle.x > 180.0)
      angle.x = 360.0 - angle.x; */
    if (angle.x == 360.0)
	angle.x = 0.0;

    /* step 1, check for rolling */
    if ((QUERY_FLAG(obj, FLAG_ROLL) || QUERY_FLAG(obj, FLAG_ROLLX) ||
	QUERY_FLAG(obj, FLAG_ROLLZ)) && ((angle.x > 0.0 && angle.x < 90.0)
					 || obj->cli.velocity > 0.0)) {
	float tneeded, ares, torque, radius, circum, carea, ovel, an;
	vector1_t *v;
	vertex_t dest;

	ovel = obj->cli.velocity;
	tneeded = (float)obj->cli.stats.weight / 1000.0 * ROLL_FRICTION_COEF;
	if (obj->cli.velocity == 0.0) /* inertia */
	    tneeded += (float)obj->cli.stats.weight / 1000.0;
	radius = dist_between(box->side[0], box->side[7])/(2.0*POINTS_PER_METER);
	carea = M_PI * radius * radius;
	ares = 0.5 * DRAG_COEF * AIR_DENSE * carea *
	    (obj->cli.velocity/POINTS_PER_METER) *
	    (obj->cli.velocity/POINTS_PER_METER) * TIMESLICE;
	/* Acceleration (m/sec^2)= 9.67 m/sec^2)* sin(0 ) */

	torque = 0.0 - GRAVITY * sin(DEGREES_TO_RADIANS(angle.x)) * TIMESLICE;
	v = vector_1d_cross(obj->cli.direction.y,
			    obj->cli.velocity*obj->cli.stats.weight,
			    angle.y, torque*obj->cli.stats.weight);
	v->m /= (float)obj->cli.stats.weight;
	v->a = add_euler(0.0, v->a);
	obj->cli.velocity = v->m;
	if (obj->cli.velocity > (tneeded + ares)) {
            obj->cli.velocity -= tneeded + ares;
        } else {
            obj->cli.velocity = 0.0;
        }
	if (obj->cli.velocity > 0.0) {
	    /* we roll */
	    circum = M_PI * radius * 2.0;
		
	    obj->cli.facing.y = v->a;
	    obj->cli.direction.y = v->a;
	    obj->cli.facing.x = add_euler(obj->cli.facing.x,
		 ((obj->cli.velocity*TIMESLICE)/circum) / (circum/360.0));
	    new_coord(obj->cli.direction, obj->cli.velocity, obj->cli.coord,
		      &dest);
	    dest.y = groundheight_at_mapcoord(obj->maphead, dest);
	    if (dest.y > obj->cli.coord.y) {
		/* we need to douse the velocity */
		an = RADIANS_TO_DEGREES(atan2(dest.y-obj->cli.coord.y,
					      obj->cli.velocity));
		obj->cli.velocity += GRAVITY * sin(DEGREES_TO_RADIANS(an)) * TIMESLICE;
		if (obj->cli.velocity < 0.0)
		    obj->cli.velocity = 0.0;
	    }
	    free(box);
	    free(tri);
	    free(v);
	    if (ovel == 0.0 && obj->cli.velocity == 0.0) {
		CLEAR_FLAG(obj, FLAG_STABLE_SLIDE);
		CLEAR_FLAG(obj, FLAG_STABLE_TIP);
		SET_FLAG(obj, FLAG_STABLE_ROLL);
	    }
	    return(obj->cli.velocity);
	}
	free(v);
    } else if (!QUERY_FLAG(obj, FLAG_NOTIP) &&
	       !QUERY_FLAG(obj, FLAG_STABLE_TIP)) {
	int side[4], sog, unstable;
	vertex_t oangle;

	/* lets tip some objects over */

	/* should tip first, level if no tip, not other way round */

	sog = find_lowest_bbox_side(box, obj->maphead, side);

	unstable = calc_tip_unstability(sog, side, box, obj);

	if (unstable > 0) {
	    float za=0.0, xa=0.0, zd, xd, torque, radius, cir;
	    vertex_t nnp, dangle;
	    int j, k, nrof;

	    /* the object's center is beyond it's lower bbox points,
	       thus it will tip.  which direction though? */

	    if (unstable > 3) {
		za = obj->cli.facing.z;
		xa = obj->cli.facing.x;
	    }
	    if (xa == 0.0 && za == 0.0) {
		if (unstable == 1 || unstable == 3) {
		    if (obj->cli.facing.x > 180.0) {
			xa = -1.0;
		    } else if (obj->cli.facing.x > 0.0) {
			xa = 1.0;
		    } else {
			xa = (float)rndm(-1, 1);
		    }
		    xa = add_euler(obj->cli.facing.x, xa);
		}
		if (unstable == 2 || unstable == 3) {
		    if (obj->cli.facing.z > 180.0) {
			za = -1.0;
		    } else if (obj->cli.facing.z > 0.0) {
			za = 1.0;
		    } else {
			za = (float)rndm(-1, 1);
		    }
		    za = add_euler(obj->cli.facing.z, za);
		}
		if (unstable == 4) {
		    xa = add_euler(0.0, (float)rndm(-1, 1));
		    za = add_euler(0.0, (float)rndm(-1, 1));
		}
	    }

	    /* k is lowest point
	       opposite point is k +/- 4 */
	    k = find_lowest_bbox_point(box);
	    /* if the lowest point is not on the ground, pick the point
	       closest to the ground */
	    for (nrof=0, j=0; nrof < sog; nrof++) {
		if (k != side[nrof])
		    j++;
	    }
	    if (j == sog)
		k = side[0];

	    if (k > 3)
		j = k - 4;
	    else
		j = k + 4;
	    /* if the opposite point is on the ground, find the other
	       opposite point */
	    for (nrof=0; nrof < 4; nrof++) {
		if (j == side[nrof]) {
		    if (j > 3)
			j -= 2;
		    else
			j += 2;
		    break;
		}
	    }

	    radius = dist_between(box->side[k], box->side[j]);

	    cir = M_PI * radius * 2.0;
	    torque = -GRAVITY * sin(DEGREES_TO_RADIANS(za)) * TIMESLICE;
	    zd = torque/cir*360.0;
	    torque = -GRAVITY * sin(DEGREES_TO_RADIANS(xa)) * TIMESLICE;
	    xd = torque/cir*360.0;

	    xd = calc_fall_rotation(k, j, radius, xd, obj, box, 1);
	    zd = calc_fall_rotation(k, j, radius, zd, obj, box, 0);

	    dangle.x = diff_euler(xd, obj->cli.facing.x);
	    if (xd < obj->cli.facing.x)
		dangle.x = -dangle.x;
	    dangle.z = diff_euler(zd, obj->cli.facing.z);
	    if (zd < obj->cli.facing.z)
		dangle.z = -dangle.z;
	    dangle.y = 0.0;

	    if ((floor(dangle.z)*10.0 == 0.0 && floor(dangle.x)*10.0 == 0.0) ||
		(floor(obj->cli.facing.x)*10.0 == floor(fabs(xd))*10.0 &&
		 floor(obj->cli.facing.z)*10.0 == floor(fabs(zd))*10.0)) {
		SET_FLAG(obj, FLAG_STABLE_TIP);
	    } else {
		obj->cli.facing.x = fabs(xd);
		obj->cli.facing.z = fabs(zd);

		obj->cli.state = STATE_FALLING;
		obj->cli.coord.x += obj->coff.x;
		obj->cli.coord.y += obj->coff.y;
		obj->cli.coord.z += obj->coff.z;
		rotate_point_on_center(obj->cli.coord, box->side[k],
				       dangle, &nnp);
		nnp.x -= obj->coff.x;
		nnp.y -= obj->coff.y;
		nnp.z -= obj->coff.z;
		obj->cli.coord = nnp;
	    }

	} else if (sog < 3) {
	    vertex_t gangle, dangle, vangle, mangle;

	    /* the center of the box must be dropped, first */
	    oangle.y = oangle.x = oangle.z = 0.0;
	    if (obj->cli.facing.x < 180.0 && angle.x > 180.0)
		oangle.x = 360.0;
	    if (obj->cli.facing.z < 180.0 && angle.z > 180.0)
		oangle.z = 360.0;
	    if (sog < 2) {
		oangle.x = -angle.x;
		oangle.z = -angle.z;
	    } else if (sog == 2) {	
		dangle = box->side[side[3]];
		dangle.y = groundheight_at_mapcoord(obj->maphead, dangle);
		angle_between(box->side[side[0]], dangle, &mangle);
		if (diff_euler(angle.x, obj->cli.facing.x) >
		    diff_euler(mangle.x, obj->cli.facing.x))
		    angle.x = mangle.x;
		if (diff_euler(angle.z, obj->cli.facing.z) >
		    diff_euler(mangle.z, obj->cli.facing.z))
		    angle.z = mangle.z;

		if ((box->side[side[0]].x == box->side[side[2]].x &&
		     box->side[side[1]].x == box->side[side[3]].x) ||
		    (box->side[side[1]].x == box->side[side[2]].x &&
		     box->side[side[0]].x == box->side[side[3]].x))
		    oangle.x = -angle.x;
		else if ((box->side[side[0]].z == box->side[side[2]].z &&
		     box->side[side[1]].z == box->side[side[3]].z) ||
		    (box->side[side[1]].z == box->side[side[2]].z &&
		     box->side[side[0]].z == box->side[side[3]].z))
		    oangle.z = -angle.z;
		/* if neither is true, its a fairly complex process to guess
		   properly, so we just wing it for speed's sake */
		else {
		    oangle.x = -angle.x;
		    oangle.z = -angle.z;
		}
	    }
	    gangle.x = gangle.z = 0.0;
	    if (obj->cli.facing.z > 90.0 && obj->cli.facing.z < 180.0)
		gangle.z = 90.0;
	    else if (obj->cli.facing.z > 180.0 && obj->cli.facing.z < 270.0)
		gangle.z = 180.0;
	    else if (obj->cli.facing.z > 270.0 && obj->cli.facing.z < 360.0)
		gangle.z = 270.0;
	    if (obj->cli.facing.x > 90.0 && obj->cli.facing.x < 180.0)
		gangle.x = 90.0;
	    else if (obj->cli.facing.x > 180.0 && obj->cli.facing.x < 270.0)
		gangle.x = 180.0;
	    else if (obj->cli.facing.x > 270.0 && obj->cli.facing.x < 360.0)
		gangle.x = 270.0;

	    /* record the old angle */
	    vangle = obj->cli.facing;
	    obj->cli.facing.x = add_euler(gangle.x, oangle.x);
	    obj->cli.facing.z = add_euler(gangle.z, oangle.z);

	    dangle.x = diff_euler(vangle.x, obj->cli.facing.x);
	    dangle.z = diff_euler(vangle.z, obj->cli.facing.z);
	    dangle.y = 0.0;

	    if (floor(dangle.x)*10.0 == 0.0 && floor(dangle.z)*10.0 == 0.0) {
		/* we have tried to move the object, but it did not move.
		   therefore we place it at rest, to prevent further waste
		   of CPU */
		SET_FLAG(obj, FLAG_STABLE_TIP);
		CLEAR_FLAG(obj, FLAG_STABLE_SLIDE);
	    } else {
		rotate_point_on_center(obj->cli.coord, box->side[side[0]],
				       dangle, &newp);
		obj->cli.coord = newp;
	    }
	} else if (sog > 2 && !unstable) {
	    SET_FLAG(obj, FLAG_STABLE_TIP);
	    CLEAR_FLAG(obj, FLAG_STABLE_SLIDE);
	}
    } else if (!QUERY_FLAG(obj, FLAG_ANCHOR) &&
	       !QUERY_FLAG(obj, FLAG_STABLE_SLIDE)) {
	/* sliding, should be just like rolling, except the friction should
	 be way higher. */
	float tneeded, ares, torque, carea, ovel, oppangle;
	vector1_t *v;

	tneeded = (float)obj->cli.stats.weight / 1000.0 * FRICTION_COEF;
	tneeded *= cos(DEGREES_TO_RADIANS(angle.x));
	carea = dist_between(obj->bmax, obj->bmin) / POINTS_PER_METER;
	carea *= carea;
	ares = 0.5 * DRAG_COEF * AIR_DENSE * carea *
	    (obj->cli.velocity/POINTS_PER_METER) *
	    (obj->cli.velocity/POINTS_PER_METER) * TIMESLICE;
	ovel = obj->cli.velocity;
	/* Acceleration (m/sec^2)= 9.67 m/sec^2)* sin(0 ) */

	oppangle = add_euler(angle.y, 180.0);

	torque = fabs(GRAVITY * sin(DEGREES_TO_RADIANS(angle.x))) * TIMESLICE;
	v = vector_1d_cross(obj->cli.direction.y,
			    obj->cli.velocity*obj->cli.stats.weight,
			    oppangle, torque*obj->cli.stats.weight);
	v->m /= (float)obj->cli.stats.weight;
	v->a = add_euler(0.0, v->a);
	obj->cli.velocity = v->m;

	if (obj->cli.velocity > (tneeded + ares)) {
	    obj->cli.velocity -= tneeded + ares;
	} else {
	    obj->cli.velocity = 0.0;
	}
	obj->cli.facing.y = v->a;
	obj->cli.direction.y = v->a;
	/* if we weren't moving, and we aren't moving, we are done moving */
	if (ovel == 0.0 && obj->cli.velocity == 0.0) {
	    SET_FLAG(obj, FLAG_STABLE_SLIDE);
	}
	free(box);
	free(tri);
	free(v);
	return(obj->cli.velocity);
    }

    free(tri);
    free(box);
    return(0.0);
}

/**
  \brief attempt to move an object
  \return success/fail
  \param obj object to attempt to move
  \warning calls dtTest()

  Move the object by calc_dis. if we succeed leave them there and
  return 1.  if not, put them back where we found them.
*/

int attempt_move_obj(object_t *obj)
{
    vertex_t ncoord;
    float distance;

    /* Step 1. Assume the current rotation and location are correct.  Move. */
    distance = calc_obj_trajectory(obj);
    if (distance == 0.0) {
	obj->cli.velocity = 0.0;
	return(0);
    }
    /*printf("DIST = %s %f\n", obj->cli.name, distance);*/

    new_coord(obj->cli.direction, distance, obj->cli.coord, &ncoord);
    ncoord.x += obj->cli.coord.x;
    ncoord.z += obj->cli.coord.z;
    if (obj->cli.state != STATE_FALLING) {
	ncoord.y = groundheight_at_mapcoord(obj->maphead, ncoord);
    } else {
	ncoord.y += obj->cli.coord.y;
    }

    /*printf("WAS AT %f %f %f\n", obj->cli.coord.x, obj->cli.coord.y, obj->cli.coord.z);*/
    /* are we leaving the map boundaries? */
    if (coord_out_of_mapgroup(obj->maphead, ncoord))
	return(0);

    /*printf("OBJ AT %f %f %f\n", ncoord.x, ncoord.y, ncoord.z);*/

    place_obj_and_test(obj, ncoord);

    /* if we collide, we should attempt to move as close as possible to the
       offending object without colliding */

    if (obj->collision) {
	int moveok=0;
	float zwidth, dis, ghy;
	vertex_t newcoord, moved;

	obj->collision = 0;
	zwidth = fabs(obj->bmin.z - obj->bmax.z);
	if (zwidth > distance)
	    zwidth = distance/25.0;
	newcoord = obj->cli.coord;
	for (dis=zwidth; dis <= distance; dis+=zwidth) {
	    new_coord(obj->cli.direction, dis, newcoord, &moved);
	    newcoord.x += moved.x;
	    newcoord.y += moved.y;
	    newcoord.z += moved.z;
	    if (obj->cli.state != STATE_FALLING) {
		ghy = groundheight_at_mapcoord(obj->maphead, newcoord);
		if (ghy > newcoord.y)
		    moved.y += ghy - newcoord.y;
	    }
	    if (place_obj_and_test(obj, newcoord) > 0) {
		/*printf("Distance FAIL: %f\n", dis);*/
		dis -= zwidth;
		break;
	    } else {
		moveok++;
		/*printf("Distance OK: %f\n", dis);*/
	    }
	}
	if (moveok) {
	    /*printf("Distance TO MOVE: %f\n", dis);*/
	    new_coord(obj->cli.direction, dis, obj->cli.coord, &moved);
	    obj->cli.coord.x += moved.x;
	    obj->cli.coord.y += moved.y;
	    obj->cli.coord.z += moved.z;
	    if (obj->cli.state != STATE_FALLING) {
		ghy = groundheight_at_mapcoord(obj->maphead, obj->cli.coord);
		if (ghy > newcoord.y)
		    moved.y += ghy - obj->cli.coord.y;
	    }
	    obj->collision = 0;
	} else
	    obj->collision = 1;
    }

    if (obj->collision) {
	obj->collision = 0;
	return(0);
    }
    obj->cli.prevc = obj->cli.coord;
    obj->cli.coord = ncoord;
    return(1);
}
