/* $Id: player.c,v 1.66 2004/04/02 02:39:00 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#include <math.h>
#include <sys/param.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "object.h"
#include "map.h"
#include "shstr.h"
#include "utils.h"
#include "models.h"
#include "tod.h"
#include "xmlparse.h"
#include "event.h"
#include "player.h"
#include "network.h"
#include "matrix.h"
#include "netparser.h"
#include "ninterface.h"
#include "collision.h"
#include "stattables.h"
#include "physics.h"
#include "servdata.h"
#include "sutils.h"

extern persist_data_t persist;
extern npersist_data_t nonpersist;
extern pthread_mutex_t maps_mutex;
extern TAILQ_HEAD(, _mapgroup_t) maps;
extern pthread_mutex_t players_mutex;

TAILQ_HEAD(, _player_t) players = TAILQ_HEAD_INITIALIZER(players);

/** \file player.c
    \brief Player related functions

    All functions that deal directly with players, including the events
    that specifically involve them, and are bound to them.
*/

/**
   \brief allocate a new player struct
*/

player_t *new_player(void)
{
    player_t *pl;

    pl = smalloc(player_t);
    return(pl);
}

/**
   \brief find a player knowledge structure
   \param pl player to look in
   \param num object number to look for
   \return NULL if none, knowledge structure if one found
*/

pl_known_t *find_pl_knowledge(player_t *pl,struct _object_t*obj)
{
    pl_known_t *known;

    TAILQ_FOREACH(known, &pl->known, next)
	if (uid_eq(&obj->cli.uid,&known->uid))
	    return(known);
    return(NULL);
}

/**
   \brief tell all the players that an obj no longer exists on the map
   \param event event that called us
   \return event status (allways 1)

   This does not necc. mean that the object was taken away from the map
   entirely.  However, this object is no longer visible to other players
   on the map.  It may have been picked up by a player or NPC.
*/

int obj_removed_from_map(event_t *event)
{
    object_t *obj;
    player_t *pl;
    pl_known_t *known;

    obj = event->initiator;

    pthread_mutex_lock(&players_mutex);
    TAILQ_FOREACH(pl, &players, next) {
	known = find_pl_knowledge(pl, obj);
	if (known != NULL) {
	    TAILQ_REMOVE(&pl->known, known, next);
	    tell_rmobj(pl, obj);
	    free_string(known->name);
	    free_string(known->name_pl);
	    free_string(known->model);
	    free(known);
	}
    }
    pthread_mutex_unlock(&players_mutex);
    return(1);
}

/**
   \brief tell all the players that something performed an action
   \param event event that called us
   \return event status (allways 1)
*/

int obj_eact(event_t *event)
{
    object_t *obj;
    player_t *pl;
    pl_known_t *known;

    obj = event->initiator;  /* the object doing the action */

    pthread_mutex_lock(&players_mutex);
    TAILQ_FOREACH(pl, &players, next) {
	known = find_pl_knowledge(pl, obj);
	if (known != NULL || obj->pl == pl) {
	    tell_eact(pl, obj, event->action);
	}
    }
    pthread_mutex_unlock(&players_mutex);
    return(1);
}

/**
   \brief tell all the players that something wore something
   \param event event that called us
   \return event status (allways 1)
*/

int obj_wore_obj(event_t *event)
{
    object_t *obj, *eq;
    player_t *pl;
    pl_known_t *known;

    obj = event->initiator;  /* the object doing the wearing */
    eq = evarg_getobj(event,0); /* the first obj in involved is the EQ */

    pthread_mutex_lock(&players_mutex);
    TAILQ_FOREACH(pl, &players, next) {
	if (obj->pl == pl)
	    continue;
	known = find_pl_knowledge(pl, obj);
	if (known != NULL) {
	    tell_new_obj(pl, eq, NO_EXTENDED, NO_DRAW);
	    tell_equip_obj(pl, obj, eq, eq->cli.equip.loc,
			   eq->cli.equip.visible);
	}
    }
    pthread_mutex_unlock(&players_mutex);
    return(1);
}

/**
   \brief tell all the players, something was unworn
   \param event event that called us
   \return event status (allways 1)
*/

int obj_unwore_obj(event_t *event)
{
    object_t *obj, *eq;
    player_t *pl;
    pl_known_t *known;

    obj = event->initiator;  /* the object doing the wearing */
    eq = evarg_getobj(event,0); /* the first obj in involved is the EQ */

    pthread_mutex_lock(&players_mutex);
    TAILQ_FOREACH(pl, &players, next) {
	if (obj->pl == pl)
	    continue;
	known = find_pl_knowledge(pl, obj);
	if (known != NULL) {
	    tell_rmobj(pl, eq);
	    tell_unequip_obj(pl, obj, eq, eq->cli.equip.visible);
	}
    }
    pthread_mutex_unlock(&players_mutex);
    return(1);
}

/**
   \brief update a player with knowledge of obj
   \return 1 if changed values, 0 for new obj
   \param pl player to update
   \param obj object to inform player about
*/

int update_pl_knowledge(player_t *pl, object_t *obj)
{
    pl_known_t *known;
    int new=0;
    unsigned int bits=0;

    known = find_pl_knowledge(pl, obj);
    if (known == NULL) {
	known = smalloc(pl_known_t);
	new++;
    }
    if (new || !uid_eq(&known->uid,&obj->cli.uid)) {
	uid_cpy(&known->uid,&obj->cli.uid);
	bits |= PLKN_NUM;
    }
    if (new || known->state != obj->cli.state) {
	known->state = obj->cli.state;
	bits |= PLKN_STATE;
    }
    /* this is not a bug */
    if (new || known->name != obj->cli.name) {
	known->name = add_string(obj->cli.name);
	bits |= PLKN_NAME;
    }
    if (new || known->name_pl != obj->cli.name_pl) {
	known->name_pl = add_string(obj->cli.name_pl);
	bits |= PLKN_NAMEPL;
    }
    if (new || known->model != obj->cli.model) {
	known->model = add_string(obj->cli.model);
	bits |= PLKN_MODEL;
    }
    if (new || known->velocity != obj->cli.velocity) {
	known->velocity = obj->cli.velocity;
	bits |= PLKN_VELOC;
    }
    if (new || known->facing.x != obj->cli.facing.x ||
	known->facing.y != obj->cli.facing.y ||
	known->facing.z != obj->cli.facing.z) {
	known->facing = obj->cli.facing;
	bits |= PLKN_FACING;
    }
    if (new || known->direction.x != obj->cli.direction.x ||
	known->direction.y != obj->cli.direction.y ||
	known->direction.z != obj->cli.direction.z) {
	known->direction = obj->cli.direction;
	bits |= PLKN_DIR;
    }
    if (new || known->coord.x != obj->cli.coord.x ||
	known->coord.y != obj->cli.coord.y ||
	known->coord.z != obj->cli.coord.z) {
	known->coord = obj->cli.coord;
	bits |= PLKN_COORD;
    }
    if (new || known->scale.x != obj->cli.scale.x ||
	known->scale.y != obj->cli.scale.y ||
	known->scale.z != obj->cli.scale.z) {
	known->scale = obj->cli.scale;
	bits |= PLKN_SCALE;
    }
    known->lu = 10;
    if (new) {
	if (TAILQ_EMPTY(&pl->known))
	    TAILQ_INIT(&pl->known);
	TAILQ_INSERT_TAIL(&pl->known, known, next);
	known->bits = 0;
	return(0);
    }
    known->bits = bits;
    return(1);
}

/**
   \brief remove knowledge from the player's knowledge database
   \param known what knowledge to remove
   \param pl what player to remove it from
*/

void remove_player_knowledge(pl_known_t *known, player_t *pl)
{
    tell_rmobjbyid(pl, &known->uid);
    free_string(known->name);
    free_string(known->name_pl);
    free_string(known->model);
    TAILQ_REMOVE(&pl->known, known, next);
    free(known);
}

/**
   \brief Have the player glance around, to see what is out there
   This transfers nothing, but requires a bit of math. (1 sqrt per
   dist_between() )  On a modern CPU I consider this no big deal.
   \param event event that called us
   \return event status
   \note only checks the current maph.  should be extended to look across the
   border if we are close.
*/

int player_perc_surrounds(event_t *event)
{
    object_t *obj, *tmp, *eq;

    obj = event->initiator;

    TAILQ_FOREACH(tmp, &obj->maphead->map->objs, next) {
	if (obj == tmp || tmp->cli.model == NULL)
	    continue;
	if (dist_between(tmp->cli.coord, obj->cli.coord) <
	    vision_dist(obj->cli.stats.Per))
	    if (update_pl_knowledge(obj->pl, tmp) == 0) {
		tell_new_obj(obj->pl, tmp, NO_EXTENDED, DO_DRAW);
		/* we need to tell the client about the EQ of other objs */
		TAILQ_FOREACH(eq, &tmp->bucket.equip, stack) {
		    tell_new_obj(obj->pl, eq, NO_EXTENDED, NO_DRAW);
		    tell_inv_obj(obj->pl, eq, tmp, BUCKET_EQUIP);
		    tell_equip_obj(obj->pl, tmp, eq, eq->cli.equip.loc,
				   eq->cli.equip.visible);
		}
	    }
    }
    return(0);
}

/**
   \brief tell all players in range about me
   \param event event that called us
   \return event status (allways 1)
*/

int object_teach_players_self(event_t *event)
{
    object_t *obj, *eq;
    player_t *pl;
    event_t *tpsc;

    obj = event->initiator;

    pthread_mutex_lock(&players_mutex);
    TAILQ_FOREACH(pl, &players, next) {
	float dist;

	if (pl->obj == obj || pl->obj->mapgroup != obj->mapgroup)
	    continue;
	dist = dist_between(obj->cli.coord, pl->obj->cli.coord);
	if (dist < vision_dist(pl->obj->cli.stats.Per)) {
	    if (update_pl_knowledge(pl, obj) == 0) {
		tell_new_obj(pl, obj, NO_EXTENDED, DO_DRAW);
		TAILQ_FOREACH(eq, &obj->bucket.equip, stack) {
		    tell_new_obj(pl, eq, NO_EXTENDED, NO_DRAW);
		    tell_inv_obj(pl, eq, obj, BUCKET_EQUIP);
		    tell_equip_obj(pl, obj, eq, eq->cli.equip.loc,
				   eq->cli.equip.visible);
		}
	    }
	    else if (lod_dist(pl->obj->cli.stats.Per, dist) * 100.0 >
		     (float)rndm(1,100)) { 
		/* else schedule a player tell now 
		   We use LOD dist to update things farther away less
		   frequently than things close to us.
		 */
		tpsc = new_obj_event(Q_PLAYER, 0, "tell_player_surr_changes",
				     pl->obj);
		tpsc->action = 1;
		queue_obj_event(tpsc);
	    }
	}
    }
    pthread_mutex_unlock(&players_mutex);
    return(1);
}

/**
   \brief Tell the player about things in his surroundings that have changed
   \param event event that called us
   \return event status
   \note event->action determines if this is recurring or not
   New things should be handled elsewhere.  This should run fairly often, as it
   is normally a nop for the network traffic, and not so hard on the engine.
*/

int tell_player_surr_changes(event_t *event)
{
    pl_known_t *known, *knext;
    object_t *obj;
    char buf[256], buf2[256],b2[40];

    obj = event->initiator;

    for (known=TAILQ_FIRST(&obj->pl->known); known != NULL; known = knext) {
	knext = TAILQ_NEXT(known, next);
	if (!event->action)
	    known->lu--;
	if (known->bits > 0) {
	    sprintf(buf, "chob id:%s", uid_sprint(b2,&known->uid));
	    if (known->bits & PLKN_NROF) {
		sprintf(buf2, " nrof:%d", known->nrof);
		strcat(buf, buf2);
	    }
	    if (known->bits & PLKN_STATE) {
		sprintf(buf2, " st:%d", known->state);
		strcat(buf, buf2);
	    }
	    if (known->bits & PLKN_NAME) {
		sprintf(buf2, " nm:\"%s\"", known->name);
		strcat(buf, buf2);
	    }
	    if (known->bits & PLKN_NAMEPL) {
		sprintf(buf2, " nmp:\"%s\"", known->name_pl);
		strcat(buf, buf2);
	    }
	    if (known->bits & PLKN_MODEL) {
		sprintf(buf2, " mdl:\"%s\"", known->model);
		strcat(buf, buf2);
	    }
	    if (known->bits & PLKN_FACING) {
		sprintf(buf2, " fx:%0.1f fy:%0.1f fz:%0.1f", known->facing.x,
			known->facing.y, known->facing.z);
		strcat(buf, buf2);
	    }
	    if (known->bits & PLKN_DIR) {
		sprintf(buf2, " dx:%0.1f dy:%0.1f dz:%0.1f",known->direction.x,
			known->direction.y, known->direction.z);
		strcat(buf, buf2);
	    }
	    if (known->bits & PLKN_COORD) {
		sprintf(buf2, " cx:%0.1f cy:%0.1f cz:%0.1f", known->coord.x,
			known->coord.y, known->coord.z);
		strcat(buf, buf2);
	    }
	    if (known->bits & PLKN_SCALE) {
		sprintf(buf2, " sx:%f sy:%f sz:%f", known->scale.x,
			known->scale.y, known->scale.z);
		strcat(buf, buf2);
	    }
	    if (known->bits & PLKN_VELOC) {
		sprintf(buf2, " v:%0.2f", known->velocity);
		strcat(buf, buf2);
	    }
	    strcat(buf, " draw:1\r");
	    lock_do_write(obj->pl->d, buf);
	    known->bits = 0;
	} else if (known->bits < 1 && known->lu <= 0) {
	    remove_player_knowledge(known, obj->pl);
	}
    }
    return(event->action);
}

/**
   \brief Do a collision based look for surrounding objects.
   \param event event that called us
   \return event status
   \warning calls dtTest()
   \note event->action determines if this is recurring or not
   \todo make the distance perception based
   This shouldn't be called very often, as it is more cpu intense.  Perhaps
   as a "search" functionality.
*/

int player_find_surrounds(event_t *event)
{
    object_t *coll;

    if (event->initiator->pl == NULL )
	return(1);

    coll = create_finder_sphere(event->initiator->pl, 20.0);
    dtTest();
    remove_collision_obj(coll);
    return(event->action);
}

/**
   \brief queue up the standard set of events that a new player needs
   \param obj player object
*/

void queue_player_finders(object_t *obj)
{
    event_t *event;

    event = new_obj_event(Q_PLAYER, ETYPE_TIMER, "player_perc_surrounds", obj);
    event->dnob = 100;
    queue_obj_event(event);

    event = new_obj_event(Q_PLAYER, ETYPE_TIMER, "tell_player_surr_changes", obj);
    event->dnob = 50;
    event->action = 0;
    queue_obj_event(event);

    event = new_obj_event(Q_PLAYER, ETYPE_TIMER,
			  "player_check_border_distance", obj);
    event->dnob = 60;
    event->action = 0;
    queue_obj_event(event);

    event = new_obj_event(Q_PLAYER, ETYPE_NONE, "send_motd", obj);
    queue_event(event, NULL);
}

/**
   \warning this function needs to be rewritten
  Add the player to the map.  Send all the data to the socket the player
  needs to establish a connection.
*/

void add_player_map(object_t *obj) 
{
    mapgroup_t *mapg;
    maphead_t *maph;

    /* put him on the first active map with starting flag set we can find */
    TAILQ_FOREACH(mapg, &maps, next) {
        pthread_mutex_lock(&mapg->mutex);
        TAILQ_FOREACH(maph, &mapg->maps, next) {
            if (QUERY_FLAG(maph, MAPFLAG_LOADED) &&
		QUERY_FLAG(maph, MAPFLAG_STARTING)) {
		send_player_mapg(obj, mapg);
		pthread_mutex_unlock(&mapg->mutex);
		return;
            }
        }
        pthread_mutex_unlock(&mapg->mutex);
    }

}

/**
   \warning this function needs to be rewritten
  In reality, this procedure will need to be far more complex.  For now, we
  make it simple, and we can redo it later.  Note for now we just won't save
  any characters, so passwords and loading is not an issue.
  Assumes d is locked on entry.
*/

int add_player_game(descriptor_data_t *d, char *username, char *charname,
		    char *password)
{
    player_t *pl, *tpl;
    object_t *obj;
    char buf[1024];

    pthread_mutex_lock(&players_mutex);
    TAILQ_FOREACH(tpl, &players, next) {
	if (!strcmp(charname, tpl->charname)) {
	    pthread_mutex_unlock(&players_mutex);
	    sprintf(buf, "no plexists:0\r");
	    do_write(d, buf);
	    return(0); /* name in use */
	}
    }
    pthread_mutex_unlock(&players_mutex);
    pl = new_player();
    pl->charname = add_string(charname);
    pl->username = add_string(charname); /* XXX */
    pl->d = d;
    d->pl = pl;
    obj = new_obj_from_arch("new_player",DO_NOREC); /*XXX */
    if (obj->cli.nrof == 0)
	obj->cli.nrof = 1;
    obj->pl = pl;
    obj->cli.name = add_string(charname);
    obj->cli.name_pl = add_string(charname);
    pl->obj = obj;
    pthread_mutex_lock(&players_mutex);
    TAILQ_INSERT_TAIL(&players, pl, next);
    pthread_mutex_unlock(&players_mutex);
    nonpersist.numusers++;
    if (nonpersist.numusers > persist.userhwm)
	persist.userhwm = nonpersist.numusers;
    
    add_player_map(obj);
    tell_servdefs(pl);

    return(nonpersist.objectcount);
}

/**
   \brief send an event to put a player on a mapg
   \param obj object to send to map
   \param mapg mapg to send object to
*/

void send_player_mapg(object_t *obj, mapgroup_t *mapg) {
    event_t *event;
    
    event = new_obj_event(Q_PLAYER, ETYPE_NONE, "player_enter_mapg", obj);
    queue_event(event, mapg);
}

/**
   \brief make a player enter a mapg
   \param event event that called us
   \return event status (allways 1)
*/

int player_enter_mapg(event_t *event) {
    player_t *pl;
    object_t *obj;
    object_t *sloc;
    int found=0, ofound=0;
    char buf[1024],b2[40];
    mapgroup_t *mapg;
    maphead_t *maph;

    obj = event->initiator;
    pl = obj->pl;

    if (obj->mapgroup) /* MUST be null */
	return(1);
    if(!event_mapg(event))  /* can't be NULL */
	return(1);
    mapg = obj->mapgroup = event_mapg(event);

    
    /* lock mapg ? */
    TAILQ_FOREACH(maph, &mapg->maps, next) {
         if (QUERY_FLAG(maph, MAPFLAG_LOADED) &&
	    QUERY_FLAG(maph, MAPFLAG_STARTING) && !found) {
	    found++;
	    TAILQ_FOREACH(sloc, &maph->map->objs, next) {
		if (!strcmp(sloc->arch, "starting_location")) {
		    obj->cli.coord.x = sloc->cli.coord.x;
		    obj->cli.coord.z = sloc->cli.coord.z;
		    obj->cli.coord.y = sloc->cli.coord.y;
		    ofound++;
		}
	    }
	    if (!ofound) {
		obj->cli.coord.x = maph->x+5.0;
		obj->cli.coord.z = maph->z+5.0;
		obj->cli.coord.y = -99999.0;
	    }
	    put_obj_on_map_closeto(obj, maph, POOM_SPACE);
	}
    }
     /* now we have to tell the player we created him a new char and put it on
     the map */
    tell_new_obj(pl, obj, DO_EXTENDED, DO_DRAW);
    tell_new_map(pl, 0);
    /* we have to recurse into his inventory, and describe it to him */
    /* but we don't currently load players from a file, so why bother? */
    sprintf(buf, "ur id:%s\r", uid_sprint(b2,&obj->cli.uid));
    lock_do_write(pl->d, buf);
    queue_player_finders(obj);
    return 1;
};

/**
   \brief remove a player from the game
   \param event event that called us
   \return event status (allways 1)
*/

int remove_player_game(event_t *event)
{
    player_t *pl;
    object_t *obj;

    obj = event->initiator;
    pl = obj->pl;


    if ( pl == NULL)
	return(1);

    pthread_mutex_lock(&players_mutex);
    TAILQ_REMOVE(&players, pl, next);
    pthread_mutex_unlock(&players_mutex);

    remove_obj_from_map(pl->obj, pl->obj->maphead, POOM_SPACE, 1);
    if (pl->username)
	free_string(pl->username);
    if (pl->charname)
	free_string(pl->charname);
    if (pl->obj)
	remove_obj(pl->obj);
    free(pl);

    nonpersist.numusers--;

    return(1);
}    

/**
   \brief shift what maps the player knows about, given a direction of shift
   \param pl player to shift known maps for
   \param dir direction to shift
*/
void shift_known_maps(player_t *pl, int dir)
{
    int old;

    old = pl->known_maps;

    if (!dir)
	return;

    pl->known_maps = 0;

    if (dir+4 > 8)
	pl->known_maps |= (1<<(dir-4));
    else
	pl->known_maps |= (1<<(dir+4));

    if (dir == 1) {
	if (old & (1<<8))
	    pl->known_maps |= (1<<6);
	if (old & (1<<2))
	    pl->known_maps |= (1<<4);
    } else if (dir == 2) {
	if (old & (1<<1))
	    pl->known_maps |= (1<<8);
	if (old & (1<<3))
	    pl->known_maps |= (1<<4);
    } else if (dir == 3) {
	if (old & (1<<2))
	    pl->known_maps |= (1<<8);
	if (old & (1<<4))
	    pl->known_maps |= (1<<6);
    } else if (dir == 4) {
	if (old & (1<<3))
	    pl->known_maps |= (1<<2);
	if (old & (1<<5))
	    pl->known_maps |= (1<<6);
    } else if (dir == 5) {
	if (old & (1<<4))
	    pl->known_maps |= (1<<2);
	if (old & (1<<6))
	    pl->known_maps |= (1<<8);
    } else if (dir == 6) {
	if (old & (1<<7))
	    pl->known_maps |= (1<<8);
	if (old & (1<<5))
	    pl->known_maps |= (1<<4);
    } else if (dir == 7) {
	if (old & (1<<8))
	    pl->known_maps |= (1<<2);
	if (old & (1<<6))
	    pl->known_maps |= (1<<4);
    } else if (dir == 8) {
	if (old & (1<<1))
	    pl->known_maps |= (1<<2);
	if (old & (1<<7))
	    pl->known_maps |= (1<<6);
    }
}

/**
   \brief inform players about maps on the border
   \param pl player to inform
   \param mapnum what map to inform player about given as direction
*/

void inform_border_maps(player_t *pl, int mapnum)
{
    /* handle the four corner maps */
    if (mapnum == 1) {
	if (!(pl->known_maps & (1<<8))) {
	    tell_new_map(pl, 8);
	    pl->known_maps |= (1<<8);
	}
	if (!(pl->known_maps & (1<<1))) {
	    tell_new_map(pl, 1);
	    pl->known_maps |= (1<<1);
	}
	if (!(pl->known_maps & (1<<2))) {
	    tell_new_map(pl, 2);
	    pl->known_maps |= (1<<2);
	}
    } else if (mapnum == 3) {
	if (!(pl->known_maps & (1<<2))) {
	    tell_new_map(pl, 2);
	    pl->known_maps |= (1<<2);
	}
	if (!(pl->known_maps & (1<<3))) {
	    tell_new_map(pl, 3);
	    pl->known_maps |= (1<<3);
	}
	if (!(pl->known_maps & (1<<4))) {
	    tell_new_map(pl, 4);
	    pl->known_maps |= (1<<4);
	}
    } else if (mapnum == 5) {
	if (!(pl->known_maps & (1<<4))) {
	    tell_new_map(pl, 4);
	    pl->known_maps |= (1<<4);
	}
	if (!(pl->known_maps & (1<<5))) {
	    tell_new_map(pl, 5);
	    pl->known_maps |= (1<<5);
	}
	if (!(pl->known_maps & (1<<6))) {
	    tell_new_map(pl, 6);
	    pl->known_maps |= (1<<6);
	}
    } else if (mapnum == 7) {
	if (!(pl->known_maps & (1<<6))) {
	    tell_new_map(pl, 6);
	    pl->known_maps |= (1<<6);
	}
	if (!(pl->known_maps & (1<<7))) {
	    tell_new_map(pl, 7);
	    pl->known_maps |= (1<<7);
	}
	if (!(pl->known_maps & (1<<8))) {
	    tell_new_map(pl, 8);
	    pl->known_maps |= (1<<8);
	}
    } else {
	if (!(pl->known_maps & (1<<mapnum))) {
	    tell_new_map(pl, mapnum);
	    pl->known_maps |= (1<<mapnum);
	}
    }
}

/**
   \brief test to see if the player is in range of another map
   \param event the event that called us
   \return event status
   test player's known_maps to see if the player has wandered out of range
   of any of the border maps.  If so, tell the client we don't need them
   anymore, and update known_maps.
*/

int player_check_border_distance(event_t *event)
{
    player_t *pl;
    object_t *obj;
    int i;
    double d;
    char buf[256];

    obj = event->initiator;
    pl = obj->pl;


    d = vision_dist(obj->cli.stats.Per) * 2.0;

    for (i=1; i < 9; i++) {
	if (pl->known_maps & (1<<i))
	    if (!in_range_from_border(obj, i, d)) {
		pl->known_maps &= ~(1<<i);
		sprintf(buf, "mapd act:%d\n", i);
		lock_do_write(pl->d, buf);
	    }
    }
    return(0);
}

/**
   \brief handle a modpl comamnd
   \param event event that called us
   \return event status (allways 1)
*/

int player_modpl(event_t *event)
{
    object_t *obj;

    obj = event->initiator;

    if (obj->cli.state != STATE_FALLING &&
	obj->cli.state != STATE_DEAD)
	obj->cli.state = event->action;
    return(1);
}

/**
   \brief handle a rotate command
   \param event event that called us
   \return event status (allways 1)
*/

int player_rotate(event_t *event)
{
    object_t *obj;

    obj = event->initiator;

    if (evarg_isset(event,0))
	obj->cli.facing.x = evarg_getfloat(event,0);
    if (evarg_isset(event,1))
	obj->cli.facing.y = evarg_getfloat(event,1);;
    if (evarg_isset(event,2))
	obj->cli.facing.z = evarg_getfloat(event,2);;
    if (evarg_isset(event,3))
	obj->cli.direction.x = evarg_getfloat(event,3);;
    if (evarg_isset(event,4))
	obj->cli.direction.y = evarg_getfloat(event,4);;
    if (evarg_isset(event,5))
	obj->cli.direction.z = evarg_getfloat(event,5);;
    return(1);
}

/**
   \brief update the players that are supposed to be moving
   \param event the event that called us
   \return event status
*/

int player_update_all_motion(event_t *event)
{
    player_t *pl;
    object_t *obj;
    char buf[256], buf2[256],b2[40];
    event_t *etel;
    maphead_t *maph;
    int border;

    pthread_mutex_lock(&players_mutex);
    TAILQ_FOREACH(pl, &players, next) {
	obj = pl->obj;
	if (obj == NULL || obj->mapgroup != event_mapg(event)) {
	    continue;
	}
	if (obj->cli.state > STATE_IDLE) {
	    if (attempt_move_obj(obj)) {
		/* we succeed, tell the player */
		sprintf(buf, "chob id:%s",uid_sprint(b2,&obj->cli.uid));
		if (!same_within(obj->cli.coord.x, obj->cli.prevc.x, 10.0)) {
		    sprintf(buf2, " cx:%0.1f", obj->cli.coord.x);
		    strcat(buf, buf2);
		}
		if (!same_within(obj->cli.coord.y, obj->cli.prevc.y, 10.0)) {
		    sprintf(buf2, " cy:%0.1f", obj->cli.coord.y);
		    strcat(buf, buf2);
		}
		if (!same_within(obj->cli.coord.z, obj->cli.prevc.z, 10.0)) {
		    sprintf(buf2, " cz:%0.1f", obj->cli.coord.z);
		    strcat(buf, buf2);
		}
		sprintf(buf2, "\r");
		strcat(buf, buf2);
		lock_do_write(pl->d, buf);
		/* schedule an event to tell other players we moved */
		etel = new_obj_event(Q_PLAYER, 0, "object_teach_players_self",
				     pl->obj);
		queue_obj_event(etel);
	    } else {
		/* we may have moved but not told prior to the nomv.  Tell. */
		sprintf(buf, "chob id:%s cx:%0.1f cy:%0.1f cz:%0.1f\r",
			uid_sprint(b2,&obj->cli.uid), obj->cli.coord.x, obj->cli.coord.y,
			obj->cli.coord.z);
		lock_do_write(pl->d, buf);
		/* nomv consitutes a state change to IDLE.  The client
		   should just do that on it's own. */
		sprintf(buf, "nomv\r");
		obj->cli.state = STATE_IDLE;
		lock_do_write(pl->d, buf);
	    }
	    maph = obj_on_map(obj);
	    if (maph != obj->maphead) {
		tell_mapshift(pl, maph);
		obj->maphead = maph;
	    }
	    border = obj_close_to_border(obj);
	    if (border)
		inform_border_maps(pl, border);
	}
    }
    pthread_mutex_unlock(&players_mutex);
    return(0);
}

