/* $Id: sutils.c,v 1.12 2004/03/30 06:45:55 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <inttypes.h>
#include <time.h>
#include <math.h>
#include <sys/param.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "object.h"
#include "map.h"
#include "utils.h"
#include "sutils.h"

/** \file sutils.c
    \brief Server Utility Functions

 * The random functions here take luck into account when rolling random
 * dice or numbers.  This function has less of an impact the larger the
 * difference becomes in the random numbers.  IE, the effect is lessened
 * on a 1-1000 roll, vs a 1-6 roll.  This can be used by crafty programmers,
 * to specifically disable luck in certain rolls, simply by making the
 * numbers larger (ie, 1d1000 > 500 vs 1d6 > 3)
 */

/**
   \brief Roll a random number between min and max
   \param min minimum result
   \param max maximum result
   \param obj object to base luck roll off
   \param goodbad prefer high rolls, or low rolls? See PREFER_*
   \return the int result

 * Roll a random number between min and max.  Uses obj to determine luck,
 * and if goodbad is non-zero, luck increases the roll, if zero, it decreases.
 * Generally, obj should be the player/caster/hitter requesting the roll,
 * not the recipient (ie, the poor slob getting hit).
 */

int random_roll(int min, int max, object_t *obj, int goodbad)
{
    int omin, diff, luck, base;

    omin = min;
    diff = max - min + 1;
    ((diff > 2) ? (base = 20) : (base = 50)); /* d2 and d3 are corner cases */

    if (max < 1 || diff < 1) {
      LOG(LOG_ERROR, "Calling random_roll with min=%d max=%d", min, max);
      return(min); /* avoids a float exception */
    }

    if (obj->cli.stats.type != T_PLAYER)
	return((RANDOM()%diff)+min);

    luck = obj->cli.stats.luck;
    if (RANDOM()%base < MIN(10, abs(luck))) {
	/* we have a winner */
	((luck > 0) ? (luck = 1) : (luck = -1));
	diff -= luck;
	if (diff < 1)
	    return(omin); /*check again*/
	((goodbad) ? (min += luck) : (diff));

	return(MAX(omin, MIN(max, (RANDOM()%diff)+min)));
    }
    return((RANDOM()%diff)+min);
}

/**
   \brief Roll a number of dice (2d3, 4d6)
   \param num number of dice to roll
   \param size size of dice to roll
   \param obj object to base luck roll off
   \param goodbad prefer high rolls, or low rolls? See PREFER_*
   \return the int result

 * Roll a number of dice (2d3, 4d6).  Uses obj to determine luck,
 * If goodbad is non-zero, luck increases the roll, if zero, it decreases.
 * Generally, obj should be the player/caster/hitter requesting the roll,
 * not the recipient (ie, the poor slob getting hit).
 * The args are num D size (ie 4d6)
 */

int die_roll(int num, int size, object_t *obj, int goodbad)
{
    int min, diff, luck, total, i, gotlucky, base;

    diff = size;
    min = 1;
    luck = total = gotlucky = 0;
    ((diff > 2) ? (base = 20) : (base = 50)); /* d2 and d3 are corner cases */
    if (size < 2 || diff < 1) {
      LOG(LOG_ERROR, "Calling die_roll with num=%d size=%d", num, size);
      return(num); /* avoids a float exception */
    }

    if (obj->cli.stats.type == T_PLAYER)
	luck = obj->cli.stats.luck;

    for (i = 0; i < num; i++) {
	if (RANDOM()%base < MIN(10, abs(luck)) && !gotlucky) {
	    /* we have a winner */
	    gotlucky++;
	    ((luck > 0) ? (luck = 1) : (luck = -1));
	    diff -= luck;
	    if (diff < 1)
		return(num); /*check again*/
	    ((goodbad) ? (min += luck) : (diff));
	    total += MAX(1, MIN(size, (RANDOM()%diff)+min));
	} else {
	    total += RANDOM()%size+1;
	}
    }
    return(total);
}

/** 
    \brief get the mapnum inside the mapgroup that corresponds to the map for x,z
    \param maph the maphead to check
    \param x the x coordinate
    \param z the z coordinate
    \return the map number that corresponds to x,z
*/

int mapnum_at_xz(maphead_t *maph, float x, float z)
{
    int mapnum, zz, xx;
    uint64_t ix, iz;

    ix = (uint64_t)ceil(x);
    iz = (uint64_t)ceil(z);
    zz = iz/((maph->width-1)*SCALE_XZ);
    xx = ix/((maph->height-1)*SCALE_XZ);

    mapnum = zz*maph->mapg->width+xx;
    return(mapnum);
}

/**
   \brief return the groundheight at the current x,z, given a maphead_t
   \param maph the maphead to check
   \param x the x coordinate
   \param z the z coordinate
   \return the groundheight y coordinate
*/

float groundheight_at_mapxz(maphead_t *maph, float x, float z)
{
    int mapnum;
    maphead_t *tmap;

    if (xz_out_of_mapgroup(maph, x, z))
	return(0.0);

    if (!xz_out_of_map(maph, x, z)) {
	return(groundheight_at_xz(maph->map->elevation, x-maph->x, z-maph->z,
				  maph->width, maph->height)+maph->y);
    } else {
	mapnum = mapnum_at_xz(maph, x, z);
	tmap = find_map_bynum(mapnum, maph->mapg);
	/* we have to load this map, as we are about to reference it */
	set_map_loaded(tmap);
	return(groundheight_at_xz(tmap->map->elevation, x-tmap->x, z-tmap->z,
				  tmap->width, tmap->height)+tmap->y);
    }
}

/**
   \brief Initialize a mutex
   \note the mutex MUST NOT be locked or valid
   \param mutex the mutex to initialize
   \return int 0 for success.  bailout for failure

   Attempts to initialize a mutex to type PTHREAD_MUTEX_RECURSIVE. This type
   allows a thread to recursively lock a mutex, and recursively unlock that
   mutex.
*/

int init_mutex(pthread_mutex_t *mutex)
{
    pthread_mutexattr_t attr;

    if (pthread_mutexattr_init(&attr) != 0)
	bailout();
 
#ifndef _PTH_PTHREAD_H_
 #ifdef __linux__
    if (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE_NP) != 0)
	bailout();
 #else
    if (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE) != 0)
	bailout();
 #endif
#endif

    if (pthread_mutex_init(mutex, &attr) != 0)
	bailout();

    return(0);
}
