/* $Id: tec1.c,v 1.14 2004/03/26 06:23:15 kstenger Exp $ */

/* This program is Copyright (c) 1990 David Allen.  It may be freely
   distributed as long as you leave my name and copyright notice on it.
   I'd really like your comments and feedback; send e-mail to
   davea@vlsi.ll.mit.edu, or send us-mail to David Allen, 10 O'Moore Ave,
   Maynard, MA 01754. 
   
   Copyright (C) 2003 Karla Stenger
   Copyright (C) 2003 Genecys Developer Team
*/

/* This is the file containing all of the important functions except for
   trysplit (), which splits a continent into pieces.  Also, all of the main
   arrays are declared here, even a couple that are only used by functions in
   tec2.c.  The array declarations are first, followed by the sequencing
   function onestep () and some miscellaneous routines including the text
   output routines; initialization routines and the routines that do all
   the interesting stuff are last. */


#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#include "const.h"
#include "var.h"
#include "d_queue.h"
#include "plates.h"

#include "defines.h"
#include "utils.h"

extern int rndm(int min, int max);

/* These defines are used for the PostScript output section of tecst(). */
#define D ((double) 432 / ((XSIZE > YSIZE) ? XSIZE : YSIZE))
#define ZMAX 128
#define XX(x) (108 + ((x) * D))
#define YY(y) (612 - ((y) * D))
#define ZZ(z) (1 - (float) ((z > ZMAX) ? ZMAX : z) / (ZMAX))
#define TOGGLE(a) (a=!(a))
#define MIN(x,y) (((x) > (y)) ? (y) : (x))

#define OCEAN 255

/* The following arrays are global and most are used by functions in both
   source files.  The two main ones are m and t.  Each is set up to be two
   2-d arrays, where each array is the size of the whole world.  M is the
   map; elements in m are indices of plates, showing which squares are
   covered by which plate.  T is the topography; elements in t are altitudes. */

int** m[2];
int** t[2];
int** b; /* initial blob */
int XBLOB, YBLOB;

/* Several arrays are used by the binary blob segmenter, segment() in tec2.c.
   These include r, which is used to store fragment indices; many fragments
   make up one region during a segmentation.  Kid is a lookup table; fragment
   k belongs to region kid[k] after a segmentation is finished.  Karea[k]
   is the area of fragment k. */

int** r;
int* kid;
int* karea;

/* The merge routine gets information from the move routine; when the move
   routine puts a square of one plate on top of another plate, that information
   is recorded in the merge matrix mm. */

int** mm;

/* The erosion routine needs an array to store delta information; during an
   erosion, the increases or decreases in elevation are summed in e and then
   applied all at once to the topography. */

int** e;

/* Several routines need temporary storage for areas and plate identifiers. */

int* tarea;
int* ids;


/* Step counter. */
int step;


/* prototypes */
extern void panic (char *s);
extern void draw (int src);
extern void trysplit (int src);
extern int segment (int src, int match, int *frag, int *reg);


void makefrac (int src, int match);
void draw_bar (struct dot* d, int length);
void move (int src, int dest);
void onestep (void);
int plate_add (void);
void plate_remove (int n);
void tecst (int src, int drawmode, int index);
void init (void);
int singlefy (int src, int match);
void newdxy (void);
void merge (int dest);
int bump (int dest, int a, int b);
void onerode (int dest, int i, int j, int di, int dj);
void erode (int dest);
int** alloc_array_XY(int x, int y);
/*void plates_alloc(void);
void plates_free(void);
int plate_add (void);
struct plate *get_plate(int n);
void plate_remove (int n);*/


int** alloc_array_XY(int x, int y){ 
    /* allocates a 2D array of x*y integers */
    int **b= safer_malloc(x * sizeof(int *));
    int j;
    for (j=0; j<x; j++)
	b[j] = safer_malloc(y * sizeof(int));
    return b;
}


void onestep (void) {
    /* This is the sequencing routine called by main once per step.
       It just calls the important subfunctions in order:
       - trysplit   finds a plate to break up, and computes new velocities
       - newdxy     computes the deltas to move each plate this step
       - move       moves the plates
       - merge      determines results when plates rub together
       - erode      erodes the terrain, adding or subtracting at the margins
       - draw       draw the resulting array once every DRAWEVERY steps
       The m and t arrays are double-buffered in the sense that operations go
       from m[0] to m[1] or vice-versa; src and dest determine which is which. */

    int src, dest;
    if (!step)
	printf( "--- Steps:");
    printf( " %d", step);
    fflush(NULL);

    src = step % 2;
    dest = 1 - src;
    
    if (rndm (0,100) < RIFTPCT)
	trysplit (src);
    
    newdxy ();
    move (src, dest);
    merge (dest);
    
    if (DOERODE)
	erode(dest);
    if (step==MAXSTEP){
	erode(dest);
	erode(dest);
    }

    if (!(step % DRAWEVERY))
	draw (dest);
    
    if (step==MAXSTEP) {
	free(m[0]); 
	free(m[1]);
	free(t[0]);
	free(t[1]);
	free(karea);
	free(kid);
	free(mm);
	free(tarea);
	free(ids);
	plates_free();
    }
}



void tecst (int src, int drawmode, int index) {
    /* This function is called whenever map output is called for.  It looks
       at the parameter `drawmode' to decide between long text, simple text,
       and PostScript output formats.  Note that the default for this
       function is no output at all, corresponding to DRAWMODE_NONE. */

    int i, j, k;
    char filename[15];
    FILE *fp;
    
    /*for (i=0; i<XSIZE; i++)
	for (j=0; j<YSIZE; j++)
	    if ((t[0][i][j]!=0) && (t[0][i][j]!=ZINIT)) 
		printf("o");*/

    if (drawmode == DRAWMODE_GENERIC) 
	  for (j=0; j < YSIZE; j++) 
		for (i=0, k=0; i < XSIZE; i++) {
		      printf ("%4d", t[src][i][j]);
		      if (!(++k % 18))
			  printf ("\n");
		}
    else if (drawmode == DRAWMODE_PGM) {
	  /* determine filename */
	  (void) sprintf (filename, "image%03d.pgm", index);
	  fp = fopen (filename, "w");
	  
	  /* write PGM header */
	  fprintf (fp, "P2\n%d %d\n%d\n", (int)XSIZE, (int)YSIZE, 255);
	  for (i=0; i < XSIZE; i++) 
	      for (j=0; j < YSIZE; j++) {
		  fprintf (fp, "%3d\n", t[src][i][j]);
		  //if ((t[src][i][j]!=0) && (t[src][i][j]!=ZINIT)) printf("x");/**** this should give no output if we call genterrain -t0, that is is we just make a blob ****/
	      }
	  (void) fclose (fp);
    }
    else if (drawmode == DRAWMODE_TEXT) {
	  for (j=0; j < YSIZE; j++) {
		for (i=0; i < XSIZE; i++) {
		      if ((k = t[src][i][j]) < ZCOAST)
			  k = 0;
		      else if (k > ZMOUNTAIN)
			  k = 2;
		      else
			  k = 1;
		      printf ("%d", k);
		}
		printf ("\n");
	  }
	  printf ("\n");
    }
    else if (drawmode == DRAWMODE_GRAY) {
	  printf ("%%!PS-Adobe-1.0\n/d %4.2f def\n", D);
	  printf ("37.5 45 { dup mul exch dup mul add 1 exch sub } setscreen\n");
	  printf ("/r { setgray moveto d 0 rlineto 0 d rlineto ");
	  printf ("d neg 0 rlineto fill } bind def\n");
	  printf ("%6.2f %6.2f moveto\n", XX (-1), YY (-1));
	  printf ("%6.2f %6.2f lineto\n", XX (XSIZE), YY (-1));
	  printf ("%6.2f %6.2f lineto\n", XX (XSIZE), YY (YSIZE));
	  printf ("%6.2f %6.2f lineto\n", XX (-1), YY (YSIZE));
	  printf ("%6.2f %6.2f lineto\nstroke\n", XX (-1), YY (-1));

	  for (j=0; j < YSIZE; j++)
	      for (i=0; i < XSIZE; i++)
		  if ((k = t[src][i][j]) > ZCOAST)
		      printf ("%6.2f %6.2f %5.4f r\n", XX (i), YY (j), ZZ (k));
	  printf ("\nshowpage\n");
    }
}


void init (void) {
    /* This is the catchall function that initializes everything.  First,
       it links together the plates onto the free list and starts the used list
       at empty.  The first plate is created by a fractal technique and then
       improved.  Finally, the fractal is copied to the data array and drawn.
       There are two kinds of improvement done here.  First, islands are
       eliminated by segmenting the blob and erasing all the regions except
       for the biggest.  Second, oceans inside the blob (holes) are eliminated
       by segmenting the _ocean_ and filling in all regions except the biggest. */

    int besti, x;
    int i, j;

    /* Allocate space for every global array */
    XBLOB = XSIZE*XPERC/100 -(XPERC/100);
    YBLOB = YSIZE*YPERC/100 -(YPERC/100);
    b     = alloc_array_XY(XBLOB, YBLOB);
    
    m[0] = alloc_array_XY(XSIZE, YSIZE);
    m[1] = alloc_array_XY(XSIZE, YSIZE);
    r    = alloc_array_XY(XSIZE, YSIZE);
    e    = alloc_array_XY(XSIZE, YSIZE);
    t[0] = alloc_array_XY(XSIZE, YSIZE);
    t[1] = alloc_array_XY(XSIZE, YSIZE);
    
    karea= safer_malloc(MAXFRAG * sizeof(int *));
    kid  = safer_malloc(MAXFRAG * sizeof(int *));
    
    tarea= safer_malloc(MAXPLATE * sizeof(int *));
    ids  = safer_malloc(MAXPLATE * sizeof(int *));
    mm   = alloc_array_XY(MAXPLATE, MAXPLATE);
     
    /* Allocate the plates structure. Get a first plate to use for the blob */
    plates_alloc();
    x = plate_add();
    
    /* Make a blob. Copy into m. Free b after we are done with it. */
    makefrac (0, x);
    free(b);
    
    /* Segment m[0] looking for x, set besti to the largest region, */
    /* and zero out all the other regions.  This eliminates islands. */
    besti = singlefy (0, x);
    if (besti > 0)
	for (i = 0; i < XSIZE; i++)
	    for (j = 0; j < YSIZE; j++)
		if (kid[r[i][j]] != besti)
		    m[0][i][j] = 0;
    
    /* Segment m[0] looking for 0 (ocean), set besti to the largest region, */
    /* and fill in all the other regions.  This eliminates holes in the blob. */
    besti = singlefy (0, 0);
    if (besti > 0)
	for (i=0; i < XSIZE; i++)
	    for (j=0; j < YSIZE; j++)
		if (kid[r[i][j]] != besti)
		    m[0][i][j] = x;
    
    /* Fill the topo structure with the blob shape while finding its area */
    for (i=0; i < XSIZE; i++)
	for (j=0; j < YSIZE; j++)
	    if (m[0][i][j]) {
		t[0][i][j] = ZINIT;
		p.first->area++;
	      }
    
    /* Draw the blob */
    draw (0);
}


void makefrac (int src, int match) {
    /* Creates a blob of max size XBLOB x YBLOB placing a random dot and
     * creating an initial bar in a certain direction, then chooses some dots
     * of the bar and repeats the process with a different direction starting
     * at those dots. 
     *
     * Squares in the blob are set to the value in `match'. */

    int i, j, r1, r2, x, y;
    int xofs, yofs;    /* offsets for locating the blob into the map */
    
    int length;	    /* length of drawn bars */

    struct dots *q = d_new_queue();	/* queue of dots */
    struct dot  *d = d_new_dot(rndm(XBLOB*3/10, YBLOB*7/10),
			       rndm(YBLOB*3/10, YBLOB*7/10),
			       rndm(0, 1));
    d_queue_dot(d, q);
    
    r1 = rndm(XBLOB*YBLOB/4, XBLOB*YBLOB/4+1 );
    for (i=0; i < r1; i++) { 
	if ((d = d_dequeue_dot(q))) {
	    /* length is restricted to the maximum diameter of an elipse
	     * centered in the current dot and that fits into the blob.
	     * Discard dots on the borders of the blob */
	    if (d->dir) /* h */
		if ((d->x == 0) || (d->x == (XBLOB-1)))
		    continue;
		else
		    length = rndm(1, MIN(d->x, (XBLOB - d->x)));
	    else	/* v */
		if ((d->y == 0) || (d->y == (YBLOB-1)))
		    continue;
		else
		    length = rndm(1, MIN(d->y, (YBLOB - d->y)));
		
	    draw_bar(d, length);
	    
	    x = d->x;
	    y = d->y;
	    TOGGLE(d->dir);
	    r2 = rndm(1, 8);
	    for (j=0; j < r2; j++) { /* add dots to the queue */
		if (!(d->dir)) {
		    do    /* h */
		    d->x = rndm(x -length/2, x +length/2 -((length%2)?0:1));
		    while ((d->x == x) && (length > 2));    
		}
		else {
		    do	  /* v */
		    d->y = rndm(y -length/2, y + length/2 -((length%2)?0:1));
		    while ((d->y == y) && (length > 2));
		}
		d_queue_dot(d, q);
	    }
	}   
    }
    while (q->fst) { 
	i++;
	d = d_dequeue_dot(q);
	if ((length = rndm(0, MIN(MIN(d->x, XBLOB - d->x ),
			          MIN(d->y, YBLOB - d->y )) + 1 )))
	    draw_bar(d, length);
    }
    free(d);
    free(q);
    
    /* Find the position (offset) of blob[0][0] in the whole map randomly */
    xofs = rndm((XSIZE-XBLOB)/4, (XSIZE-XBLOB)*3/4);
    yofs = rndm((YSIZE-YBLOB)/4, (YSIZE-YBLOB)*3/4);
    
    /* If height is greater than BLOBLEVEL, there will be land there. */
    for (i=0; i<XBLOB; i++)
	for (j=0; j<YBLOB; j++)
	    if (b[i][j] > BLOBLEVEL)
		m[src][i + xofs][j + yofs] = match;
}


void draw_bar(struct dot* d, int length){
    /* draws a bar centered on d increasing the altitude of each square 
     * corresponding to it. The direction is set in the dot itself. */
    int i,j;
    
    if (length){
	if (d->dir)  /* horizontal */
	    for (i=(d->x -length/2); i<(d->x +length/2 -((length%2)?0:1)); i++) 
		for (j=0; j<3; j++)
		    if ((i >= 0) && (i < XBLOB))
			b[i][d->y]++;
		    else
			b[d->x +length/2 -((length%2)?0:1) -i][d->y]++;
	    
	else	 /* vertical */
	    for (i=(d->y -length/2); i<(d->y +length/2 -((length%2)?0:1)); i++)
		for (j=0; j<3; j++)
		    if ((i >= 0) && (i < YBLOB))
			b[d->x][i]++;
		    else
			b[d->x][d->y +length/2 -((length%2)?0:1) -i]++;
    }
}


int singlefy (int src, int match) {
    /* This is a subfunction of init() which is called twice to improve the
       fractal blob.  It calls segment() and then interprets the result.  If
       only one region was found, no improvement is needed; otherwise, the
       area of each region must be computed by summing the areas of all its
       fragments, and the index of the largest region is returned. */

    int i, reg, frag, besti, besta;

    if (!(segment (src, match, &frag, &reg)))
	return -1;
    if (reg == 1)
	return -1; /* No improvement needed */

    /* Initialize the areas to zero, then sum frag areas */
    for (i=1; i <= reg; i++)
	tarea[i] = 0;
    for (i=1; i <= frag; i++)
	tarea[kid[i]] += karea[i];

    /* Pick largest area of all regions and return it */
    for (i=1, besta=0, besti=0; i <= reg; i++)
	if (besta < tarea[i]) {
	      besti = i;
	      besta = tarea[i];
	  }
    return besti;
}


void newdxy (void) {
    /* For each plate, compute how many squares it should move this step.
       Multiply the plate's basic movement vector odx,ody by the age modifier
       MR[], then add the remainders rx,ry from the last move to get some large
       integers.  Then turn the large integers into small ones by dividing by
       REALSCALE and putting the remainders back into rx,ry.  This function also
       increases the age of each plate, but doesn't let the age of any plate
       go above MAXLIFE.  This is done to make sure that MR[] does not need to
       be a long vector. */

    struct plate *pl = p.first;
    int a, i;

    for (i=1; i < MAXPLATE; i++) {
	if (pl->in_use) { 
	    a =(pl->odx * MR[pl->age]) + pl->rx;
	    pl->dx = a / REALSCALE;
	    pl->rx = a % REALSCALE;
	
	    a =(pl->ody * MR[pl->age]) + pl->ry;
	    pl->dy = a / REALSCALE;
	    pl->ry = a % REALSCALE;
	
	    if (pl->age < MAXLIFE - 1)
		pl->age++;
	}
	pl = pl->next;
    }
}


void move (int src, int dest) {
    /* This function moves all the plates that are drifting.  The amount to
       move by is determined in newdxy().  The function simply steps through
       every square in the array; if there's a plate in a square, its new location
       is found and the topography is moved there.  Overlaps between plates are
       detected and recorded so that merge() can resolve the collision; mountain
       growing is performed.  If two land squares wind up on top of each other,
       folded mountains are produced. */

    int i, j, x, y;
    struct plate *pl_a, *pl_c; /* Plates */

    /* Clear out the merge matrix and the destination arrays */
    for (i=1; i < MAXPLATE; i++)
	for (j=1; j < MAXPLATE; j++)
	    mm[i][j] = 0;
    for (i=0; i < XSIZE; i++)
	for (j=0; j < YSIZE; j++) {
	    m[dest][i][j] = 0;
	    t[dest][i][j] = 0;
	}

    /* Look at every square which belongs to a plate */
    for (i=0; i < XSIZE; i++)
	for (j=0; j < YSIZE; j++)
	    if (m[src][i][j] > 0) {
		if (!(pl_a = get_plate(m[src][i][j])))
		    printf("plate Nr.%d is NULL.\n",m[src][i][j]);
		
		/* Get the square's new location. */
		x = pl_a->dx + i;
		y = pl_a->dy + j;
		
		if ((x >= XSIZE) || (x < 0) || (y >= YSIZE) || (y < 0)) 
		    /* It is off the map so throw it away */
		    pl_a->area--;
		else {  	/* It IS on the map */
		    /* If the destination is occupied, remove the other guy
		     * remembering the two plates overlapped.
		     * Set the new height */
		    if (m[dest][x][y]) { /* overlapped */
			pl_c = get_plate(m[dest][x][y]);
			
			mm[pl_a->id][pl_c->id]++;
			pl_c->area--;
			t[dest][x][y] += t[src][i][j]*2 ;
			if (t[dest][x][y] > 255)
			    t[dest][x][y] = 255 - t[dest][x][y]%255;
		    } else {	/* Destination is free. */
			t[dest][x][y] = t[src][i][j];
		    }
		    
		    /* Prevent erosion from beeing to bad by increasing the 
		     * plate randomly, depending on how hard we are eroding*/
		    if (step%(ERODERND*ERODERND)) 
			t[dest][x][y]+=rndm(0,1);
		    
		    /* Plate A now owns this square */
		    m[dest][x][y] = pl_a->id;
		}
	    }
}


void merge (int dest) {
    /* Since move has set up the merge matrix, most of the work is done.  This
       function calls bump once for each pair of plates which are rubbing; note
       that a and b below loop through the lower diagonal part of the matrix.
       One subtle feature is that a plate can bump with several other plates in
       a step; suppose that the plate is merged with the first plate it bumped.
       The loop will try to bump the vanished plate with the other plates, which
       would be wrong.  To avoid this, the lookup table lut is used to provide
       a level of indirection.  When a plate is merged with another, its lut
       entry is changed to indicate that future merges with the vanished plate
       should be applied to the plate it has just been merged with. */

    int* lut = safer_malloc(MAXPLATE * sizeof(int *));
    int a, aa, b, bb, i;

    for (a=1; a < MAXPLATE; a++)
	lut[a] = a;
    for (a=2; a < MAXPLATE; a++)
	for (b=1; b < a; b++)
	    if (mm[a][b] || mm[b][a]) {
		  aa = lut[a];
		  bb = lut[b];
		  if (aa != bb)
		      if (bump (dest, aa, bb)) {
			    lut[aa] = bb;
			    for (i=1; i < MAXPLATE; i++)
				if (lut[i] == aa)
				    lut[i] = bb;
			}
	      }
}


int bump (int dest, int a, int b) {
    /* Plates a and b have been moved on top of each other by some amount;
       alter their movement rates for a slow collision, possibly merging them.
       The collision "strength" is a ratio of the area overlap (provided by
       move ()) to the total area of the plates involved.  A fraction of each
       plate's current movement vector is subtracted from the movement vector
       of the other plate.  If the two vectors are now within some tolerance
       of each other, they are almost at rest so merge them with each other. */

    double maa, mab, ta, tb, rat, area;
    struct plate *pl_a = get_plate(a);
    struct plate *pl_b = get_plate(b);
    int i, j, x;

    /* Find a ratio describing how strong the collision is */
    x = mm[pl_a->id][pl_b->id] + mm[pl_b->id][pl_a->id];
    area = pl_a->area + pl_b->area;
    rat = x / (MAXBUMP + area/20);
    if (rat > 1.0)
	rat = 1.0;

    /* Do some math to update the move vectors.  This looks complicated */
    /* because a plate's actual movement vector must be multiplied by */
    /* MR[age], and because I have rewritten the equations to maximize */
    /* use of common factors.  Trust me, it's just inelastic collision. */
    maa = pl_a->area * MR[pl_a->age];
    mab = pl_b->area * MR[pl_b->age];
    ta = MR[pl_a->age] * area;
    pl_a->odx = (pl_a->odx * maa + pl_b->odx * mab * rat) / ta;
    pl_a->ody = (pl_a->ody * maa + pl_b->ody * mab * rat) / ta;
    tb = MR[pl_b->age] * area;
    pl_b->odx = (pl_b->odx * mab + pl_a->odx * maa * rat) / tb;
    pl_b->ody = (pl_b->ody * mab + pl_a->ody * maa * rat) / tb;

    /* For each axis, compute the remaining relative velocity.  If it is */
    /* too large, return without merging the plates */
    if (abs(pl_a->odx * MR[pl_a->age] - pl_b->odx * MR[pl_b->age]) > BUMPTOL)
	return 0;
    if (abs(pl_a->ody * MR[pl_a->age] - pl_b->ody * MR[pl_b->age]) > BUMPTOL)
	return 0;

    /* The relative velocity is small enough, so merge the plates.  Replace */
    /* all references to a with b, free a, and tell merge() a was freed. */
    for (i=0; i < XSIZE; i++)
	for (j=0; j < YSIZE; j++)
	    if (m[dest][i][j] == pl_a->id)
		m[dest][i][j] = pl_b->id;
    pl_b->area += pl_a->area;
    plate_remove(pl_a->id);
    return pl_a->id;
}

void erode (int dest) {
   /* This function takes the topography in t[dest] and smooths it, lowering
   mountains and raising lowlands and continental margins.  It does this by
   stepping across the entire array and doing a computation once for each
   pair of 8-connected pixels.  The computation is done by onerode(), below.
   The computation result for a pair is a small delta for each square, which
   is summed in the array e.  When the computation is finished, the delta
   is applied; if this pushes an ocean square high enough, it is added to
   an adjacent plate if one can be found.  Also, if a land square is eroded
   low enough, it is turned into ocean and removed from its plate. */
 
   int i, j, x, z, xx;
   struct plate *pl;
 
   /* Zero out the array for the deltas first */
   for (i=0; i<XSIZE; i++) 
       for (j=0; j<YSIZE; j++) 
	   e[i][j] = 0;
   /*checkmouse();*/

   /* Step across the entire array; each pixel is adjacent to 8 others, and */
   /* it turns out that if four pairs are considered for each pixel, each */
   /* pair is considered exactly once.  This is important for even erosion */
   for (i=1; i<XSIZE; i++) 
       for (j=1; j<YSIZE; j++) {
	   onerode (dest, i, j,  0, -1);
	   onerode (dest, i, j, -1, -1);
	   onerode (dest, i, j, -1,  0);
	   if (i < XSIZE-1) 
	       onerode (dest, i, j, -1, 1); 
       }
 
   /* Now go back across the array, applying the delta values from e[][] */
   for (i=0; i<XSIZE; i++) 
       for (j=0; j<YSIZE; j++) {
	   z = t[dest][i][j] + e[i][j];
	   if (z < 0) z = -z;
	   if (z > 255) z = 255 - z%255;
	   
	   /* If the square just rose above shelf level, look at the four */
	   /* adjacent squares.  If one is a plate, add the square to that plate */
	   if ((z >= ZSHELF) && (t[dest][i][j] < ZSHELF)) { 
	       xx = 0;
	       if (i > 1)       if ((x = m[dest][i-1][j])) xx = x;
	       if (i < XSIZE-1) if ((x = m[dest][i+1][j])) xx = x;
	       if (j > 1)       if ((x = m[dest][i][j-1])) xx = x;
	       if (j < YSIZE-1) if ((x = m[dest][i][j+1])) xx = x;
	       if (xx) { 
		   pl = get_plate(xx);
		   pl->area++; 
		   m[dest][i][j] = xx; 
	       }
	   }
 
	   /* Add the increment to the old value; if the square is lower than */
	   /* shelf level but belongs to a plate, remove it from the plate */
	   t[dest][i][j] = z;
	   
	   if ((z < ZSHELF) && (x = m[dest][i][j])) {
	       pl = get_plate(x);
	       pl->area--;
	       m[dest][i][j] = 0;
	   }
       }
}


void onerode (int dest, int i, int j, int di, int dj) {
    /* This function is called once per pair of squares in the array.  The
    amount each square loses to an adjacent but lower square in each step is
    one-eighth the difference in altitude.  The code must be repeated to avoid
    shifting negative numbers.  Also, since erosion is reduced below the
    waterline, if an altitude is lower than ZERODE, ZERODE is used instead. */
    
    int z, t1, t2;
    
    t1 = t[dest][i][j];
    t2 = t[dest][i+di][j+dj];
    z = 0;
    if ((t1 > t2) && (t1 > ZERODE)) {
	z = (t1 - t2 + ERODERND) / 16; 
    } else if ((t2 > t1) && (t2 > ZERODE)) {
	z = -(t2 - t1 + ERODERND) / 16; 
    } else if((t2 == t1) && t1) {
	z = rndm(0,4);
    }

    

    if (step==MAXSTEP) {
	if ((t1 > t2) && (t1 > ZERODE)) {
	    z += (t1 - t2 + 4) / 16;
	} else if ((t2 > t1) && (t2 > ZERODE)) {
	    z += -(t2 - t1 + 4) / 16;
	}
    }

    if (z) { 
	e[i][j] -= z; 
	e[i+di][j+dj] += z; 
    } 
}

